<?php

namespace Controller;

use Exception;
use Zeedhi\Framework\DTO\Response\Message;
use Zeedhi\Framework\DTO\Response;
use Zeedhi\Framework\DTO\Request;
use Zeedhi\Framework\DataSource\DataSet;

class Order extends \Zeedhi\Framework\Controller\Simple
{
    // --- Messages 
    const ERROR_MESSAGE     = 'E';
    const GENERAL_MESSAGE   = 'A';

    // --- Address
    const PRINCIPAL_ADDRESS = 'P';

    // --- Table/Positions 
    const REQUEST_ACCOUNT   = 'S';
    const BUSY              = 'O';
    const PAID              = 'P';

    const SCHEDULED_ORDER         = 'A';
    const PENDING_SCHED_ORDER     = 'N';
    const NOT_PENDING_SCHED_ORDER = 'S';

    const MAX_CONTROL_FILE_UPDATE_TIME = 60;

    protected $util;
    protected $orderService;
    protected $geocodeURL;
    protected $printMode;
    protected $impressaoUtil;
    protected $menuService;
    protected $projectUrl;
    protected $maxParallelClose;
    protected $executionDebug;
    protected $maxSaleCancelTime;

    public function __construct(
        \Util\Util $util,
        \Service\Order $orderService,
        $geocodeURL,
        \Zeedhi\Framework\DependencyInjection\InstanceManager $instanceManager,
        \Odhen\API\Lib\ImpressaoUtil $impressaoUtil,
        \Service\Menu $menuService
    ) {

        date_default_timezone_set($instanceManager->getParameter('TIMEZONE'));

        $this->util              = $util;
        $this->orderService      = $orderService;
        $this->geocodeURL        = $geocodeURL;
        $this->printMode         = $instanceManager->getParameter('PRINT_MODE');
        $this->systemPath        = $instanceManager->getParameter('SYSTEM_PATH');
        $this->logPath           = $instanceManager->getParameter('LOG_PATH');
        $this->projectUrl        = $instanceManager->getParameter('PROJECT_URL');
        $this->maxParallelClose  = $instanceManager->getParameter('MAX_PARALLEL_CLOSE');
        $this->executionDebug    = !empty($instanceManager->getParameter('EXECUTION_DEBUG')) ? $instanceManager->getParameter('EXECUTION_DEBUG') : false;
        $this->maxSaleCancelTime = !empty($instanceManager->getParameter('MAX_SALE_CANCEL_TIME')) ? $instanceManager->getParameter('MAX_SALE_CANCEL_TIME') : 30;
        $this->impressaoUtil     = $impressaoUtil;
        $this->menuService       = $menuService;
    }

    /*
        Salva uma venda delivery/toGo
    */
    public function saveDeliveryOrder(Request $request, Response $response)
    {
        $inputJSON = file_get_contents('php://input');
        $input     = json_decode($inputJSON, true);

        try {

            if (!isset($input))
                throw new \Exception('Não foram recebidos parâmetros');
        } catch (\Exception $e) {
            $message = 'Erro ao inserir comanda delivery. ' . $e->getMessage();
            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
            return;
        }

        $data   = array();
        $params = $input;

        try {

            $taaSaleId = isset($input['taaSaleId']) ? $input['taaSaleId'] : null;

            if (!empty($taaSaleId)) {
                $orderById = $this->orderService->getOrderById($taaSaleId);

                if (!empty($orderById)) {

                    if ($this->util->databaseIsOracle())
                        $format = 'Y-m-d H:i:s';
                    else
                        $format = 'Y-m-d H:i:s.u';

                    $orderTime = \DateTime::createFromFormat($format, $orderById['DTHRABERMESA']);
                    $orderTime = $orderTime->format($format);

                    throw new \Exception("Pedido $taaSaleId já existente no banco de dados. Pedido feito em $orderTime");
                }
            }
        } catch (\Exception $e) {
            $response->addMessage(new Message($e->getMessage(), self::GENERAL_MESSAGE));
            $response->addDataSet(
                new DataSet(
                    '',
                    array(
                        array(
                            'saleId'   => $orderById['NRCOMANDAEXT'],
                            'saleCode' => $orderById['NRCOMANDA']
                        )
                    )
                )
            );
            return;
        }

        $order    = $params['order'];
        $fileName = date("Ymd-H-i-s") . "-";
        $pagerId  = isset($order['pagerId']) ? $order['pagerId'] : null;
        $fileName .= empty($pagerId) ? $order['subsidiaryCode'] : $pagerId;

        $this->util->saveLogOrder($inputJSON, 'Order/' . date("Y/m/d"), $fileName);
        $this->util->beginTransaction();

        try {
            $subsidiaryCode   = $order['subsidiaryCode'];
            $comandavenSeq    = isset($order['saleCode']) ? $order['saleCode'] : null;
            $dscomanda        = isset($order['orderId']) ? $order['orderId'] : null;
            $nrmesa           = isset($order['tableNumber']) ? $order['tableNumber'] : null;
            $tableData        = $this->orderService->getTableData($nrmesa, $subsidiaryCode);
            $comandavenData   = $this->orderService->getComandaData($comandavenSeq, $subsidiaryCode);

            if (empty($comandavenData) && empty($tableData)) {
                $vendarestSeq = $this->orderService->generateVendarestSeq($subsidiaryCode);
                $comandavenSeq = $this->orderService->generateComandavenSeq($subsidiaryCode);
            }

            $this->util->commit();
        } catch (\Exception $e) {
            $this->util->rollBack();
            $message = 'Erro ao gerar novo codigo. ' . $e->getMessage();
            $this->util->saveLogOrder($message, 'Error/' . date("Y/m/d"), $fileName);
            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
            return;
        }

        $now = strtotime("now");

        $this->util->beginTransaction();

        try {
            $paymentOnlyWithCashback = false;

            $consumerCode     = !empty($order['consumerCode']) ? $order['consumerCode'] : null;
            $clientCode       = $order['clientCode'];
            $billParams       = $order['orderDescription'];
            $cashBack         = isset($billParams['cashBack']) ? $billParams['cashBack'] : null;
            $orderType        = $billParams['orderType'];
            $orderDiscount    = isset($billParams['orderDiscount']) ? $billParams['orderDiscount'] : 0;
            $holdTime         = isset($billParams['holdTime']) ? $billParams['holdTime'] : 0;
            $productsValue    = $billParams['productsValue'];
            $remaingValue     = $billParams['productsValue'];
            $toTravel         = !empty($billParams['toTravel']);
            $driveThru        = !empty($billParams['driveThru']);
            $orderDescription = isset($billParams['description']) ? $billParams['description'] : null;
            $sellerCode       = isset($order['sellerCode']) ? $order['sellerCode'] : null;
            $closeByApi       = isset($order['closeByApi']) ? $order['closeByApi'] : false;
            $orderCPF = $billParams['orderCPF'];
            $items            = $order['items'];

            $closeByApi = !empty($closeByApi) ? 'S' : 'N';

            $realItemsQuant = 0;
            for ($i = 0; $i < count($items); $i++) {
                $realItemsQuant += $this->divideItems($items, $items[$i]);
            }

            $standardTable = null;
            if (!empty($nrmesa)) {
                $this->orderService->validateTable($nrmesa);
            }

            if (!empty($cashBack))
                $this->validateCashBack($clientCode, $consumerCode, $cashBack, $subsidiaryCode, $orderCPF);
            /*
                Para que seja entendido e o código abaixo não seja retirado até o #~#, é necessário entender(tentar) a lógica utilizada.
                
                Para que uma venda seja válida, TODO item tem que ter um valor de pelo menos 0.01 centavo.
                
                Para saber qual valor máximo de desconto(DESCONTO+CASHBACK) que o pedido pode ter, basta descobrir a quantidade de items existentes
                no pedido e dividir por 100, o resultado de (qtdItems/100) irá retornar a quantidade de items representados em centavos.
                
                SE o valor do CASHBACK + DESCONTO - quantidade de items em centavos for > que o valor de todos os produtos
                o CASHBACK é somado ao DESCONTO já existente, isso valida casos como: produtos custam 50 reais, temos 20 reais
                de desconto, 15 reais de cashback, e 7 items, então o máximo de cashback possível pra esse pedido é 
                50 - 20 - 7/100 que daria 29.93, 29.93 é o valor máximo de cashback possível
                SENÃO respeitar a regra acima, o CASHBACK a ser aplicado é recuperado da seguinte forma: VALORPRODUTOS - DESCONTO
                - quantidade de items em centavos.
                DESCONTO recebe DESCONTO + valor encontrado acima.
                e o CASHBACK quer será utilizado para salvar no campo VRDESCFID da tabela COMANDAVEN que informa quanto de valor
                cashback foi utilizado no pedido, é dado por CASHBACK que será aplicado + quantidade de itens em centavos.
                EXEMPLO: 12 ITEMS com valor total de 50 reais, desconto de 20 reais, com cashback de 60. este caso não irá respeitar
                a primeira regra, portanto: 50 - 20 - 12/100 = 29.88 ESTE é o valor máximo de CASHBACK que será aplicado,
                pode ser valido por 29.88 + 20 = 29.88 sobrando assim 0.01 centavo pra cada item.
                E o valor que será salvo na VRDESCFID é basicamente os 29.88 + 0.12 que dá 30. Este seria o valor de cashback
                utilizado caso pudéssemos zerar uma venda. 50-20-30 = 0.
            */

            $items = $this->recalculateComboDiscount($subsidiaryCode, $items);

            $realItemsQuant = $this->util->div($realItemsQuant, 100);
            if (!empty($cashBack)) {
                if ($this->util->add($cashBack, $orderDiscount) < $productsValue) {
                    $orderDiscount = $this->util->add($orderDiscount, $cashBack);
                } else {
                    $paymentOnlyWithCashback = true;
                    $cashToApply = $this->util->subtract($this->util->subtract($productsValue, $orderDiscount), $realItemsQuant);
                    $orderDiscount = $this->util->add($orderDiscount, $cashToApply);
                    $userCashBack = $cashBack;
                    $cashBack = $this->util->add($cashToApply, $realItemsQuant);


                    if ($userCashBack < $this->util->add($cashBack, $billParams['deliveryFee'])) {
                        $cashBack = $userCashBack;
                    } else {
                        $cashBack = $this->util->add($cashBack, $billParams['deliveryFee']);
                    }
                }
            }
            /* #~# */
            $remaingDiscount = $orderDiscount;
            // --- Order items
            // --- Order payments
            $payments      = $billParams['payments'];
            $giftCards     = isset($billParams['giftCards']) ? $billParams['giftCards'] : array();

            if (!empty($payments) || !empty($giftCards) || !empty($cashBack))
                $this->validatePayment($billParams['totalValue'], $payments, $giftCards, $cashBack);

            // --- Consumer description
            $consumerDescription = isset($order['consumerId']) ? $order['consumerId'] : null;

            // verifica a existência da filial
            if (!$this->util->subsidiaryExists($subsidiaryCode)) {
                throw new \Exception('Filial inexistente.');
            }

            // if (!empty($cashBack) && (empty($clientCode) && empty($consumerCode))) {
            //     throw new \Exception('Não pode ser utilizado saldo fidelidade sem consumidor identificado.');
            // }

            if (!empty($clientCode) && !empty($consumerCode)) {
                if (!$this->util->consumerExists($clientCode, $consumerCode)) {
                    throw new \Exception('Consumidor inexistente.');
                }
            } else {
                $defaultClient = $this->util->getDefaultClient($subsidiaryCode);

                if (!empty($defaultClient))
                    $clientCode = $defaultClient;
                else
                    throw new \Exception('Cliente padrão inexistente.');
            }

            $orderType     = $this->validateOrderType($orderType, $toTravel, $driveThru);
            $storeCode     = null;
            $sellerPos     = null;
            $idposobsped   = null;
            $idlugarmesa   = null;

            if (isset($order['storeCode'])) {
                $storeCode   = $order['storeCode'];
                $storeReturn = $this->util->getStoreDetails($subsidiaryCode, $storeCode);

                if (!empty($storeReturn)) {
                    if (empty($standardTable)) $standardTable = $storeReturn['NRMESA'];
                    $idposobsped = $storeReturn['IDPOSOBSPED'];
                    $idlugarmesa = $storeReturn['IDLUGARMESA'];
                }
            }

            if (isset($order['posCode'])) {
                $posCode   = $order['posCode'];
                $sellerPos = $posCode;
            }

            if (empty($sellerCode))
                $sellerReturn = $this->orderService->getStandardSeller($subsidiaryCode);
            else
                $sellerReturn = $this->orderService->getVendorInfo($sellerCode);

            if (!$sellerReturn['error']) {

                $standardSeller = $sellerReturn['data'];
                $sellerCode     = $standardSeller['CDVENDEDOR'];
                $operatorCode   = $standardSeller['CDOPERADOR'];

                if ($storeCode == null)
                    $storeCode = $standardSeller['CDLOJA'];

                if ($sellerPos == null)
                    $sellerPos = $standardSeller['CDCAIXA'];

                if ($standardTable == null)
                    $standardTable = $standardSeller['NRMESA'];

                if ($idposobsped == null)
                    $idposobsped = $standardSeller['IDPOSOBSPED'];

                if ($idlugarmesa == null)
                    $idlugarmesa = $standardSeller['IDLUGARMESA'];

                $orderExists = false;

                $table = !empty($nrmesa) ? $nrmesa : $standardTable;

                if (!empty($dscomanda))
                    $oderByDscomanda = $this->orderService->getOrderDscomanda($dscomanda, $subsidiaryCode);

                $tableData = $this->orderService->getTableData($nrmesa, $subsidiaryCode);
                $comandavenData = $this->orderService->getComandaData($comandavenSeq, $subsidiaryCode);

                if (!empty($comandavenData)) {
                    $vendarestSeq = $comandavenData['NRVENDAREST'];
                    $comandavenSeq = $comandavenData['NRCOMANDA'];
                    $orderExists = true;
                } else if (!empty($tableData)) {
                    $vendarestSeq = $tableData['NRVENDAREST'];
                    $comandavenSeq = $tableData['NRCOMANDA'];
                    $orderExists = true;
                } else if (!empty($oderByDscomanda)) {
                    $vendarestSeq = $oderByDscomanda['NRVENDAREST'];
                    $comandavenSeq = $oderByDscomanda['NRCOMANDA'];
                    $orderExists = true;
                } else {
                    $returnVendarest = $this->insertVendaRest($subsidiaryCode, $vendarestSeq, $storeCode, $sellerCode, $operatorCode, $clientCode, $consumerCode, $table);

                    if ($returnVendarest['error'])
                        $response->addMessage(new Message($returnVendarest['message'], self::ERROR_MESSAGE));

                    if (!empty($nrmesa))
                        $this->orderService->updateTableStatus($subsidiaryCode, $storeCode, $nrmesa);
                }

                if ($paymentOnlyWithCashback === true) {
                    $cdtiporece     = $this->util->getMoney();
                    $movcaixadlvSeq = str_pad(1, 10, '0', STR_PAD_LEFT);
                    $paymentValue   = $this->util->add($realItemsQuant, $billParams['deliveryFee']);

                    $this->insertMovCaixaDlv($subsidiaryCode, $vendarestSeq, $movcaixadlvSeq, 'E', $clientCode, $consumerCode, $cdtiporece, $paymentValue, null, 1, null);
                }

                if ($this->isToGo($orderType)) {
                    $toGo = 'S';
                    $subsidiaryAddress = $this->util->getSubsidiaryAddress($subsidiaryCode, self::PRINCIPAL_ADDRESS);

                    if (!empty($subsidiaryAddress)) {
                        $cityCode      = $subsidiaryAddress['CDMUNICIPIO'];
                        $state         = $subsidiaryAddress['SGESTADO'];
                        $country       = $subsidiaryAddress['CDPAIS'];
                    } else {
                        $cityCode      = null;
                        $state         = null;
                        $country       = null;
                    }

                    $zipCode           = null;
                    $streetAddress     = null;
                    $neighborhood      = null;
                    $number            = null;
                    $reference         = null;
                    $phone             = null;
                    $cellphone         = null;
                    $addressLine2      = null;
                    $fullStreetAddress = null;
                    $deliveryFee       = 0;
                } else {
                    $toGo           = 'N';
                    $addressDetails = $billParams['deliveryAddress'];
                    $state          = $addressDetails['state'];
                    $cityCodeExt    = $addressDetails['city'];
                    $cityCode       = $this->util->getCityCode($cityCodeExt, $state);
                    $zipCode        = $addressDetails['zipCode'];
                    $streetAddress  = $addressDetails['streetAddress'];
                    $neighborhood   = $addressDetails['neighborhood'];
                    $number         = $addressDetails['number'];
                    $reference      = $addressDetails['reference'];
                    $phone          = $addressDetails['phone'];
                    $cellphone      = $addressDetails['cellphone'];
                    $addressLine2   = $addressDetails['addressLine2'];
                    $deliveryFee    = (float)$billParams['deliveryFee'];
                    $fullStreetAddress = $streetAddress . ', ' . $number;
                    $country        = '0055';
                    $phoneToUpdate  = !empty($phone) ? $phone : $cellphone;
                    $this->util->updateConsumerPhone($clientCode, $consumerCode, $phoneToUpdate);
                }

                $idcontroprod = $this->getIdcontroprod($storeCode, $subsidiaryCode);
                $IDSTCOMANDA  = '3';

                if ($idcontroprod === 'S')
                    $IDSTCOMANDA = '1';

                if (empty($taaSaleId)) {
                    $orderToken = sha1($subsidiaryCode . $vendarestSeq . $comandavenSeq . $clientCode . $consumerCode);
                } else {
                    $orderToken = $taaSaleId;
                }

                $scheduledDate = null;
                $isScheduled   = $billParams['isScheduled'];
                $scheduledDate = $isScheduled ? $billParams['scheduledDate'] : $scheduledDate;
                $IDSTCOMANDA   = $isScheduled ? self::SCHEDULED_ORDER : $IDSTCOMANDA;
                $syncScheduled = $isScheduled ? self::PENDING_SCHED_ORDER : self::NOT_PENDING_SCHED_ORDER;
                $kdsChangeStatusOnExp = '0';

                $taaOrder = strpos($orderType, 'TAA') !== false;

                if ($this->printMode == 'EXPEDICAO') {

                    if ($toGo == 'S' || $taaOrder)
                        $kdsChangeStatusOnExp = '1';
                } else if ($this->printMode == 'INICIO') {
                    if ($taaOrder)
                        $kdsChangeStatusOnExp = '2';
                    else if ($toGo == 'S')
                        $kdsChangeStatusOnExp = '1';
                }

                if (empty($dscomanda)) {
                    if ($orderType == "MES_MOB" || $orderType == "CMD_MOB") {
                        $dscomanda = "PKR_$comandavenSeq";
                        $toGo = 'N';
                    } else {
                        $dscomanda = "DLV_$comandavenSeq";
                    }
                } else {
                    $dscomanda = str_pad($dscomanda, 10, 0, STR_PAD_LEFT);
                }
                
                if ($orderExists === false) {
                    $returnComandaven = $this->insertComandaVen($subsidiaryCode, $vendarestSeq, $comandavenSeq, $storeCode, $IDSTCOMANDA, $deliveryFee, $country, $state, $cityCode, $zipCode, $neighborhood, $fullStreetAddress, $addressLine2, $reference, $toGo, $orderToken, $syncScheduled, $scheduledDate, $kdsChangeStatusOnExp, $consumerDescription, $orderType, $cashBack, $orderDescription, $dscomanda, $closeByApi, $orderCPF);
                    if ($returnComandaven['error']) {
                        $message = $returnComandaven['message'];
                        $response->addMessage(new Message($message, self::ERROR_MESSAGE));
                    }
                }

                // --- Save payments
                if (!empty($payments)) {
                    $savePaymentsreturn = $this->savePayments($payments, $subsidiaryCode, $vendarestSeq, $clientCode, $consumerCode, (float)$billParams['changeValue'], true);
                    if ($savePaymentsreturn['error']) {
                        $message = $savePaymentsreturn['message'];
                        throw new \Exception($message);
                    }
                }

                // --- Save gift cards
                if (!empty($giftCards)) {
                    $DTABERCAIX        = $this->util->getPosDate($subsidiaryCode, $sellerPos);
                    $totalValue        = $billParams['totalValue'] - $cashBack;
                    $giftCardsResponse = $this->saveGiftCards($giftCards, $subsidiaryCode, $vendarestSeq, $sellerPos, $DTABERCAIX, $totalValue);

                    if ($giftCardsResponse['error']) {
                        $message = $giftCardsResponse['message'];
                        throw new \Exception($message);
                    }
                }

                // --- save items
                if (count($items) > 0)
                    $itemsSeq = $this->orderService->generateItemsSeq($subsidiaryCode);

                // $orderCPF = $billParams['orderCPF'];
                $itemNum  = 0;
                $positionNumber = '01';

                $toTravelData = array();
                $driveThruData = array();

                // if(!empty($toTravel)) $toTravelData = $this->getToTravelData();
                if (!empty($driveThru)) $driveThruData = $this->getDriveThruData();

                if (!$taaOrder && !empty($pagerId)) {
                    $pagerId = substr($pagerId, 0, 4);
                    $pagerId = 'A' . $pagerId;
                }

                $sumValue = 0;
                $totalDiscount = $orderDiscount;

                while ($orderDiscount > 0) {
                    $sumValue = 0;
                    foreach ($items as &$item) {
                        $this->shareDiscount($item, $orderDiscount, $productsValue, $remaingValue, $remaingDiscount, false);
                    }
                    unset($item);
                    $remaingValue = $productsValue;
                    foreach ($items as $item) {
                        $sumValue += $this->getDiscountValue($item);
                    }
                    $orderDiscount = $this->util->subtract($totalDiscount, $sumValue);
                    $remaingDiscount = $orderDiscount;
                    unset($item);
                }
                $quantItems = $this->orderService->getQuantItens($subsidiaryCode, $vendarestSeq, $comandavenSeq);

                foreach ($items as &$item)
                    $this->saveItem($item, $subsidiaryCode, $vendarestSeq, $comandavenSeq, $itemsSeq, $positionNumber, $idcontroprod, $sellerPos, $storeCode, $sellerCode, $orderCPF, $itemNum, null, $toTravelData, $pagerId, $holdTime, $driveThruData, $orderType, false, $quantItems);

                unset($item);

                if (empty($consumerCode) && !empty($giftCards)) {
                    $this->orderService->updateConsumerVendarest($giftCards[0]['clientCode'], $giftCards[0]['consumerCode'], $vendarestSeq, $subsidiaryCode);
                }

                $this->util->commit();

                if (!empty($consumerDescription) && !empty($pagerId))
                    $pagerId .= " - $consumerDescription";

                $printItems = $this->formatItemsToPrint($items);

                $toGoString = ($toGo == 'S') ? 'PEDIDO PARA RETIRADA NA LOJA' : 'PEDIDO PARA ENTREGA DELIVERY';
                $this->printOrder($subsidiaryCode, $comandavenSeq, $storeCode, $printItems, $sellerCode, $table, $pagerId, 'B', $toGoString, $taaOrder);
                $response->addDataSet(
                    new DataSet(
                        '',
                        array(
                            array(
                                'saleId' => $orderToken,
                                'saleCode' => $comandavenSeq,
                                'pagerId' => $pagerId
                            )
                        )
                    )
                );
            } else {
                $this->util->rollBack();
                $message = $sellerReturn['message'];
                $response->addMessage(new Message($message, self::ERROR_MESSAGE));
            }
        } catch (\Exception $e) {
            $this->util->rollBack();

            $message = 'Erro ao inserir comanda delivery. ' . $e->getMessage();
            $this->util->saveLogOrder($message, 'Error/' . date("Y/m/d"), $fileName);

            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
        }
    }

    public function saveDeliveryOrderEx(Request $request, Response $response)
    {
        $inputJSON = file_get_contents('php://input');
        $input     = json_decode($inputJSON, true);

        try {

            if (!isset($input))
                throw new \Exception('Não foram recebidos parâmetros');
        } catch (\Exception $e) {
            $message = 'Erro ao inserir comanda delivery. ' . $e->getMessage();
            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
            return;
        }

        $data   = array();
        $params = $input;

        try {

            $taaSaleId = isset($input['taaSaleId']) ? $input['taaSaleId'] : null;

            if (!empty($taaSaleId)) {
                $orderById = $this->orderService->getOrderById($taaSaleId);

                if (!empty($orderById)) {

                    if ($this->util->databaseIsOracle())
                        $format = 'Y-m-d H:i:s';
                    else
                        $format = 'Y-m-d H:i:s.u';

                    $orderTime = \DateTime::createFromFormat($format, $orderById['DTHRABERMESA']);
                    $orderTime = $orderTime->format($format);

                    throw new \Exception("Pedido $taaSaleId já existente no banco de dados. Pedido feito em $orderTime");
                }
            }
        } catch (\Exception $e) {
            $response->addMessage(new Message($e->getMessage(), self::GENERAL_MESSAGE));
            $response->addDataSet(
                new DataSet(
                    '',
                    array(
                        array(
                            'saleId'   => $orderById['NRCOMANDAEXT'],
                            'saleCode' => $orderById['NRCOMANDA']
                        )
                    )
                )
            );
            return;
        }

        $order    = $params['order'];
        $fileName = date("Ymd-H-i-s") . "-";
        $pagerId  = (isset($order['pagerId']) && !empty($order['pagerId'])) ? $order['pagerId'] : mt_rand(1000, 9999);
        $fileName .= empty($pagerId) ? $order['subsidiaryCode'] : $pagerId;

        $this->util->saveLogOrder($inputJSON, 'Order/' . date("Y/m/d"), $fileName);

        $this->util->beginTransaction();

        try {
            $subsidiaryCode   = $order['subsidiaryCode'];
            $comandavenSeq    = isset($order['saleCode']) ? $order['saleCode'] : null;
            $dscomanda        = isset($order['orderId']) ? $order['orderId'] : null;
            $nrmesa           = isset($order['tableNumber']) ? $order['tableNumber'] : null;
            $tableData        = $this->orderService->getTableData($nrmesa, $subsidiaryCode);
            $comandavenData   = $this->orderService->getComandaData($comandavenSeq, $subsidiaryCode);

            if (empty($comandavenData) && empty($tableData)) {
                $vendarestSeq = $this->orderService->generateVendarestSeq($subsidiaryCode);
                $comandavenSeq = $this->orderService->generateComandavenSeq($subsidiaryCode);
            }

            $this->util->commit();
        } catch (\Exception $e) {
            $this->util->rollBack();
            $message = 'Erro ao gerar novo codigo. ' . $e->getMessage();
            $this->util->saveLogOrder($message, 'Error/' . date("Y/m/d"), $fileName);
            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
            return;
        }

        $now = strtotime("now");

        $this->util->beginTransaction();

        try {
            $paymentOnlyWithCashback = false;

            $consumerCode     = !empty($order['consumerCode']) ? $order['consumerCode'] : null;
            $clientCode       = $order['clientCode'];
            $billParams       = $order['orderDescription'];
            $cashBack         = isset($billParams['cashBack']) ? $billParams['cashBack'] : null;
            $orderType        = $billParams['orderType'];
            $orderDiscount    = isset($billParams['orderDiscount']) ? $billParams['orderDiscount'] : 0;
            $holdTime         = isset($billParams['holdTime']) ? $billParams['holdTime'] : 0;
            $productsValue    = $billParams['productsValue'];
            $remaingValue     = $billParams['productsValue'];
            $toTravel         = !empty($billParams['toTravel']);
            $driveThru        = !empty($billParams['driveThru']);
            $orderDescription = isset($billParams['description']) ? $billParams['description'] : null;
            $sellerCode       = isset($order['sellerCode']) ? $order['sellerCode'] : null;
            $closeByApi       = isset($order['closeByApi']) ? $order['closeByApi'] : false;
            $items            = $order['items'];

            $closeByApi = !empty($closeByApi) ? 'S' : 'N';

            $realItemsQuant = 0;
            for ($i = 0; $i < count($items); $i++) {
                $realItemsQuant += $this->divideItems($items, $items[$i]);
            }

            $standardTable = null;
            if (!empty($nrmesa)) {
                $this->orderService->validateTable($nrmesa);
            }

            if (!empty($cashBack))
                $this->validateCashBack($clientCode, $consumerCode, $cashBack, $subsidiaryCode);
            /*
                Para que seja entendido e o código abaixo não seja retirado até o #~#, é necessário entender(tentar) a lógica utilizada.
                
                Para que uma venda seja válida, TODO item tem que ter um valor de pelo menos 0.01 centavo.
                
                Para saber qual valor máximo de desconto(DESCONTO+CASHBACK) que o pedido pode ter, basta descobrir a quantidade de items existentes
                no pedido e dividir por 100, o resultado de (qtdItems/100) irá retornar a quantidade de items representados em centavos.
                
                SE o valor do CASHBACK + DESCONTO - quantidade de items em centavos for > que o valor de todos os produtos
                o CASHBACK é somado ao DESCONTO já existente, isso valida casos como: produtos custam 50 reais, temos 20 reais
                de desconto, 15 reais de cashback, e 7 items, então o máximo de cashback possível pra esse pedido é 
                50 - 20 - 7/100 que daria 29.93, 29.93 é o valor máximo de cashback possível
                SENÃO respeitar a regra acima, o CASHBACK a ser aplicado é recuperado da seguinte forma: VALORPRODUTOS - DESCONTO
                - quantidade de items em centavos.
                DESCONTO recebe DESCONTO + valor encontrado acima.
                e o CASHBACK quer será utilizado para salvar no campo VRDESCFID da tabela COMANDAVEN que informa quanto de valor
                cashback foi utilizado no pedido, é dado por CASHBACK que será aplicado + quantidade de itens em centavos.
                EXEMPLO: 12 ITEMS com valor total de 50 reais, desconto de 20 reais, com cashback de 60. este caso não irá respeitar
                a primeira regra, portanto: 50 - 20 - 12/100 = 29.88 ESTE é o valor máximo de CASHBACK que será aplicado,
                pode ser valido por 29.88 + 20 = 29.88 sobrando assim 0.01 centavo pra cada item.
                E o valor que será salvo na VRDESCFID é basicamente os 29.88 + 0.12 que dá 30. Este seria o valor de cashback
                utilizado caso pudéssemos zerar uma venda. 50-20-30 = 0.
            */

            $items = $this->recalculateComboDiscount($subsidiaryCode, $items);

            $realItemsQuant = $this->util->div($realItemsQuant, 100);
            if (!empty($cashBack)) {
                if ($this->util->add($cashBack, $orderDiscount) < $productsValue) {
                    $orderDiscount = $this->util->add($orderDiscount, $cashBack);
                } else {
                    $paymentOnlyWithCashback = true;
                    $cashToApply = $this->util->subtract($this->util->subtract($productsValue, $orderDiscount), $realItemsQuant);
                    $orderDiscount = $this->util->add($orderDiscount, $cashToApply);
                    $userCashBack = $cashBack;
                    $cashBack = $this->util->add($cashToApply, $realItemsQuant);


                    if ($userCashBack < $this->util->add($cashBack, $billParams['deliveryFee'])) {
                        $cashBack = $userCashBack;
                    } else {
                        $cashBack = $this->util->add($cashBack, $billParams['deliveryFee']);
                    }
                }
            }
            /* #~# */
            $remaingDiscount = $orderDiscount;
            // --- Order items
            // --- Order payments
            $payments      = $billParams['payments'];
            $giftCards     = isset($billParams['giftCards']) ? $billParams['giftCards'] : array();

            if (!empty($payments) || !empty($giftCards) || !empty($cashBack))
                $this->validatePayment($billParams['totalValue'], $payments, $giftCards, $cashBack);

            // --- Consumer description
            $consumerDescription = isset($order['consumerId']) ? $order['consumerId'] : null;

            // verifica a existência da filial
            if (!$this->util->subsidiaryExists($subsidiaryCode)) {
                throw new \Exception('Filial inexistente.');
            }

            if (!empty($cashBack) && (empty($clientCode) && empty($consumerCode))) {
                throw new \Exception('Não pode ser utilizado saldo fidelidade sem consumidor identificado.');
            }

            if (!empty($clientCode) && !empty($consumerCode)) {
                if (!$this->util->consumerExists($clientCode, $consumerCode)) {
                    throw new \Exception('Consumidor inexistente.');
                }
            } else {
                $defaultClient = $this->util->getDefaultClient($subsidiaryCode);

                if (!empty($defaultClient))
                    $clientCode = $defaultClient;
                else
                    throw new \Exception('Cliente padrão inexistente.');
            }

            $orderType     = $this->validateOrderType($orderType, $toTravel, $driveThru);
            $storeCode     = null;
            $sellerPos     = null;
            $idposobsped   = null;
            $idlugarmesa   = null;

            if (isset($order['storeCode'])) {
                $storeCode   = $order['storeCode'];
                $storeReturn = $this->util->getStoreDetails($subsidiaryCode, $storeCode);

                if (!empty($storeReturn)) {
                    if (empty($standardTable)) $standardTable = $storeReturn['NRMESA'];
                    $idposobsped = $storeReturn['IDPOSOBSPED'];
                    $idlugarmesa = $storeReturn['IDLUGARMESA'];
                }
            }

            if (isset($order['posCode'])) {
                $posCode   = $order['posCode'];
                $sellerPos = $posCode;
            }

            if (empty($sellerCode))
                $sellerReturn = $this->orderService->getStandardSeller($subsidiaryCode);
            else
                $sellerReturn = $this->orderService->getVendorInfo($sellerCode);

            if (!$sellerReturn['error']) {

                $standardSeller = $sellerReturn['data'];
                $sellerCode     = $standardSeller['CDVENDEDOR'];
                $operatorCode   = $standardSeller['CDOPERADOR'];

                if ($storeCode == null)
                    $storeCode = $standardSeller['CDLOJA'];

                if ($sellerPos == null)
                    $sellerPos = $standardSeller['CDCAIXA'];

                if ($standardTable == null)
                    $standardTable = $standardSeller['NRMESA'];

                if ($idposobsped == null)
                    $idposobsped = $standardSeller['IDPOSOBSPED'];

                if ($idlugarmesa == null)
                    $idlugarmesa = $standardSeller['IDLUGARMESA'];

                $orderExists = false;

                $table = !empty($nrmesa) ? $nrmesa : $standardTable;

                if (!empty($dscomanda))
                    $oderByDscomanda = $this->orderService->getOrderDscomanda($dscomanda, $subsidiaryCode);

                $tableData = $this->orderService->getTableData($nrmesa, $subsidiaryCode);
                $comandavenData = $this->orderService->getComandaData($comandavenSeq, $subsidiaryCode);

                if (!empty($comandavenData)) {
                    $vendarestSeq = $comandavenData['NRVENDAREST'];
                    $comandavenSeq = $comandavenData['NRCOMANDA'];
                    $orderExists = true;
                } else if (!empty($tableData)) {
                    $vendarestSeq = $tableData['NRVENDAREST'];
                    $comandavenSeq = $tableData['NRCOMANDA'];
                    $orderExists = true;
                } else if (!empty($oderByDscomanda)) {
                    $vendarestSeq = $oderByDscomanda['NRVENDAREST'];
                    $comandavenSeq = $oderByDscomanda['NRCOMANDA'];
                    $orderExists = true;
                } else {
                    $returnVendarest = $this->insertVendaRest($subsidiaryCode, $vendarestSeq, $storeCode, $sellerCode, $operatorCode, $clientCode, $consumerCode, $table);

                    if ($returnVendarest['error'])
                        $response->addMessage(new Message($returnVendarest['message'], self::ERROR_MESSAGE));

                    if (!empty($nrmesa))
                        $this->orderService->updateTableStatus($subsidiaryCode, $storeCode, $nrmesa);
                }

                if ($paymentOnlyWithCashback === true) {
                    $cdtiporece     = $this->util->getMoney();
                    $movcaixadlvSeq = str_pad(1, 10, '0', STR_PAD_LEFT);
                    $paymentValue   = $this->util->add($realItemsQuant, $billParams['deliveryFee']);

                    $this->insertMovCaixaDlv($subsidiaryCode, $vendarestSeq, $movcaixadlvSeq, 'E', $clientCode, $consumerCode, $cdtiporece, $paymentValue, null, 1);
                }

                if ($this->isToGo($orderType)) {
                    $toGo = 'S';
                    $subsidiaryAddress = $this->util->getSubsidiaryAddress($subsidiaryCode, self::PRINCIPAL_ADDRESS);

                    if (!empty($subsidiaryAddress)) {
                        $cityCode      = $subsidiaryAddress['CDMUNICIPIO'];
                        $state         = $subsidiaryAddress['SGESTADO'];
                        $country       = $subsidiaryAddress['CDPAIS'];
                    } else {
                        $cityCode      = null;
                        $state         = null;
                        $country       = null;
                    }

                    $zipCode           = null;
                    $streetAddress     = null;
                    $neighborhood      = null;
                    $number            = null;
                    $reference         = null;
                    $phone             = null;
                    $cellphone         = null;
                    $addressLine2      = null;
                    $fullStreetAddress = null;
                    $deliveryFee       = 0;
                } else {
                    $toGo           = 'N';
                    $addressDetails = $billParams['deliveryAddress'];
                    $state          = $addressDetails['state'];
                    $cityCodeExt    = $addressDetails['city'];
                    $cityCode       = $this->util->getCityCode($cityCodeExt, $state);
                    $zipCode        = $addressDetails['zipCode'];
                    $streetAddress  = $addressDetails['streetAddress'];
                    $neighborhood   = $addressDetails['neighborhood'];
                    $number         = $addressDetails['number'];
                    $reference      = $addressDetails['reference'];
                    $phone          = $addressDetails['phone'];
                    $cellphone      = $addressDetails['cellphone'];
                    $addressLine2   = $addressDetails['addressLine2'];
                    $deliveryFee    = (float)$billParams['deliveryFee'];
                    $fullStreetAddress = $streetAddress . ', ' . $number;
                    $country        = '0055';
                    $phoneToUpdate  = !empty($phone) ? $phone : $cellphone;
                    $this->util->updateConsumerPhone($clientCode, $consumerCode, $phoneToUpdate);
                }

                $idcontroprod = $this->getIdcontroprod($storeCode, $subsidiaryCode);
                $IDSTCOMANDA  = '3';

                if ($idcontroprod === 'S')
                    $IDSTCOMANDA = '1';

                if (empty($taaSaleId)) {
                    $orderToken = sha1($subsidiaryCode . $vendarestSeq . $comandavenSeq . $clientCode . $consumerCode);
                } else {
                    $orderToken = $taaSaleId;
                }

                $scheduledDate = null;
                $isScheduled   = $billParams['isScheduled'];
                $scheduledDate = $isScheduled ? $billParams['scheduledDate'] : $scheduledDate;
                $IDSTCOMANDA   = $isScheduled ? self::SCHEDULED_ORDER : $IDSTCOMANDA;
                $syncScheduled = $isScheduled ? self::PENDING_SCHED_ORDER : self::NOT_PENDING_SCHED_ORDER;
                $kdsChangeStatusOnExp = '0';

                $taaOrder = strpos($orderType, 'TAA') !== false;

                if ($this->printMode == 'EXPEDICAO') {

                    if ($toGo == 'S' || $taaOrder)
                        $kdsChangeStatusOnExp = '1';
                } else if ($this->printMode == 'INICIO') {
                    if ($taaOrder)
                        $kdsChangeStatusOnExp = '2';
                    else if ($toGo == 'S')
                        $kdsChangeStatusOnExp = '1';
                }

                if (empty($dscomanda)) {
                    if ($orderType == "MES_MOB" || $orderType == "CMD_MOB") {
                        $dscomanda = "PKR_$comandavenSeq";
                        $toGo = 'N';
                    } else {
                        $dscomanda = "DLV_$comandavenSeq";
                    }
                } else {
                    $dscomanda = str_pad($dscomanda, 10, 0, STR_PAD_LEFT);
                }

                if ($orderExists === false) {
                    $returnComandaven = $this->insertComandaVen($subsidiaryCode, $vendarestSeq, $comandavenSeq, $storeCode, $IDSTCOMANDA, $deliveryFee, $country, $state, $cityCode, $zipCode, $neighborhood, $fullStreetAddress, $addressLine2, $reference, $toGo, $orderToken, $syncScheduled, $scheduledDate, $kdsChangeStatusOnExp, $consumerDescription, $orderType, $cashBack, $orderDescription, $dscomanda, $closeByApi);
                    if ($returnComandaven['error']) {
                        $message = $returnComandaven['message'];
                        $response->addMessage(new Message($message, self::ERROR_MESSAGE));
                    }
                }

                // --- Save payments
                if (!empty($payments)) {
                    $savePaymentsreturn = $this->savePayments($payments, $subsidiaryCode, $vendarestSeq, $clientCode, $consumerCode, (float)$billParams['changeValue'], true);
                    if ($savePaymentsreturn['error']) {
                        $message = $savePaymentsreturn['message'];
                        throw new \Exception($message);
                    }
                }

                // --- Save gift cards
                if (!empty($giftCards)) {
                    $DTABERCAIX        = $this->util->getPosDate($subsidiaryCode, $sellerPos);
                    $totalValue        = $billParams['totalValue'] - $cashBack;
                    $giftCardsResponse = $this->saveGiftCards($giftCards, $subsidiaryCode, $vendarestSeq, $sellerPos, $DTABERCAIX, $totalValue);

                    if ($giftCardsResponse['error']) {
                        $message = $giftCardsResponse['message'];
                        throw new \Exception($message);
                    }
                }

                // --- save items
                if (count($items) > 0)
                    $itemsSeq = $this->orderService->generateItemsSeq($subsidiaryCode);

                $orderCPF = $billParams['orderCPF'];
                $itemNum  = 0;
                $positionNumber = '01';

                $toTravelData = array();
                $driveThruData = array();

                // if(!empty($toTravel)) $toTravelData = $this->getToTravelData();
                if (!empty($driveThru)) $driveThruData = $this->getDriveThruData();

                if (!$taaOrder && !empty($pagerId)) {
                    $pagerId = substr($pagerId, 0, 4);
                    $pagerId = 'A' . $pagerId;
                }

                $sumValue = 0;
                $totalDiscount = $orderDiscount;

                while ($orderDiscount > 0) {
                    $sumValue = 0;
                    foreach ($items as &$item) {
                        $this->shareDiscount($item, $orderDiscount, $productsValue, $remaingValue, $remaingDiscount, false);
                    }
                    unset($item);
                    $remaingValue = $productsValue;
                    foreach ($items as $item) {
                        $sumValue += $this->getDiscountValue($item);
                    }
                    $orderDiscount = $this->util->subtract($totalDiscount, $sumValue);
                    $remaingDiscount = $orderDiscount;
                    unset($item);
                }
                $quantItems = $this->orderService->getQuantItens($subsidiaryCode, $vendarestSeq, $comandavenSeq);
                $quantCombos = 1;

                foreach ($items as &$item) {
                    $code = $item['productCode'];
                    if (empty($item['subProductsGroupPrint']) && empty($item['productPrint'])) throw new \Exception("Não foram enviadas informações de impressão para o item de código $code .");
                    
                    $this->saveItemsEx($item, $subsidiaryCode, $vendarestSeq, $comandavenSeq, $itemsSeq, $positionNumber, $idcontroprod, $sellerPos, $storeCode, $sellerCode, $orderCPF, $itemNum, null, $toTravelData, $pagerId, $holdTime, $driveThruData, $orderType, false, $quantItems, $quantCombos);
                }

                unset($item);


                if (empty($consumerCode) && !empty($giftCards)) {
                    $this->orderService->updateConsumerVendarest($giftCards[0]['clientCode'], $giftCards[0]['consumerCode'], $vendarestSeq, $subsidiaryCode);
                }

                $this->util->commit();

                if (!empty($consumerDescription) && !empty($pagerId))
                    $pagerId .= " - $consumerDescription";

                $printItems = $this->formatItemsToPrint($items);

                $toGoString = ($toGo == 'S') ? 'PEDIDO PARA RETIRADA NA LOJA' : 'PEDIDO PARA ENTREGA DELIVERY';
                $this->printOrder($subsidiaryCode, $comandavenSeq, $storeCode, $printItems, $sellerCode, $table, $pagerId, 'B', $toGoString, $taaOrder);
                $response->addDataSet(
                    new DataSet(
                        '',
                        array(
                            array(
                                'saleId' => $orderToken,
                                'saleCode' => $comandavenSeq,
                                'pagerId' => $pagerId
                            )
                        )
                    )
                );
            } else {
                $this->util->rollBack();
                $message = $sellerReturn['message'];
                $response->addMessage(new Message($message, self::ERROR_MESSAGE));
            }
        } catch (\Exception $e) {
            $this->util->rollBack();

            $message = 'Erro ao inserir comanda delivery. ' . $e->getMessage();
            $this->util->saveLogOrder($message, 'Error/' . date("Y/m/d"), $fileName);

            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
        }
    }

    public function V3_saveDeliveryOrderEx(Request $request, Response $response)
    {
        $inputJSON = file_get_contents('php://input');
        $input     = json_decode($inputJSON, true);

        try {
            if (!isset($input))
                throw new \Exception('Não foram recebidos parâmetros');
        } catch (\Exception $e) {
            $message = 'Erro ao inserir comanda delivery. ' . $e->getMessage();
            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
            return;
        }

        $data   = array();
        $params = $input;

        try {

            $taaSaleId = isset($input['taaSaleId']) ? $input['taaSaleId'] : null;

            if (!empty($taaSaleId)) {
                $orderById = $this->orderService->getOrderById($taaSaleId);

                if (!empty($orderById)) {

                    if ($this->util->databaseIsOracle())
                        $format = 'Y-m-d H:i:s';
                    else
                        $format = 'Y-m-d H:i:s.u';

                    $orderTime = \DateTime::createFromFormat($format, $orderById['DTHRABERMESA']);
                    $orderTime = $orderTime->format($format);

                    throw new \Exception("Pedido $taaSaleId já existente no banco de dados. Pedido feito em $orderTime");
                }
            }
        } catch (\Exception $e) {
            $response->addMessage(new Message($e->getMessage(), self::GENERAL_MESSAGE));
            $response->addDataSet(
                new DataSet(
                    '',
                    array(
                        array(
                            'saleId'   => $orderById['NRCOMANDAEXT'],
                            'saleCode' => $orderById['NRCOMANDA']
                        )
                    )
                )
            );
            return;
        }

        $order    = $params['order'];
        $fileName = date("Ymd-H-i-s") . "-";
        $pagerId  = (isset($order['pagerId']) && !empty($order['pagerId'])) ? $order['pagerId'] : (string)$this->geraSenhaProducao($order['storeCode'], $order['subsidiaryCode'], $order['posCode']);
        $fileName .= empty($pagerId) ? $order['subsidiaryCode'] : $pagerId;

        $this->util->saveLogOrder($inputJSON, 'Order/' . date("Y/m/d"), $fileName);

        $this->util->beginTransaction();

        try {
            $subsidiaryCode   = $order['subsidiaryCode'];
            $comandavenSeq    = isset($order['saleCode']) ? $order['saleCode'] : null;
            $dscomanda        = isset($order['orderId']) ? $order['orderId'] : null;
            $nrmesa           = isset($order['tableNumber']) ? $order['tableNumber'] : null;

            $vendarestSeq = $this->orderService->generateVendarestSeq($subsidiaryCode);
            $comandavenSeq = $this->orderService->generateComandavenSeq($subsidiaryCode);

            $this->util->commit();
        } catch (\Exception $e) {
            $this->util->rollBack();
            $message = 'Erro ao gerar novo codigo. ' . $e->getMessage();
            $this->util->saveLogOrder($message, 'Error/' . date("Y/m/d"), $fileName);
            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
            return;
        }

        $now = strtotime("now");

        $this->util->beginTransaction();

        try {
            $consumerCode     = !empty($order['consumerCode']) ? $order['consumerCode'] : null;
            $clientCode       = $order['clientCode'];
            $billParams       = $order['orderDescription'];
            $orderType        = $billParams['orderType'];
            $orderDiscount    = isset($billParams['orderDiscount']) ? $billParams['orderDiscount'] : 0;
            $holdTime         = isset($billParams['holdTime']) ? $billParams['holdTime'] : 0;
            $productsValue    = $billParams['productsValue'];
            $remaingValue     = $billParams['productsValue'];
            $toTravel         = !empty($billParams['toTravel']);
            $driveThru        = !empty($billParams['driveThru']);
            $orderDescription = isset($billParams['description']) ? $billParams['description'] : null;
            $sellerCode       = isset($order['sellerCode']) ? $order['sellerCode'] : null;
            $closeByApi       = isset($order['closeByApi']) ? $order['closeByApi'] : false;
            $items            = $order['items'];

            $closeByApi = !empty($closeByApi) ? 'S' : 'N';

            $standardTable = null;
            if (!empty($nrmesa)) {
                $this->orderService->validateTable($nrmesa);
            }

            $payments = $billParams['payments'];
            // ---> validate payments <---

            // --- Consumer description
            $consumerDescription = isset($order['consumerId']) ? $order['consumerId'] : null;

            // verifica a existência da filial
            if (!$this->util->subsidiaryExists($subsidiaryCode)) {
                throw new \Exception('Filial inexistente.');
            }

            if (!empty($clientCode) && !empty($consumerCode)) {
                if (!$this->util->consumerExists($clientCode, $consumerCode)) {
                    throw new \Exception('Consumidor inexistente.');
                }
            } else {
                $defaultClient = $this->util->getDefaultClient($subsidiaryCode);

                if (!empty($defaultClient))
                    $clientCode = $defaultClient;
                else
                    throw new \Exception('Cliente padrão inexistente.');
            }

            $orderType     = $this->validateOrderType($orderType, $toTravel, $driveThru);
            $storeCode     = null;
            $sellerPos     = null;
            $idposobsped   = null;
            $idlugarmesa   = null;

            if (isset($order['storeCode'])) {
                $storeCode   = $order['storeCode'];
                $storeReturn = $this->util->getStoreDetails($subsidiaryCode, $storeCode);

                if (!empty($storeReturn)) {
                    if (empty($standardTable)) $standardTable = $storeReturn['NRMESA'];
                    $idposobsped = $storeReturn['IDPOSOBSPED'];
                    $idlugarmesa = $storeReturn['IDLUGARMESA'];
                }
            }

            if (isset($order['posCode'])) {
                $posCode   = $order['posCode'];
                $sellerPos = $posCode;
            }

            if (empty($sellerCode)){
                $sellerReturn = $this->orderService->getStandardSeller($subsidiaryCode);

                if($sellerReturn['error']){
                    throw new \Exception($sellerReturn['message']);
                }
            }else{
                $sellerReturn = $this->orderService->getVendorInfo($sellerCode);
            }

            if (!$sellerReturn['error']) {

                $standardSeller = $sellerReturn['data'];
                $sellerCode     = $standardSeller['CDVENDEDOR'];
                $operatorCode   = $standardSeller['CDOPERADOR'];

                if ($storeCode == null)
                    $storeCode = $standardSeller['CDLOJA'];

                if ($sellerPos == null)
                    $sellerPos = $standardSeller['CDCAIXA'];

                if ($standardTable == null)
                    $standardTable = $standardSeller['NRMESA'];

                if ($idposobsped == null)
                    $idposobsped = $standardSeller['IDPOSOBSPED'];

                if ($idlugarmesa == null)
                    $idlugarmesa = $standardSeller['IDLUGARMESA'];

                $table = !empty($nrmesa) ? $nrmesa : $standardTable;

                if (!empty($dscomanda))
                    $oderByDscomanda = $this->orderService->getOrderDscomanda($dscomanda, $subsidiaryCode);

                $returnVendarest = $this->insertVendaRest($subsidiaryCode, $vendarestSeq, $storeCode, $sellerCode, $operatorCode, $clientCode, $consumerCode, $table);

                if ($returnVendarest['error'])
                    $response->addMessage(new Message($returnVendarest['message'], self::ERROR_MESSAGE));

                if (!empty($nrmesa))
                    $this->orderService->updateTableStatus($subsidiaryCode, $storeCode, $nrmesa);

                if ($this->isToGo($orderType)) {
                    $toGo = 'S';
                    $subsidiaryAddress = $this->util->getSubsidiaryAddress($subsidiaryCode, self::PRINCIPAL_ADDRESS);

                    if (!empty($subsidiaryAddress)) {
                        $cityCode      = $subsidiaryAddress['CDMUNICIPIO'];
                        $state         = $subsidiaryAddress['SGESTADO'];
                        $country       = $subsidiaryAddress['CDPAIS'];
                    } else {
                        $cityCode      = null;
                        $state         = null;
                        $country       = null;
                    }

                    $zipCode           = null;
                    $streetAddress     = null;
                    $neighborhood      = null;
                    $number            = null;
                    $reference         = null;
                    $phone             = null;
                    $cellphone         = null;
                    $addressLine2      = null;
                    $fullStreetAddress = null;
                    $deliveryFee       = 0;
                } else {
                    $toGo           = 'N';
                    $addressDetails = $billParams['deliveryAddress'];
                    $state          = $addressDetails['state'];
                    $cityCodeExt    = $addressDetails['city'];
                    $cityCode       = $this->util->getCityCode($cityCodeExt, $state);
                    $zipCode        = $addressDetails['zipCode'];
                    $streetAddress  = $addressDetails['streetAddress'];
                    $neighborhood   = $addressDetails['neighborhood'];
                    $number         = $addressDetails['number'];
                    $reference      = $addressDetails['reference'];
                    $phone          = $addressDetails['phone'];
                    $cellphone      = $addressDetails['cellphone'];
                    $addressLine2   = $addressDetails['addressLine2'];
                    $deliveryFee    = (float)$billParams['deliveryFee'];
                    $fullStreetAddress = $streetAddress . ', ' . $number;
                    $country        = '0055';
                    $phoneToUpdate  = !empty($phone) ? $phone : $cellphone;
                    $this->util->updateConsumerPhone($clientCode, $consumerCode, $phoneToUpdate);
                }

                $idcontroprod = $this->getIdcontroprod($storeCode, $subsidiaryCode);
                $IDSTCOMANDA  = '3';

                if ($idcontroprod === 'S')
                    $IDSTCOMANDA = '1';

                if (empty($taaSaleId)) {
                    $orderToken = sha1($subsidiaryCode . $vendarestSeq . $comandavenSeq . $clientCode . $consumerCode);
                } else {
                    $orderToken = $taaSaleId;
                }

                $scheduledDate = null;
                $isScheduled   = $billParams['isScheduled'];
                $scheduledDate = $isScheduled ? $billParams['scheduledDate'] : $scheduledDate;
                $IDSTCOMANDA   = $isScheduled ? self::SCHEDULED_ORDER : $IDSTCOMANDA;
                $syncScheduled = $isScheduled ? self::PENDING_SCHED_ORDER : self::NOT_PENDING_SCHED_ORDER;
                $kdsChangeStatusOnExp = '0';

                $taaOrder = strpos($orderType, 'TAA') !== false;

                if ($this->printMode == 'EXPEDICAO') {

                    if ($toGo == 'S' || $taaOrder)
                        $kdsChangeStatusOnExp = '1';
                } else if ($this->printMode == 'INICIO') {
                    if ($taaOrder)
                        $kdsChangeStatusOnExp = '2';
                    else if ($toGo == 'S')
                        $kdsChangeStatusOnExp = '1';
                }

                if (empty($dscomanda)) {
                    if ($orderType == "MES_MOB" || $orderType == "CMD_MOB") {
                        $dscomanda = "PKR_$comandavenSeq";
                        $toGo = 'N';
                    } else {
                        $dscomanda = "DLV_$comandavenSeq";
                    }
                } else {
                    $dscomanda = str_pad($dscomanda, 10, 0, STR_PAD_LEFT);
                }

                $orderCPF     = $billParams['orderCPF'];
                $consumerName = isset($billParams['consumerName']) ? $billParams['consumerName'] : null;
                if(!empty($orderCPF) && !$this->util->isValidCPF($orderCPF)) {
                  throw new \Exception("O CPF informado no pedido é inválido.");
                }

                $returnComandaven = $this->insertComandaVenEx($subsidiaryCode, $vendarestSeq, $comandavenSeq, $storeCode, $IDSTCOMANDA, $deliveryFee, $country, $state, $cityCode, $zipCode, $neighborhood, $fullStreetAddress, $addressLine2, $reference, $toGo, $orderToken, $syncScheduled, $scheduledDate, $kdsChangeStatusOnExp, $consumerDescription, $orderType, $orderDescription, $dscomanda, $closeByApi, 0, $orderCPF);
                if ($returnComandaven['error']) {
                    $message = $returnComandaven['message'];
                    $response->addMessage(new Message($message, self::ERROR_MESSAGE));
                }

                // --- Save payments
                if (!empty($payments)) {
                    // ---> verify function savePayments cause cashBack was removed <---
                    $savePaymentsreturn = $this->savePayments($payments, $subsidiaryCode, $vendarestSeq, $clientCode, $consumerCode, (float)$billParams['changeValue'], true);
                    if ($savePaymentsreturn['error']) {
                        $message = $savePaymentsreturn['message'];
                        throw new \Exception($message);
                    }
                }

                // ---> new saveItems <---
                $itemsArray = $this->orderService->formatItems($items, $orderType, $subsidiaryCode, $vendarestSeq, $comandavenSeq, $storeCode, $idcontroprod, $sellerPos, $sellerCode, $orderCPF, $consumerName, $pagerId, $holdTime);
                
                foreach($itemsArray['ITCOMANDAVEN'] as $validateItems) {
                    if($validateItems['VRPRECCOMVEN'] < 0.01)
                      throw new \Exception("Não é permitido gravar produtos com preço zero. Produto {$validateItems['NMPRODUTO']}|código {$validateItems['CDPRODUTO']}.");
                }

                if($orderDiscount > 0){
                    $this->orderService->shareDiscount($orderDiscount, $itemsArray['ITCOMANDAVEN']);
                }
                $this->orderService->saveOrderItems($itemsArray);

                $this->util->commit();

                if (!empty($consumerDescription) && !empty($pagerId))
                    $pagerId .= " - $consumerDescription";

                $printItems = $this->formatItemsToPrint($items);
                

                $toGoString = ($toGo == 'S') ? 'PEDIDO PARA RETIRADA NA LOJA' : 'PEDIDO PARA ENTREGA DELIVERY';
                $this->printOrder($subsidiaryCode, $comandavenSeq, $storeCode, $printItems, $sellerCode, $table, $pagerId, 'B', $toGoString, $taaOrder);
                $response->addDataSet(
                    new DataSet(
                        '',
                        array(
                            array(
                                'saleId' => $orderToken,
                                'saleCode' => $comandavenSeq,
                                'pagerId' => $pagerId
                            )
                        )
                    )
                );
            }
        } catch (\Exception $e) {
            $this->util->rollBack();

            $message = 'Erro ao inserir comanda delivery. ' . $e->getMessage();
            $this->util->saveLogOrder($message, 'Error/' . date("Y/m/d"), $fileName);

            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
        }
    }

    private function validateOrderType($orderType, $toTravel, $driveThru)
    {
        if ($orderType == 'DLV') {
            $orderType = 'DLV_APP';
        } else if ($orderType == 'TAA') {
            if ($toTravel) {
                $orderType = 'TGO_TAA';
            } else if ($driveThru) {
                $orderType = 'THR_TAA';
            } else {
                $orderType = 'BAL_TAA';
            }
        } else if ($orderType == 'MES') {
            $orderType = 'MES_MOB';
        } else if ($orderType == 'CMD') {
            $orderType = 'CMD_MOB';
        }

        if ($orderType === 'BGR_APP')
            $orderType = 'BAL_APP';

        if ($orderType === 'BGR_TAA')
            $orderType = 'BAL_TAA';

        return $orderType;
    }

    public function formatItemsToPrint($items)
    {
        $formatedItems = array();

        foreach ($items as $item) {
            $formatedItem = array();
            $childItems   = array();

            if (!empty($item['childItems'])) {

                $formatedItem['NMPRODPROMOCAO'] = $item['productName'];
                $formatedItem['IDTIPOCOMPPROD'] = '3';
                $formatedItem['ORDEMIMP']       = 0;

                $childItems = $this->formatItemsToPrint($item['childItems']);
            } else {
                $formatedItem['IDIMPPROMOCAO']  = 'N';
                $formatedItem['IDTIPOCOMPPROD'] = null;
            }

            $formatedItem['CDPRODUTO']      = $item['productCode'];
            $formatedItem['NRLUGARMESA']    = 1;
            $formatedItem['CDPRODUTO']      = $item['productCode'];
            $formatedItem['NMPRODUTO']      = $item['productName'];
            $formatedItem['QTPRODCOMVEN']   = $item['quantity'];
            $formatedItem['ATRASOPROD']     = null;
            $formatedItem['IDORIGEMVENDA']  = $item['toTravel'] == true ? 'TGO': '';
            $TXPRODCOMVEN = '';
            $note = !empty($item['note']) ? $item['note'] : '';

            if (!empty($item['observations'])) {
                foreach ($item['observations'] as $obs) {
                    $TXPRODCOMVEN .= " {$obs['obsName']}\n   *  ";
                }
            }

            $formatedItem['TXPRODCOMVEN'] = substr($TXPRODCOMVEN, 0, -4).$note;

            array_push($formatedItems, $formatedItem);

            if (!empty($childItems))
                $formatedItems = array_merge($formatedItems, $childItems);
        }

        return $formatedItems;
    }

    public function isToGo($orderType)
    {
        if (strpos($orderType, 'TGO') !== false) return true;
        if (strpos($orderType, 'BGR') !== false) return true;
        if (strpos($orderType, 'MES') !== false) return true;
        if (strpos($orderType, 'BAL') !== false) return true;
        if (strpos($orderType, 'CMD') !== false) return true;
        if (strpos($orderType, 'THR') !== false) return true;
        return false;
    }

    private function getToTravelData()
    {
        $toTravelObservation = $this->util->getToTravelObservation();
        return array(
            'obsGroupCode'  => $toTravelObservation['CDGRPOCOR'],
            'obsCode'       => $toTravelObservation['CDOCORR'],
            'obsName'       => $toTravelObservation['DSOCORR'],
            'price'         => 0,
            'quantity'      => 1
        );
    }

    public function getDriveThruData()
    {
        $driveThruObservation = $this->util->getDriveThruObservation();
        return array(
            'obsGroupCode'  => $driveThruObservation['CDGRPOCOR'],
            'obsCode'       => $driveThruObservation['CDOCORR'],
            'obsName'       => $driveThruObservation['DSOCORR'],
            'price'         => 0,
            'quantity'      => 1
        );
    }

    public function saveGiftCards($giftCards, $subsidiaryCode, $vendarestSeq, $sellerPos, $DTABERCAIX, $totalValue)
    {
        $savedValue = $this->util->getPaymentValue($subsidiaryCode, $vendarestSeq);

        $this->validateConsumerType($giftCards);

        $paymentValue = bcsub($totalValue, $savedValue, 2);
        $paymentCode  = $this->util->getCreditPayment();

        if (count($giftCards) == 1) {
            $giftCard     = $giftCards[0];
            $clientCode   = $giftCard['clientCode'];
            $consumerCode = $giftCard['consumerCode'];

            if ($giftCard['value'] < $paymentValue)
                throw new \Exception("Valor insuficiente para realização do pagamento com cartão presente.");
        } else {

            $destinyCard    = array_shift($giftCards);
            $clientCode     = $destinyCard['clientCode'];
            $consumerCode   = $destinyCard['consumerCode'];
            $availableValue = $destinyCard['value'];

            foreach ($giftCards as $giftCard)
                $availableValue = bcadd($availableValue, $giftCard['value'], 2);

            if ($availableValue < $paymentValue)
                throw new \Exception("Valor insuficiente para realização do pagamento com cartão presente.");

            $extratoCons = new \SimpleXMLElement('<ExtratoCons/>');
            $xmlFields = $this->getXMLFields();

            foreach ($giftCards as $giftCard) {
                $registro = $extratoCons->addChild('Registro');

                foreach ($xmlFields as $field)
                    $registro->addChild($field);

                $registro->cdcliente    = $giftCard['clientCode'];
                $registro->cdconsumidor = $giftCard['consumerCode'];
                $registro->cdfamilisald = $giftCard['cdfamilisald'];
                $registro->dtmovextcons = date('d/m/Y H:i:s');
                $registro->nrseqmovext  = '001';
                $registro->cdtiporece   = '001';
                $registro->cdfilial     = $subsidiaryCode;
                $registro->cdcaixa      = $sellerPos;
                $registro->dsopeextcons = "Transf. Saldo:" . $destinyCard['cardNumber'] . " - " . $destinyCard['clientCode'];
                $registro->vrmovextcons = $giftCard['value'];
                $registro->vrsaldconext = 0;
                $registro->idtpmovext   = 'D';
                $registro->idimpextrato = 'S';
                $detiny = $extratoCons->addChild('Registro');

                foreach ($xmlFields as $field)
                    $detiny->addChild($field);

                $detiny->cdcliente    = $destinyCard['clientCode'];
                $detiny->cdconsumidor = $destinyCard['consumerCode'];
                $detiny->cdfamilisald = $destinyCard['cdfamilisald'];
                $detiny->dtmovextcons = date('d/m/Y H:i:s');
                $detiny->nrseqmovext  = '001';
                $detiny->cdtiporece   = '001';
                $detiny->cdfilial     = $subsidiaryCode;
                $detiny->cdcaixa      = $sellerPos;
                $detiny->dsopeextcons = "Transf. Saldo:" . $giftCard['cardNumber'] . " - " . $giftCard['clientCode'];
                $detiny->vrmovextcons = $giftCard['value'];
                $detiny->vrsaldconext = 0;
                $detiny->idtpmovext   = 'C';
                $detiny->idimpextrato = 'S';
            }

            $xmlToSend = $extratoCons->asXML();
            $params = array(
                'P_XMLEXTCONS' => $xmlToSend
            );

            $this->util->callExtratocons($params, $subsidiaryCode, "InsereDados");
        }

        $movcaixadlvSeq    = $this->orderService->generateMovcaixadlvSeq($subsidiaryCode, $vendarestSeq);
        $movementType      = 'E';
        $returnMovcaixadlv = $this->insertMovCaixaDlv($subsidiaryCode, $vendarestSeq, $movcaixadlvSeq, $movementType, $clientCode, $consumerCode, $paymentCode, $paymentValue, null, 1, null);

        if ($returnMovcaixadlv['error'])
            return $this->returnPattern(true, $returnMovcaixadlv['message']);

        return $this->returnPattern(false, '');
    }

    public function validateConsumerType($giftCards)
    {
        $cards = $giftCards;

        foreach ($giftCards as &$giftCard)
            $giftCard['TIPOCONS'] = $this->util->getTipoCons($giftCard['clientCode'], $giftCard['consumerCode']);

        unset($giftCard);

        $consType = $giftCards[0]['TIPOCONS'];

        foreach ($giftCards as $giftCard)
            if (!in_array($consType, $giftCard)) throw new \Exception("Os tipos de consumidor devem ser iguais.");

        unset($giftCard);
    }

    public function getXMLFields()
    {
        return array("cdcliente", "cdconsumidor", "cdfamilisald", "dtmovextcons", "nrseqmovext", "cdtiporece", "cdfilial", "cdcaixa", "dsopeextcons", "vrmovextcons", "vrsaldconext", "idtpmovext", "nrdeposicons", "nrseqmovcaixa", "idimpextrato");
    }

    public function savePayments($payments, $subsidiaryCode, $vendarestSeq, $clientCode, $consumerCode, $change, $validateConsumer = false)
    {
        $sequential = 1;

        foreach ($payments as $key => $payment) {
            $paymentCode   = $payment['paymentCode'];
            $quantParcel   = isset($payment['quantParcel']) ? $payment['quantParcel'] : 1;
            $this->util->validatePaymentCode($paymentCode);
            $nsu           = isset($payment['nsu']) ? $payment['nsu'] : null;
            $authHash      = isset($payment['authHash']) ? $payment['authHash'] : null;
            $paymentValue  = $this->util->removeDecimal($payment['value']);
            $paymentId     = $this->util->getPaymentId($paymentCode);
            
            if ($paymentId === "9") {
                
                if (empty($consumerCode) || empty($clientCode))
                    throw new \Exception("Obrigatorio a identificacao do consumidor na abertura do pedido para o pagamento codigo: $paymentCode");
                
                if($validateConsumer){
                    $cdcliente = !empty($payment['clientCode']) ? $payment['clientCode'] : null;
                    $cdconsumidor = !empty($payment['consumerCode']) ? $payment['consumerCode'] : null;

                    if(empty($cdcliente)) throw new \Exception("Código de cliente não informado para o pagamento código: $paymentCode.");
                    if(empty($cdconsumidor)) throw new \Exception("Código de consumidor não informado para o pagamento código: $paymentCode.");

                    if($consumerCode != $cdconsumidor || $clientCode != $cdcliente)
                        throw new \Exception("Consumidor informado no pagamento código $paymentCode difere do consumidor informado na abertura do pedido.");
                
                }
                
                $returnVerify = $this->verifyDailyCreditLimit($paymentValue, $clientCode, $consumerCode, $subsidiaryCode);
                if ($returnVerify['error']) {
                    return $this->returnPattern(true, $returnVerify['message']);
                }
            }

            $movcaixadlvSeq    = $this->orderService->generateMovcaixadlvSeq($subsidiaryCode, $vendarestSeq);
            $movementType      = 'E';
            $returnMovcaixadlv = $this->insertMovCaixaDlv($subsidiaryCode, $vendarestSeq, $movcaixadlvSeq, $movementType, $clientCode, $consumerCode, $paymentCode, $paymentValue, $nsu, $quantParcel, $authHash);

            if ($returnMovcaixadlv['error'])
                return $this->returnPattern(true, $returnMovcaixadlv['message']);
        }

        // --- save change
        if ($change > 0) {
            $paymentCode    = $this->util->getMoney();
            $movcaixadlvSeq = $this->orderService->generateMovcaixadlvSeq($subsidiaryCode, $vendarestSeq);
            $movementType   = 'S';

            $returnMovcaixadlv = $this->insertMovCaixaDlv($subsidiaryCode, $vendarestSeq, $movcaixadlvSeq, $movementType, $clientCode, $consumerCode, $paymentCode, $change, $nsu, 1, null);

            if ($returnMovcaixadlv['error'])
                return $this->returnPattern(true, $returnMovcaixadlv['message']);
        }

        return $this->returnPattern(false, '');
    }

    public function divideItems(&$items, &$item)
    {

        if (!empty($item['childItems'])) {
            $childQuant = 0;

            foreach ($item['childItems'] as &$childItem) {
                $childQuant += $this->divideItems($item['childItems'], $childItem);
            }

            if (!empty($item['observations'])) {
                foreach ($item['observations'] as $observation) {
                    if (!empty($item['price'])) {
                        $childQuant += 1;
                    }
                }
            }

            $item['NRLUGARMESA'] = 1;

            unset($childItem);

            if ($item['quantity'] > 1) {
                $fatherToPush = $item['quantity'];
                $item['quantity'] = 1;

                for ($i = 1; $i < $fatherToPush; $i++) {
                    array_push($items, $item);
                }
            }

            return $childQuant;
        } else if ($item['quantity']) {

            $realQuantity           = 1;
            $qtToSave               = $item['quantity'];
            $item['quantity']       = 1;
            $item['NRLUGARMESA']    = 1;
            $item['IDIMPPROMOCAO']  = 'N';

            for ($i = 1; $i < $qtToSave; $i++) {
                array_push($items, $item);
            }

            if (!empty($item['observations'])) {
                foreach ($item['observations'] as $observation) {
                    if (!empty($observation['price'])) {
                        $realQuantity += 1;
                    }
                }
            }

            return $realQuantity;
        }
    }

    public function shareDiscountEx(&$items, $orderDiscount, $productsValue)
    {   

        $idxToApplyDiscount  = [];
        $totalDiscount       = $orderDiscount;
        $discountLeft        = $orderDiscount;
        $productsTotalValue  = "0.00";
        $discountGivenShare  = "0.00";
        $maxValueItem        = "0.00";
        $maxValueIdx         = -1;

        if($orderDiscount >= $productsValue) {
            throw new \Exception('Valor de total de desconto igual ou superior ao valor total de produtos');
        }

        foreach($items as &$product) {
            $product['VRDESCCOMVEN']  = 0;
            $product['discount'] = "0.00";
        }
        
        if(isset($product))
            unset($product);

        foreach($items as $i => $product) {
            if((bccomp($product["VRPRECCOMVEN"], "0.01", 2) == 1)) { 
                $unityTotalValue = bcmul($product['VRPRECCOMVEN'], $product['QTPRODCOMVEN'], 2);
                $productsTotalValue = bcadd($productsTotalValue, $unityTotalValue, 2);
                array_push($idxToApplyDiscount, $i);
            }
        }

        foreach($idxToApplyDiscount as $idxWithDiscount) {
            $items[$idxWithDiscount]['TOTAL'] = bcmul($items[$idxWithDiscount]['VRPRECCOMVEN'], $items[$idxWithDiscount]['QTPRODCOMVEN'], 2);
            $discountShare = bcdiv((bcmul($items[$idxWithDiscount]['TOTAL'], "100.00", 2)), $productsTotalValue, 2);
            $items[$idxWithDiscount]["discount"] = $discountShare;

            $discountGivenShare = bcadd($discountGivenShare, $discountShare, 2);
        }
        

        foreach($items as $i => $product) {
            if(bccomp($product['discount'], $maxValueItem, 2) == 1) {
                $maxValueIdx  = $i;
                $maxValueItem = $product['discount'];
            }            
        }

        if($maxValueIdx != -1) {
            $items[$maxValueIdx]["discount"] = bcadd($items[$maxValueIdx]["discount"], bcsub("100.00", $discountGivenShare, 2), 2);
        }

        foreach($idxToApplyDiscount as $idxWithDiscount) {
        
            $itemDiscount = bcdiv((bcmul($items[$idxWithDiscount]["discount"], $totalDiscount, 2)), "100.00", 2);
            $items[$idxWithDiscount]["VRDESCCOMVEN"] = $itemDiscount;

            $discountLeft = bcsub($discountLeft, $itemDiscount, 2);
        }

        if($maxValueIdx != -1) {
            $items[$maxValueIdx]["VRDESCCOMVEN"] = bcadd($items[$maxValueIdx]["VRDESCCOMVEN"], $discountLeft, 2);
            $discountLeft = "0.00";
        }
        
        $leftDiscount = "0.00";
        
        foreach($items as &$product) {
            $unityTotalValue = bcmul($product['VRPRECCOMVEN'], $product['QTPRODCOMVEN'], 2);
            if( (bccomp($product['VRDESCCOMVEN'], $unityTotalValue, 2) == 1) || (bccomp($product['VRDESCCOMVEN'], $unityTotalValue, 2) == 0) )  {
                $leftDiscount = bcadd($leftDiscount, bcsub($product['VRDESCCOMVEN'], $product['VRPRECCOMVEN'], 2), 2);
                $leftDiscount = bcadd($leftDiscount, "0.01", 2);
                
                $product['VRDESCCOMVEN'] = bcsub($product['VRDESCCOMVEN'], $leftDiscount, 2);
            }
        }
        
        if(isset($product))
            unset($product);

        foreach($items as &$product){
            $unityTotalValue = bcmul($product['VRPRECCOMVEN'], $product['QTPRODCOMVEN'], 2);
            $maxAddDisc = (bcsub(bcsub($unityTotalValue, "0.01", 2), $product['VRDESCCOMVEN'], 2));
            if($maxAddDisc >= $leftDiscount) {
                $addDisc = $leftDiscount;
            } else{
                $addDisc = $maxAddDisc;
            }

            $product['VRDESCCOMVEN'] = bcadd($product['VRDESCCOMVEN'], $addDisc, 2);
            $product['discount'] = $product['VRDESCCOMVEN'];
            $leftDiscount = bcsub($leftDiscount, $addDisc, 2);
        }        
    }

    public function shareDiscount(&$item, $orderDiscount, $productsValue, &$remaingValue, &$remaingDiscount, $isObs)
    {
        if (!empty($item['observations'])) {
            foreach ($item['observations'] as &$observation) {
                $this->shareDiscount($observation, $orderDiscount, $productsValue, $remaingValue, $remaingDiscount, true);
            }
            unset($observation);
        }
        if (!empty($item['childItems'])) {
            foreach ($item['childItems'] as &$childItem) {
                $this->shareDiscount($childItem, $orderDiscount, $productsValue, $remaingValue, $remaingDiscount, false);
            }
            unset($childItem);
        } else {
            $productQuantity    = $item['quantity'];
            if ($isObs) {
                $productValue = !empty($item['price']) ? $item['price'] : 0;
                $item['discount'] = !empty($item['discount']) ? $item['discount'] : 0;
                $productDiscount = $item['discount'];
            } else {
                $productValue    = $item['unitValue'];
                $productDiscount = $item['discount'];
            }
            if ($productValue != 0) {
                /*echo  ("\n-----------------------------------------------------");
                echo  ("\nProduct Value ".$productValue);  
                  echo  ("\nProduct Discount ".$productDiscount);
                    echo  ("\nRemaing Value ".$remaingValue);   
                 echo  ("\nRemaing Discount ".$remaingDiscount); 
                 if($isObs){
                     echo  ("\n============================\nobsName: {$item['obsName']}\nPrice: {$item['price']}\nQuantity: {$item['quantity']}\nDiscount: {$item['discount']}");
                 }else{
                    echo  ("\n============================\nproductName: {$item['productName']}\nUnitValue: {$item['unitValue']}\nQuantity: {$item['quantity']}\nDiscount: {$item['discount']}");
                 }*/
                if (!isset($item['elegible']) || !empty($item['elegible'])) {
                    if ($this->util->removeDecimal($remaingValue) == $this->util->removeDecimal($productValue)) {
                        if ($this->util->removeDecimal($productValue) == $this->util->removeDecimal($remaingDiscount)) {
                            $productDiscount = $remaingDiscount - 0.01;
                            $remaingDiscount = 0.01;
                        } else $productDiscount = $remaingDiscount;
                    } else if ($remaingDiscount > 0 && $remaingDiscount < 0.01) {
                        $remaingDiscount = 0.01;
                        $productDiscount = 0.01;
                    } else {
                        $productDiscount = $this->util->removeDecimal(((($this->util->subtract($productValue, $productDiscount)) * $productQuantity) * $orderDiscount) / $productsValue);
                    }
                    $actualValue = $this->util->removeDecimal($productValue - $item['discount']);
                    if ($productDiscount === 0.00 &&  $actualValue > $this->util->removeDecimal($remaingDiscount)) {
                        $productDiscount = $remaingDiscount;
                    } else if ($productDiscount === 0.00 && $actualValue < $this->util->removeDecimal($remaingDiscount)) {
                        $productDiscount = $this->util->subtract($productValue, $item['discount']) - 0.01;
                        // $remaingDiscount = $this->util->subtract($remaingDiscount, $productDiscount);    
                    } else if ($productDiscount === 0.00 && $actualValue == $this->util->removeDecimal($remaingDiscount)) {
                        $productDiscount = $remaingDiscount - 0.01;
                        $remaingDiscount = 0.01;
                    }
                    $remaingValue -= $productValue;
                    // echo "\nProduct Value: $productValue";
                    if ($productDiscount >= $productValue) {
                        $productDiscount = $productValue - 0.01;
                        $remaingDiscount = $this->util->subtract($remaingDiscount, $productDiscount);
                    } else {
                        // echo "\n_______________Product Discount: $productDiscount Remaing Discount: $remaingDiscount _______________";
                        $remaingDiscount =  $this->util->subtract($this->util->removeDecimal($remaingDiscount), $this->util->removeDecimal($productDiscount)) * $item['quantity'];
                        // echo "\n######Remaing Discount: $remaingDiscount######";
                    }
                    $item['discount'] += $this->util->removeDecimal($productDiscount);
                    $item['elegible'] = $this->util->removeDecimal($productValue - 0.01) > $item['discount'];
                    /*if($isObs){
                             echo  ("\n============================\nobsName: {$item['obsName']}\nPrice: {$item['price']}\nQuantity: {$item['quantity']}\nDiscount: {$item['discount']}\n============================");
                         }else{
                            echo  ("\n============================\nproductName: {$item['productName']}\nUnitValue: {$item['unitValue']}\nQuantity: {$item['quantity']}\nDiscount: {$item['discount']}\n============================");
                         }
                    echo  ("\nProduct Value ".$productValue);  
                  echo  ("\nProduct Discount ".$productDiscount);
                    echo  ("\nRemaing Value ".$remaingValue);   
                 echo  ("\nRemaing Discount ".$remaingDiscount);    
                echo  ("\n-----------------------------------------------------"); */
                }
            }
        }
    }

    public function getDiscountValue($item)
    {
        $discount = 0;
        // echo "\nDesconto inicial: $discount Item: {$item['productName']}";
        if (!empty($item['observations'])) {
            $observations  = $item['observations'];
            foreach ($observations as $observation) {
                $discount = $this->util->add($discount, $observation['discount']);
                // echo "\nValor: $discount Observacao: {$observation['obsName']}";
            }
        }
        if (!empty($item['childItems'])) {
            $childItems  = $item['childItems'];
            foreach ($childItems as $childItem) {
                $discount = $this->util->add($discount, $this->getDiscountValue($childItem));
            }
        } else {
            $discount = $this->util->add($discount, $item['discount']);
            // echo "\nItem sem filho: {$item['productName']} Desconto: $discount";
        }
        /*if(isset($item['productName'])){
            echo "\nDesconto: $discount\nItem: {$item['productName']}";
        }else{
            echo "\nDesconto: $discount\nObs Item: {$item['obsName']}";
        }
        echo "\n";*/
        return $discount;
    }

    public function saveItem(&$item, $subsidiaryCode, $vendarestSeq, $comandavenSeq, $itemsSeq, $positionNumber, $idcontroprod, $sellerPos, $storeCode, $sellerCode, $orderCPF, &$itemNum, $comboProductCode = null, $toTravelData, $pagerId = null, $holdTime, $driveThruData, $orderType, $toTravel, &$quantItems)
    {
        // echo json_encode($item);
        // echo "\n --------------------------------- \n";
        if (!empty($item['childItems'])) {
            if (!empty($item['note'])) $item['childItems'][0]['note'] .= '; ' . $item['note'];
            $childItems       = $item['childItems'];
            $comboProductCode = $item['productCode'];
            $comboQuantity    = $item['quantity'];
            $toTravel = !empty($item['toTravel']) ? $item['toTravel'] : null;
            for ($save = 0; $save < $comboQuantity; $save++) {
                $itemNum++;
                foreach ($childItems as &$childItem)
                    $this->saveItem($childItem, $subsidiaryCode, $vendarestSeq, $comandavenSeq, $itemsSeq, $positionNumber, $idcontroprod, $sellerPos, $storeCode, $sellerCode, $orderCPF, $itemNum, $comboProductCode, $toTravelData, $pagerId, $holdTime, $driveThruData, $orderType, $toTravel, $quantItems);
            }
        } else {

            $note = !empty($item['note']) ? $item['note'] : null;
            $realProductQuantity  = $item['quantity'];
            $item['observations'] = $item['observations'] === null ? array() : $item['observations'];
            $toTravel = isset($item['toTravel']) && $item['toTravel'] !== null ? $item['toTravel'] : $toTravel;
            $idorigemvenda = null;
            if (!empty($toTravel) && substr($orderType, 0, 3) !== 'TGO' && substr($orderType, 0, 3) !== 'DLV' && substr($orderType, 0, 3) !== 'THR') {
                $idorigemvenda = 'TGO_' . substr($orderType, 0, 3);
                // $toTravelData = $this->getToTravelData();
            }
            /*if(!empty($toTravelData)){
                $alreadyHasToTravel = false;
                foreach($item['observations'] as $observation){
                    if($observation['obsCode'] === $toTravelData['obsCode'] && $observation['obsGroupCode'] === $toTravelData['obsGroupCode'])
                        $alreadyHasToTravel = true;
                }
                if(!$alreadyHasToTravel)
                    array_push($item['observations'],$toTravelData);
            } */
            if (!empty($driveThruData)) {
                $alreadyHasDriveThru = false;
                foreach ($item['observations'] as $observation) {
                    if ($observation['obsCode'] === $driveThruData['obsCode'] && $observation['obsGroupCode'] === $driveThruData['obsGroupCode'])
                        $alreadyHasDriveThru = true;
                }
                if (!$alreadyHasDriveThru)
                    array_push($item['observations'], $driveThruData);
            }
            for ($save = 0; $save < $realProductQuantity; $save++) {
                if ($realProductQuantity < 1)
                    $productQuantity = $realProductQuantity;
                else
                    $productQuantity = 1;
                $productCode        = $item['productCode'];
                $productValue       = $item['unitValue'];
                $productDiscount    = $item['discount'];
                $productAcr         = 0;
                $itcomandavenSeq    = $quantItems;
                $quantItems++;
                $productDescription = '';
                if (!empty($item['observations']) || !empty($note))
                    $productDescription = $this->setProductDescription($note, $item['observations']);

                $idstprcomven = $this->util->getTheThing($subsidiaryCode, $storeCode, $productCode, $idcontroprod);
                $returnItcomandaven = $this->insertItComandaVen($subsidiaryCode, $vendarestSeq, $comandavenSeq, $itcomandavenSeq, $productCode, $productQuantity, $productValue, $productDescription, $idstprcomven, $productDiscount, $storeCode, $productAcr, $itemNum, $itemsSeq, $sellerPos, $comboProductCode, $itemNum, $sellerCode, $orderCPF, $positionNumber, $pagerId, $holdTime, $idorigemvenda, $note, "N");

                if ($returnItcomandaven['error']) {
                    throw new \Exception($returnItcomandaven['message']);
                }
                foreach ($item['observations'] as $observation) {
                    $obsCode      = $observation['obsCode'];
                    $obsGroupCode = $observation['obsGroupCode'];
                    $obsValue     = $observation['price'];
                    $obsQuantity  = $observation['quantity'];
                    $obsDiscount  = !empty($observation['discount']) ? $observation['discount'] : 0;
                    if (!$obsCode)
                        throw new \Exception("Código de observação inválido!");
                    if (!$obsGroupCode)
                        throw new \Exception("Código de grupo de observação inválido");
                    $returnObsitcomandaven = $this->insertObsItComandaVen($subsidiaryCode, $vendarestSeq, $comandavenSeq, $itcomandavenSeq, $obsGroupCode, $obsCode);
                    $observationProduct    = $this->util->getObservationInfo($obsGroupCode, $obsCode);
                    if (!empty($observationProduct)) {
                        $idstprcomven       = $this->util->getTheThing($subsidiaryCode, $storeCode, $observationProduct['CDPRODUTO'], $idcontroprod);
                        $itcomandavenSeqObs = $quantItems;
                        $quantItems++;
                        $returnItcomandaven = $this->insertItComandaVen($subsidiaryCode, $vendarestSeq, $comandavenSeq, $itcomandavenSeqObs, $observationProduct['CDPRODUTO'], $obsQuantity, $obsValue, '', $idstprcomven, $obsDiscount, $storeCode, 0, $itemNum, $itemsSeq, $sellerPos, null, $itemNum, $sellerCode, $orderCPF, $positionNumber, $pagerId, $holdTime, $idorigemvenda, null, "N");
                        if ($returnItcomandaven['error']) {
                            throw new \Exception($returnItcomandaven['message']);
                        }
                    }
                    if ($returnObsitcomandaven['error']) {
                        throw new \Exception($returnObsitcomandaven['message']);
                    }
                }
            }
        }
    }

    public function saveItemsEx(&$item, $subsidiaryCode, $vendarestSeq, $comandavenSeq, $itemsSeq, $positionNumber, $idcontroprod, $sellerPos, $storeCode, $sellerCode, $orderCPF, &$itemNum, $comboProductCode = null, $toTravelData, $pagerId = null, $holdTime, $driveThruData, $orderType, $toTravel, &$quantItems, &$quantCombos)
    {

        /* Exige o envio da informação imprime S imprime N */
        $productCode = $item['productCode'];
        if (empty($item['productPrint']) && empty($item['subProductsGroupPrint'])) throw new \Exception("O campo 'productPrint' deve ser enviado no item $productCode .");

        // $idImpr = !empty($item['productPrint']) ? $item['productPrint'] : $item['subProductsGroupPrint'];

        if (!empty($item['childItems'])) {
            /* Caso o item pai seja imprime N salva apenas os filhos na ITCOMANDAVEN */
            if ($item['productPrint'] == "N") {
                if (!empty($item['note'])) $item['childItems'][0]['note'] .= '; ' . $item['note'];
                $childItems       = $item['childItems'];
                $comboProductCode = $item['productCode'];
                $comboQuantity    = $item['quantity'];
                $toTravel = !empty($item['toTravel']) ? $item['toTravel'] : null;
                for ($save = 0; $save < $comboQuantity; $save++) {
                    $itemNum++;
                    $nrCombo = $quantItems - 1;
                    foreach ($childItems as &$childItem) {
                        $this->saveItemEx($childItem, $subsidiaryCode, $vendarestSeq, $comandavenSeq, $itemsSeq, $positionNumber, $idcontroprod, $sellerPos, $storeCode, $sellerCode, $orderCPF, $itemNum, $comboProductCode, $quantCombos, $toTravelData, $pagerId, $holdTime, $driveThruData, $orderType, $toTravel, $quantItems, "N", true, false, $nrCombo);
                        
                        // contempla terceiro nivel de produtos (combo do combo)
                        if(!empty($childItem['childItems'])) {
                            foreach($childItem['childItems'] as &$subChildItem){
                                $this->saveItemEx($subChildItem, $subsidiaryCode, $vendarestSeq, $comandavenSeq, $itemsSeq, $positionNumber, $idcontroprod, $sellerPos, $storeCode, $sellerCode, $orderCPF, $itemNum, $comboProductCode, $quantCombos, $toTravelData, $pagerId, $holdTime, $driveThruData, $orderType, $toTravel, $quantItems, "N", true, false, $nrCombo);
                            }
                        } 
                    }
                }
            } else if ($item['productPrint'] == "S") {
                if (!empty($item['note'])) $item['childItems'][0]['note'] .= '; ' . $item['note'];
                $childItems       = $item['childItems'];
                $comboProductCode = $item['productCode'];
                $comboQuantity    = $item['quantity'];
                $toTravel = !empty($item['toTravel']) ? $item['toTravel'] : null;

                /* Confere se a promoção combinada é um item vendido dentro dela mesma */
                $key = array_search($item['productCode'], array_column($childItems, 'productCode'));

                /* Caso NAO seja vendido dentro dela mesma, divide o preço da combinacao combinada igualmente entre os itens salvos na ITCOMANDAEST */
                if ($key === false) {
                    $generalPrice = $this->util->div($item['unitValue'], count($childItems));
                } else if ($key !== false) {
                    $generalPrice = null;
                }
                                
                /* Caso a promoção combinada seja imprime S, salva o pai na ITEMVENDA */
                $nrCombo = $quantItems - 1;
                $this->saveItemEx($item, $subsidiaryCode, $vendarestSeq, $comandavenSeq, $itemsSeq, $positionNumber, $idcontroprod, $sellerPos, $storeCode, $sellerCode, $orderCPF, $itemNum, $comboProductCode, $quantCombos, $toTravelData, $pagerId, $holdTime, $driveThruData, $orderType, $toTravel, $quantItems, "N", true, false, $nrCombo);
                                
                for ($save = 0; $save < $comboQuantity; $save++) {
                    $itemNum++;
                    $nrCombo = $quantItems - 1;
                    
                    /* Verifica se ocorrerá erro de PK  na ITCOMANDAEST, quando dois filhos de mesmo código são enviados para inserir na tabela */
                    $this->correctQuantityChildItems($childItems);

                    foreach ($childItems as &$childItem) {
                        
                        if ($childItem['subProductsGroupPrint'] == "S") {
                            $insertItComVen = isset($childItem['insertItComVen']) ? $childItem['insertItComVen'] : true;
                            $insertItComEst = isset($childItem['insertItComEst']) ? $childItem['insertItComEst'] : true;
                            $this->saveItemEx($childItem, $subsidiaryCode, $vendarestSeq, $comandavenSeq, $itemsSeq, $positionNumber, $idcontroprod, $sellerPos, $storeCode, $sellerCode, $orderCPF, $itemNum, null, $quantCombos, $toTravelData, $pagerId, $holdTime, $driveThruData, $orderType, $toTravel, $quantItems, "S", $insertItComVen, $insertItComEst, $nrCombo);
                            
                        } else if ($childItem['subProductsGroupPrint'] == "N") {
                            /* Continua a regra de preço por promocao combinda contida em si mesmo ou nao */
                            if ($generalPrice == null) {
                                $childItem['price'] = 0;
                            } else {
                                $childItem['price'] = $generalPrice;
                            }
                            
                            $quantItems--;
                            $insertItComVen = isset($childItem['insertItComVen']) ? $childItem['insertItComVen'] : false;
                            $insertItComEst = isset($childItem['insertItComEst']) ? $childItem['insertItComEst'] : true;
                            $this->saveItemEx($childItem, $subsidiaryCode, $vendarestSeq, $comandavenSeq, $itemsSeq, $positionNumber, $idcontroprod, $sellerPos, $storeCode, $sellerCode, $orderCPF, $itemNum, $comboProductCode, $quantCombos, $toTravelData, $pagerId, $holdTime, $driveThruData, $orderType, $toTravel, $quantItems, "N", $insertItComVen, $insertItComEst, $nrCombo);   
                        }

                        // contempla terceiro nivel de produtos (combo do combo)
                        if(!empty($childItem['childItems'])) {
                            $this->correctQuantityChildItems($childItem['childItems']);
                            foreach($childItem['childItems'] as &$subChildItem){
                                if ($subChildItem['subProductsGroupPrint'] == "S") {
                                    $insertItComVen = isset($subChildItem['insertItComVen']) ? $subChildItem['insertItComVen'] : true;
                                    $insertItComEst = isset($subChildItem['insertItComEst']) ? $subChildItem['insertItComEst'] : true;
                                    $this->saveItemEx($subChildItem, $subsidiaryCode, $vendarestSeq, $comandavenSeq, $itemsSeq, $positionNumber, $idcontroprod, $sellerPos, $storeCode, $sellerCode, $orderCPF, $itemNum, null, $quantCombos, $toTravelData, $pagerId, $holdTime, $driveThruData, $orderType, $toTravel, $quantItems, "S", $insertItComVen, $insertItComEst, $nrCombo);
                                    
                                } else if ($subChildItem['subProductsGroupPrint'] == "N") {
                                    /* Continua a regra de preço por promocao combinda contida em si mesmo ou nao */
                                    // if ($generalPrice == null) {
                                    //     $subChildItem['price'] = 0;
                                    // } else {
                                    //     $subChildItem['price'] = $generalPrice;
                                    // }
                                    
                                    // $quantItems--;
                                    $insertItComVen = isset($subChildItem['insertItComVen']) ? $subChildItem['insertItComVen'] : false;
                                    $insertItComEst = isset($subChildItem['insertItComEst']) ? $subChildItem['insertItComEst'] : true;
                                    $this->saveItemEx($subChildItem, $subsidiaryCode, $vendarestSeq, $comandavenSeq, $itemsSeq, $positionNumber, $idcontroprod, $sellerPos, $storeCode, $sellerCode, $orderCPF, $itemNum, $comboProductCode, $quantCombos, $toTravelData, $pagerId, $holdTime, $driveThruData, $orderType, $toTravel, $quantItems, "N", $insertItComVen, $insertItComEst, $nrCombo);   
                                }
                            }
                        } 
                    }
                }
                
            }
            $quantCombos++;
        } else {
            
            $this->saveItemEx($item, $subsidiaryCode, $vendarestSeq, $comandavenSeq, $itemsSeq, $positionNumber, $idcontroprod, $sellerPos, $storeCode, $sellerCode, $orderCPF, $itemNum, $comboProductCode, $quantCombos, $toTravelData, $pagerId, $holdTime, $driveThruData, $orderType, $toTravel, $quantItems, "N", true, false, null);
            
        }
    }

    private function saveItemEx(&$item, $subsidiaryCode, $vendarestSeq, $comandavenSeq, $itemsSeq, $positionNumber, $idcontroprod, $sellerPos, $storeCode, $sellerCode, $orderCPF, &$itemNum, $comboProductCode = null, $comboSeq, $toTravelData, $pagerId = null, $holdTime, $driveThruData, $orderType, $toTravel, &$quantItems, $idImpr, $insertItComVen, $insertItComEst, $nrCombo)
    {
        // echo json_encode($item, JSON_PRETTY_PRINT)."\n";
        // echo "insertItComVen: ".$insertItComVen." . insertItComEst: ".$insertItComEst."\n";
        $note = !empty($item['note']) ? $item['note'] : null;
        $realProductQuantity  = $item['quantity'];
        $item['observations'] = $item['observations'] === null ? array() : $item['observations'];
        $toTravel = isset($item['toTravel']) && $item['toTravel'] !== null ? $item['toTravel'] : $toTravel;
        $idorigemvenda = null;
        if (!empty($toTravel) && substr($orderType, 0, 3) !== 'TGO' && substr($orderType, 0, 3) !== 'DLV' && substr($orderType, 0, 3) !== 'THR') {
            $idorigemvenda = 'TGO_' . substr($orderType, 0, 3);
            // $toTravelData = $this->getToTravelData();
        }
        /*if(!empty($toTravelData)){
            $alreadyHasToTravel = false;
            foreach($item['observations'] as $observation){
                if($observation['obsCode'] === $toTravelData['obsCode'] && $observation['obsGroupCode'] === $toTravelData['obsGroupCode'])
                    $alreadyHasToTravel = true;
            }
            if(!$alreadyHasToTravel)
                array_push($item['observations'],$toTravelData);
        } */
        if (!empty($driveThruData)) {
            $alreadyHasDriveThru = false;
            foreach ($item['observations'] as $observation) {
                if ($observation['obsCode'] === $driveThruData['obsCode'] && $observation['obsGroupCode'] === $driveThruData['obsGroupCode'])
                    $alreadyHasDriveThru = true;
            }
            if (!$alreadyHasDriveThru)
                array_push($item['observations'], $driveThruData);
        }
        for ($save = 0; $save < $realProductQuantity; $save++) {
            if ($realProductQuantity < 1)
                $productQuantity = $realProductQuantity;
            else
                $productQuantity = 1;
            $productCode        = $item['productCode'];
            $productValue       = $item['unitValue'];
            $productDiscount    = $item['discount'];
            $productAcr         = 0;
            $itcomandavenSeq    = $quantItems;
            $quantItems++;
            $productDescription = '';
            if (!empty($item['observations']) || !empty($note))
                $productDescription = $this->setProductDescription($note, $item['observations']);

            
            $idstprcomven = $this->util->getTheThing($subsidiaryCode, $storeCode, $productCode, $idcontroprod);
            
            if ($insertItComVen) {
                // echo "productQuantity comven: ".$productQuantity."\n";
                $returnItcomandaven = $this->insertItComandaVen($subsidiaryCode, $vendarestSeq, $comandavenSeq, $itcomandavenSeq, $productCode, $productQuantity, $productValue, $productDescription, $idstprcomven, $productDiscount, $storeCode, $productAcr, $itemNum, $itemsSeq, $sellerPos, $comboProductCode, $comboSeq, $sellerCode, $orderCPF, $positionNumber, $pagerId, $holdTime, $idorigemvenda, $note, $idImpr);
                
                if ($returnItcomandaven['error']) {
                    throw new \Exception($returnItcomandaven['message']);
                }
            }
            
            if ($insertItComEst) {
                $productQuantity = (!empty($item['quantityItComandaEst'])) ? $item['quantityItComandaEst'] : $productQuantity;
                $returnItcomandaest = $this->insertItComandaEst($subsidiaryCode, $vendarestSeq, $comandavenSeq, $itcomandavenSeq, $productCode, $productQuantity, $productValue, $productDiscount, $productAcr, $idImpr, $nrCombo);
                
                if ($returnItcomandaest['error']) {
                    throw new \Exception($returnItcomandaest['message']);
                }
            }

            foreach ($item['observations'] as $observation) {
                $obsCode      = $observation['obsCode'];
                $obsGroupCode = $observation['obsGroupCode'];
                $obsValue     = $observation['price'];
                $obsQuantity  = $observation['quantity'];
                $obsDiscount  = !empty($observation['discount']) ? $observation['discount'] : 0;
                if (!$obsCode)
                    throw new \Exception("Código de observação inválido!");
                if (!$obsGroupCode)
                    throw new \Exception("Código de grupo de observação inválido");
                $returnObsitcomandaven = $this->insertObsItComandaVen($subsidiaryCode, $vendarestSeq, $comandavenSeq, $itcomandavenSeq, $obsGroupCode, $obsCode);
                $observationProduct    = $this->util->getObservationInfo($obsGroupCode, $obsCode);
                if (!empty($observationProduct)) {
                    $idstprcomven       = $this->util->getTheThing($subsidiaryCode, $storeCode, $observationProduct['CDPRODUTO'], $idcontroprod);
                    $itcomandavenSeqObs = $quantItems;
                    $quantItems++;
                    $returnItcomandaven = $this->insertItComandaVen($subsidiaryCode, $vendarestSeq, $comandavenSeq, $itcomandavenSeqObs, $observationProduct['CDPRODUTO'], $obsQuantity, $obsValue, '', $idstprcomven, $obsDiscount, $storeCode, 0, $itemNum, $itemsSeq, $sellerPos, null, null, $sellerCode, $orderCPF, $positionNumber, $pagerId, $holdTime, $idorigemvenda, null, $idImpr);

                    if ($returnItcomandaven['error']) {
                        throw new \Exception($returnItcomandaven['message']);
                    }
                }
                if ($returnObsitcomandaven['error']) {
                    throw new \Exception($returnObsitcomandaven['message']);
                }
            }
        }
        
    }

    private function setProductDescription($note, $observations)
    {
        if (!empty($note))
            $txprodcomven = ' ' . $note;
        else
            $txprodcomven = ' ';

        foreach ($observations as $observation) {
            $obsName = $observation['obsName'];
            if ($txprodcomven == null) {
                $txprodcomven =  $obsName;
            } else {
                $txprodcomven .= '; ' . $obsName;
            }
        }

        return $txprodcomven;
    }

    public function insertVendaRest($subsidiaryCode, $vendarestSeq, $storeCode, $sellerCode, $operatorCode, $clientCode, $consumerCode, $standardTable)
    {

        $params = array(
            'CDFILIAL'     => $subsidiaryCode,
            'NRVENDAREST'  => $vendarestSeq,
            'CDLOJA'       => $storeCode,
            'NRMESA'       => $standardTable,
            'CDVENDEDOR'   => $sellerCode,
            'CDOPERADOR'   => $operatorCode,
            'NRPESMESAVEN' => '1',
            'CDCLIENTE'    => $clientCode,
            'CDCONSUMIDOR' => $consumerCode,
            'IDPEDIDOPAGO' => 'S'
        );

        return $this->orderService->insertVendaRest($params);
    }

    protected function insertComandaVen($subsidiaryCode, $vendarestSeq, $comandavenSeq, $storeCode, $IDSTCOMANDA, $deliveryFee, $country, $state, $cityCode, $zipCode, $neighborhood, $fullStreetAddress, $addressLine2, $reference, $toGo, $orderToken, $syncScheduled, $scheduledDate, $kdsChangeStatusOnExp, $consumerDescription, $orderType, $cashBack, $orderDescription, $dscomanda, $closeByApi, $orderCPF = null)
    {
        
        $params = array(
            'CDFILIAL'        => $subsidiaryCode,
            'NRVENDAREST'     => $vendarestSeq,
            'NRCOMANDA'       => $comandavenSeq,
            'CDLOJA'          => $storeCode,
            'IDSTCOMANDA'     => $IDSTCOMANDA,
            'DSCOMANDA'       => $dscomanda,
            'SGSEXOCON'       => 'M',
            'TXMOTIVCANCE'    => " ",
            'VRACRCOMANDA'    => $deliveryFee,
            'CDPAIS'          => $country,
            'SGESTADO'        => $state,
            'CDMUNICIPIO'     => $cityCode,
            'NRCEPCONSCOMAND' => $zipCode,
            'CDBAIRRO'        => null,
            'DSBAIRRO'        => $neighborhood,
            'DSENDECONSCOMAN' => $fullStreetAddress,
            'DSCOMPLENDCOCOM' => $addressLine2,
            'DSREFENDCONSCOM' => $reference,
            'IDORGCMDVENDA'   => $orderType,
            'IDRETBALLOJA'    => $toGo,
            'NRCOMANDAEXT'    => $orderToken,
            'IDSINCAGENDA'    => $syncScheduled,
            'DTHRAGENDADA'    => $scheduledDate,
            'IDIMPCUPEXP'     => $kdsChangeStatusOnExp,
            'DSCONSUMIDOR'    => $consumerDescription,
            'DSOBSCOMANDA'    => $orderDescription,
            'IDFECHAMENTOAPI' => $closeByApi,
        );

        $fidelityUtil = $this->util->getFidelityUtil($subsidiaryCode);
        if(!empty($fidelityUtil) && $fidelityUtil["IDUTILFIDUDS"] == "S"){
            $params['VRDESCFID'] = 0;
            $params['VRDESCCASHUDS'] = $cashBack;
            $params['NRCPFFIDUDS'] = $orderCPF;
        } else {
            $params['VRDESCFID'] = $cashBack;
            $params['VRDESCCASHUDS'] = 0;
            $params['NRCPFFIDUDS'] = null;
        }

        return $this->orderService->insertComandaven($params);
    }

    protected function insertComandaVenEx($subsidiaryCode, $vendarestSeq, $comandavenSeq, $storeCode, $IDSTCOMANDA, $deliveryFee, $country, $state, $cityCode, $zipCode, $neighborhood, $fullStreetAddress, $addressLine2, $reference, $toGo, $orderToken, $syncScheduled, $scheduledDate, $kdsChangeStatusOnExp, $consumerDescription, $orderType, $orderDescription, $dscomanda, $closeByApi, $cashBack, $orderCPF)
    {
        $params = array(
            'CDFILIAL'        => $subsidiaryCode,
            'NRVENDAREST'     => $vendarestSeq,
            'NRCOMANDA'       => $comandavenSeq,
            'CDLOJA'          => $storeCode,
            'IDSTCOMANDA'     => $IDSTCOMANDA,
            'DSCOMANDA'       => $dscomanda,
            'SGSEXOCON'       => 'M',
            'TXMOTIVCANCE'    => " ",
            'VRACRCOMANDA'    => $deliveryFee,
            'CDPAIS'          => $country,
            'SGESTADO'        => $state,
            'CDMUNICIPIO'     => $cityCode,
            'NRCEPCONSCOMAND' => $zipCode,
            'CDBAIRRO'        => null,
            'DSBAIRRO'        => $neighborhood,
            'DSENDECONSCOMAN' => $fullStreetAddress,
            'DSCOMPLENDCOCOM' => $addressLine2,
            'DSREFENDCONSCOM' => $reference,
            'IDORGCMDVENDA'   => $orderType,
            'IDRETBALLOJA'    => $toGo,
            'NRCOMANDAEXT'    => $orderToken,
            'IDSINCAGENDA'    => $syncScheduled,
            'DTHRAGENDADA'    => $scheduledDate,
            'IDIMPCUPEXP'     => $kdsChangeStatusOnExp,
            'DSCONSUMIDOR'    => $consumerDescription,
            'VRDESCFID'       => 0,
            'VRDESCCASHUDS'   => $cashBack,
            'DSOBSCOMANDA'    => $orderDescription,
            'IDFECHAMENTOAPI' => $closeByApi,
            'NRCPFFIDUDS'     => $orderCPF
        );
        return $this->orderService->insertComandaven($params);
    }

    public function insertMovCaixaDlv($subsidiaryCode, $vendarestSeq, $movcaixadlvSeq, $movementType, $clientCode, $consumerCode, $paymentCode, $paymentValue, $nsu, $quantParcel, $authHash = null)
    {

        $params = array(
            'CDFILIAL'        => $subsidiaryCode,
            'NRVENDAREST'     => $vendarestSeq,
            'NRSEQMOVDLV'     => $movcaixadlvSeq,
            'IDTIPOMOVIVEDLV' => $movementType,
            'CDCLIENTE'       => $clientCode,
            'CDCONSUMIDOR'    => $consumerCode,
            'CDTIPORECE'      => $paymentCode,
            'VRMOVIVENDDLV'   => $paymentValue,
            'CDNSUTEFDLV'     => $nsu,
            'QTPARCRECEB'     => $quantParcel,
            'NRAUTORIZATEFDLV' => $authHash
        );

        return $this->orderService->insertMovCaixaDlv($params);
    }

    protected function insertItComandaVen($subsidiaryCode, $vendarestSeq, $comandavenSeq, $itcomandavenSeq, $productCode, $productQuantity, $productValue, $productDescription, $idstprcomven, $productDiscount, $storeCode, $productAcr, $itemNum, $itemsSeq, $sellerPos, $comboProductCode, $comboSeq, $sellerCode, $orderCPF, $positionNumber, $pagerId, $holdTime, $idorigemvenda, $note, $idImpr)
    {
        if ($comboProductCode == null) {
            $comboSeq = null;
        } 
        else {
        //     $prodSeq = str_pad($itemNum, 3, '0', STR_PAD_LEFT);
            $comboSeq = str_pad($comboSeq, 3, '0', STR_PAD_LEFT);
        }
        $params  = array(
            'CDFILIAL'       => $subsidiaryCode,
            'NRVENDAREST'    => $vendarestSeq,
            'NRCOMANDA'      => $comandavenSeq,
            'NRPRODCOMVEN'   => str_pad($itcomandavenSeq, 6, '0', STR_PAD_LEFT),
            'CDPRODUTO'      => $productCode,
            'QTPRODCOMVEN'   => $productQuantity,
            'VRPRECCOMVEN'   => $productValue,
            'TXPRODCOMVEN'   => $productDescription,
            'IDSTPRCOMVEN'   => $idstprcomven,
            'VRDESCCOMVEN'   => $productDiscount,
            'NRLUGARMESA'    => $positionNumber,
            'IDPRODIMPFIS'   => 'N',
            'CDLOJA'         => $storeCode,
            'VRACRCOMVEN'    => $productAcr,
            'NRSEQPRODCOM'   => $comboSeq,
            'NRSEQPRODCUP'   => $itemsSeq,
            'VRPRECCLCOMVEN' => 0,
            'CDCAIXACOLETOR' => $sellerPos,
            'CDPRODPROMOCAO' => $comboProductCode,
            'CDVENDEDOR'     => $sellerCode,
            'NRINSCRCONS'    => $orderCPF,
            'CDSENHAPED'     => $pagerId,
            'NRATRAPRODCOVE' => $holdTime,
            'IDORIGEMVENDA'  => $idorigemvenda,
            'DSOBSPEDDIGCMD' => $note,
            'IDIMPPRGRPROMO' => $idImpr
        );
        return $this->orderService->insertItComandaVen($params);
    }

    protected function insertItComandaEst($subsidiaryCode, $vendarestSeq, $comandavenSeq, $itcomandavenSeq, $productCode, $productQuantity, $productValue, $productDiscount, $productAcr, $idImpr, $nrCombo)
    {
        //     if($comboProductCode == null){
        //         $prodSeq = null;
        //     }else{
        //         $prodSeq = str_pad($itemNum,3,'0',STR_PAD_LEFT); 
        //     }
        $params  = array(
            'CDFILIAL' => $subsidiaryCode,
            'NRVENDAREST' =>  $vendarestSeq,
            'NRCOMANDA' => $comandavenSeq,
            // 'NRPRODCOMVEN' => str_pad($itcomandavenSeq, 6, '0', STR_PAD_LEFT),
            'NRPRODCOMVEN' => str_pad($nrCombo, 6, '0', STR_PAD_LEFT),
            'CDPRODUTO' => $productCode,
            'QTPROCOMEST' => $productQuantity,
            'VRPRECCOMEST' => $productValue,
            'VRDESITCOMEST' => $productDiscount,
            'IDIMPPRGRPROMO' => $idImpr,
            'VRACRITCOMEST' => $productAcr
        );
        return $this->orderService->insertItComandaEst($params);
    }

    public function getIdcontroprod($storeCode, $subsidiaryCode)
    {
        return $this->util->getIdcontroprod($storeCode, $subsidiaryCode);
    }

    protected function insertObsItComandaVen($subsidiaryCode, $vendarestSeq, $comandavenSeq, $itcomandavenSeq, $obsGroupCode, $obsCode)
    {

        $params = array(
            'CDFILIAL'     => $subsidiaryCode,
            'NRVENDAREST'  => $vendarestSeq,
            'NRCOMANDA'    => $comandavenSeq,
            'NRPRODCOMVEN' => str_pad($itcomandavenSeq, 6, '0', STR_PAD_LEFT),
            'CDGRPOCOR'    => $obsGroupCode,
            'CDOCORR'      => $obsCode
        );
        return $this->orderService->insertObsItComandaVen($params);
    }

    public function isValid(Request $request, Response $response)
    {
        $inputJSON  = file_get_contents('php://input');
        $this->util->saveLog($inputJSON, 'Order', 'isValid');
        $input      = json_decode($inputJSON, true);

        try {
            if (!isset($input)) {
                throw new \Exception('Não foram recebidos parâmetros');
            }
            $data   = array(); // dados a serem retornados

            $params = $input;
            // --- Order
            $order              = $params['order'];
            $clientCode         = $order['clientCode'];
            $consumerCode       = $order['consumerCode'];
            $balanceGroupCode   = $order['balanceGroup'];
            $subsidiaryCode     = $order['subsidiaryCode'];
            $qualidade          = isset($params['Qualidade']) ? $params['Qualidade'] : false;

            if (!empty($clientCode) && !empty($consumerCode)) {
                // verifica se existe cliente cadastrado com essas informações
                if (!$this->util->consumerExists($clientCode, $consumerCode)) {
                    throw new \Exception('Consumidor inexistente.');
                }
            } else {
                $defaultClient = $this->util->getDefaultClient($subsidiaryCode);

                if (!empty($defaultClient))
                    $clientCode = $defaultClient;
                else
                    throw new \Exception('Cliente padrão inexistente.');
            }

            $storeCode      = $order['storeCode'];
            $items          = $order['items'];
            $products       = array();
            $obsProducts    = array();

            $totalValue = 0;

            foreach ($items as $item) {
                $products[$item['productCode']] =  $item['productName'];

                if (!empty($item['childItems'])) {
                    $childItems = $item['childItems'];

                    foreach ($childItems as $childItem) {
                        $totalValue = +$this->util->float(($childItem['unitValue'] * $childItem['quantity']) - $childItem['discount']);
                        $products[$childItem['productCode']] =  $childItem['productName'];

                        if (!empty($childItem['observations'])) {
                            $observations = $childItem['observations'];

                            foreach ($observations as $observation) {
                                if (isset($observation['price'])) {
                                    $totalValue = +$this->util->float($observation['price'] * $observation['quantity']);
                                    $obsGroupCode = $observation['obsGroupCode'];
                                    $obsCode      = $observation['obsCode'];
                                    $obsProdCode  = $this->util->getObservationAsProduct($obsGroupCode, $obsCode);

                                    if ($obsProdCode !== null) {
                                        $obsProducts[$obsProdCode] = $observation['obsName'];
                                    }
                                }
                            }
                        }
                    }
                } else {
                    $totalValue = +$this->util->float(($item['unitValue'] * $item['quantity']) - $item['discount']);
                }

                if (!empty($item['observations'])) {
                    $observations = $item['observations'];

                    foreach ($observations as $observation) {

                        if (isset($observation['price'])) {
                            $totalValue = +$this->util->float($observation['price'] * $observation['quantity']);
                            $obsGroupCode = $observation['obsGroupCode'];
                            $obsCode      = $observation['obsCode'];
                            $obsProdCode  = $this->util->getObservationAsProduct($obsGroupCode, $obsCode);

                            if ($obsProdCode !== null) {
                                $obsProducts[$obsProdCode] = $observation['obsName'];
                            }
                        }
                    }
                }
            }

            $data = array();

            foreach ($products as $productCode => $productName) {
                if (!$this->util->productIsAvailable($subsidiaryCode, $storeCode, $productCode)) {
                    array_push($data, $productName);
                }
            }

            foreach ($obsProducts as $productCode => $productName) {
                if (!$this->util->productIsAvailable($subsidiaryCode, $storeCode, $productCode)) {
                    array_push($data, $productName);
                }
            }

            if (!empty($data)) {
                $message = 'Produtos indisponíveis';
                $response->addDataSet(new DataSet('', $data));
                $response->addMessage(new Message($message, self::ERROR_MESSAGE));
            }

            if (!empty($clientCode) && !empty($consumerCode)) {
                $balanceValue        = $this->util->getBalanceValue($clientCode, $consumerCode, $balanceGroupCode);
                $consumerBalanceInfo = $this->util->getConsumerBalanceInfo($clientCode, $consumerCode);
                $belowBalanceLimit   = false;

                if (!empty($consumerBalanceInfo['VRLIMCREDCONS'])) {
                    $belowBalanceLimit = ($this->util->float($totalValue) - $this->util->float($balanceValue)) < $consumerBalanceInfo['VRLIMCREDCONS'];
                }

                if ($balanceValue !== null && $consumerBalanceInfo['IDVERSALDCON'] == 'S' && $qualidade) {
                    if ($this->util->float($totalValue) > $this->util->float($balanceValue) && $belowBalanceLimit) {
                        $message = 'Saldo insuficiente';
                        $response->addDataSet(new DataSet('', $data));
                        $response->addMessage(new Message($message, self::ERROR_MESSAGE));
                    }
                }
            }
        } catch (\Exception $e) {
            $message = 'Erro ao validar venda. Erro: ' . $e->getMessage();
            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
        }
    }

    public function validateTableValue(Request $request, Response $response)
    {
        $inputJSON = file_get_contents('php://input');
        $input = json_decode($inputJSON, true);

        $this->util->beginTransaction();
        try {
            if (!isset($input)) {
                throw new \Exception('Não foram recebidos parâmetros');
            }

            $data   = array(); // dados a serem retornados

            $params         = $input;
            $checkinId      = $params['checkinId'];
            $usesServiceTax = $params['usesServiceTax'];
            $paymentValue   = $this->util->removeDecimal($params['paymentValue']);
            $paymentCode    = $params['paymentCode'];
            $positions      = $params['positions'];
            $currentValue   = 0;

            /*
             * Verifica se o checkinId é válido
             */
            $validCheckinId = $this->orderService->isValidCheckinId($checkinId);

            if ($validCheckinId['error']) {
                throw new \Exception($validCheckinId['message']);
            }

            /*
             * Obtém CDFILIAL E NRVENDAREST partindo de um IDCHECKIN
             */

            $resultCheckinMesPos = $this->getTableInfoByCheckin($checkinId);
            /* $resultCheckinMesPos = $this->orderService->getResultCheckinMesPos($checkinId);
            if($resultCheckinMesPos['error']){
                throw new \Exception($resultCheckinMesPos['message']);
            }
            $resultCheckinMesPos = $resultCheckinMesPos['data'];
            */

            $CDFILIAL       = $resultCheckinMesPos['CDFILIAL'];
            $NRVENDAREST    = $resultCheckinMesPos['NRVENDAREST'];
            $NRCOMANDA      = $resultCheckinMesPos['NRCOMANDA'];
            $CDLOJA         = $resultCheckinMesPos['CDLOJA'];
            $CDCLIENTE      = $resultCheckinMesPos['CDCLIENTE'];
            $CDCONSUMIDOR   = $resultCheckinMesPos['CDCONSUMIDOR'];
            /*
            * Obtém NRMESA
            */
            $resultGetNrmesa = $this->orderService->getNrmesa($CDFILIAL, $NRVENDAREST);

            if ($resultGetNrmesa['error']) {
                throw new \Exception($resultGetNrmesa['message']);
            }

            $paymentId = $this->util->getPaymentId($paymentCode);
            if ($paymentId === "9") {
                $returnVerify = $this->verifyDailyCreditLimit($paymentValue, $CDCLIENTE, $CDCONSUMIDOR, $CDFILIAL);
                if ($returnVerify['error']) {
                    throw new \Exception($returnVerify['message']);
                }
            }

            $resultGetNrmesa = $resultGetNrmesa['data'];
            $NRMESA = $resultGetNrmesa['NRMESA'];
            foreach ($positions as $position) {

                $positionNumber = $position['positionNumber'];
                $productsValue = 0;
                $orderItems = $this->util->getPositionItems($CDFILIAL, $NRVENDAREST, $NRCOMANDA, $positionNumber);

                foreach ($orderItems as $orderItem) {
                    $priceDetails = $this->getConsumerItemPrice($CDFILIAL, $CDCLIENTE, $orderItem['CDPRODUTO'], $CDLOJA, $CDCONSUMIDOR);
                    $discount     = $this->getNewDiscount($priceDetails, $orderItem);
                    $newUnitPrice = !empty($priceDetails['newPrice']) ? $priceDetails['newPrice'] : $orderItem['VRPRECCOMVEN'];

                    if ($newUnitPrice - $discount < 0.01)
                        $productsValue += 0.01;
                    else
                        $productsValue += $newUnitPrice - $discount;

                    $productsValue = bcdiv($productsValue, 1, 2);
                }

                $currentValue += $productsValue;
            }

            $serviceTax = 0;
            if ($usesServiceTax) {
                $storeServiceTax  = $this->util->getStoreServiceTax($CDFILIAL, $CDLOJA);
                $serviceTax       = $this->setServiceTax($storeServiceTax, $paymentValue);
            }

            $serviceTax   = $this->util->removeDecimal($serviceTax);
            $currentValue = $this->util->removeDecimal($currentValue);
            $currentValue = $this->util->removeDecimal($serviceTax + $currentValue);
            $paymentValue = $this->util->removeDecimal($paymentValue);

            $data         = array();
            $isValid      = $currentValue == $paymentValue;

            $responseObject = array(
                'isValid' => $isValid
            );

            array_push($data, $responseObject);

            $response->addDataSet(new DataSet('', $data));
            $this->util->commit();
        } catch (\Exception $e) {
            $this->util->rollBack();
            $message = 'Erro ao validar pagamento da mesa. Erro: ' . $e->getMessage();
            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
        }
    }

    private function setServiceTax($storeServiceTax, $paymentValue)
    {
        $storeRequiresTax = $storeServiceTax['IDCOMISVENDA'] == 'S';
        $servTaxAsProduct = $storeServiceTax['IDTRATTAXASERV'] == 'P';
        $servTaxProduct   = $storeServiceTax['CDPRODTAXASERV'];
        $servTaxValue     = $storeServiceTax['VRCOMISVENDA'];

        if (($storeRequiresTax) && (!empty($servTaxValue)) && (($servTaxAsProduct && !empty($servTaxProduct)) || (!$servTaxAsProduct))) {
            return $paymentValue / $servTaxValue;
        }

        return 0;
    }

    public function closeTableBill(Request $request, Response $response)
    {
        $inputJSON = file_get_contents('php://input');
        $this->util->saveLog($inputJSON, 'Order', 'closeTableBill');
        $params = json_decode($inputJSON, true);

        $this->util->beginTransaction();

        try {
            if (!isset($params)) {
                throw new \Exception('Não foram recebidos parâmetros');
            }

            $checkinId      = $params['checkinId'];
            $saveServiceTax = $params['usesServiceTax'];
            /*
             * Verifica se o checkinId é válido
             */
            $validCheckinId = $this->orderService->isValidCheckinId($checkinId);

            if ($validCheckinId['error']) {
                throw new \Exception($validCheckinId['message']);
            }
            /*
             * Obtém CDFILIAL E NRVENDAREST partindo de um IDCHECKIN
             */
            $resultCheckinMesPos = $this->getTableInfoByCheckin($checkinId);

            $CDFILIAL       = $resultCheckinMesPos['CDFILIAL'];
            $NRVENDAREST    = $resultCheckinMesPos['NRVENDAREST'];
            $CDLOJA         = $resultCheckinMesPos['CDLOJA'];
            $CDCLIENTE      = $resultCheckinMesPos['CDCLIENTE'];
            $CDCONSUMIDOR   = $resultCheckinMesPos['CDCONSUMIDOR'];
            $idcontroprod   = $this->getIdcontroprod($CDLOJA, $CDFILIAL);

            if ($saveServiceTax) {
                $storeServiceTax = $this->util->getStoreServiceTax($CDFILIAL, $CDLOJA);
            } else {
                $storeServiceTax = array();
            }

            /*
             * Obtém CDVENDEDOR, CDCAIXA E CDLOJA partindo de um CDFILIAL
             */
            $resultParavendAndVendedor = $this->orderService->getResultParavendAndVendedor($CDFILIAL);
            if ($resultParavendAndVendedor['error']) {
                throw new \Exception($resultParavendAndVendedor['message']);
            }

            $resultParavendAndVendedor = $resultParavendAndVendedor['data'];
            $CDVENDEDOR = $resultParavendAndVendedor['CDVENDEDOR'];
            $CDCAIXA    = $resultParavendAndVendedor['CDCAIXA'];

            /*
             * Obtém NRCOMANDA
             */
            $resultGetNrcomanda = $this->orderService->getNrcomanda($NRVENDAREST, $CDFILIAL);
            if ($resultGetNrcomanda['error']) {
                throw new \Exception($resultGetNrcomanda['message']);
            }

            $resultGetNrcomanda = $resultGetNrcomanda['data'];
            $NRCOMANDA = $resultGetNrcomanda['NRCOMANDA'];

            /*
            * Obtém NRMESA
            */
            $resultGetNrmesa = $this->orderService->getNrmesa($CDFILIAL, $NRVENDAREST);
            if ($resultGetNrmesa['error']) {
                throw new \Exception($resultGetNrmesa['message']);
            }

            $resultGetNrmesa = $resultGetNrmesa['data'];
            $NRMESA = $resultGetNrmesa['NRMESA'];
            $positions = $params['positions'];

            foreach ($positions as $positionToClose) {
                $return = $this->getProductsToRecalculate($CDFILIAL, $NRMESA, $positionToClose['positionNumber'], $CDLOJA);

                if (!$return['error']) {
                    $productsToRecalculate = $return['data'];
                    $productsData = $this->recalculateProductsPrice($productsToRecalculate, $CDFILIAL, $NRMESA, $positionToClose['positionNumber'], $CDLOJA, $CDCLIENTE, $CDCONSUMIDOR, true);
                }

                $return = $this->getCombosToRecalculate($CDFILIAL, $NRMESA, $positionToClose['positionNumber'], $CDLOJA);

                if (!$return['error']) {
                    $combosToRecalulate = $return['data'];
                    $combosData = $this->recalculateCombos($combosToRecalulate, $CDFILIAL, $CDLOJA, $CDCLIENTE, $CDCONSUMIDOR, true);
                }

                $positionNumber = $positionToClose['positionNumber'];

                if (isset($positionToClose['positionCPF']))
                    $positionCPF  = $positionToClose['positionCPF'];
                else
                    $positionCPF  = null;

                $position       = $positionNumber;
                $payments       = $positionToClose['payments'];
                $paymentsValue  = $this->util->removeDecimal($positionToClose['paymentsValue']);

                /*
                 * Altera o valor de positionNumber para T 
                 *      --> O status da mesa é alterado posteriormente
                 */
                if (empty($positionNumber)) {
                    $position = 'T';
                } else {
                    $lastPosition = $this->util->getLastPosition($CDFILIAL, $NRVENDAREST, $NRCOMANDA, $positionNumber);
                    if ($lastPosition) {
                        $position = 'T';
                    }
                }

                if ($position == 'T')
                    $this->util->updateVendaRestConsumer($CDFILIAL, $NRVENDAREST, $CDCLIENTE, $CDCONSUMIDOR);

                foreach ($payments as $payment) {
                    $paymentCode    = $payment['paymentCode'];
                    $paymentValue   = $payment['value'];

                    /*
                     * Gera NRSEQMOVMOB utiilizando um novo código
                     */

                    $NRSEQMOVMOB = $this->orderService->generateMovcaixamobSeq($CDFILIAL);

                    $NRLUGARMESA    = $position;
                    $CDTIPORECE     = $paymentCode;
                    $VRMOV          = $paymentValue;
                    $IDTIPMOV       = 'E';
                    $DSBANDEIRA     = null;
                    $IDTPTEF        = null;
                    $NRSEQVENDA     = null;
                    $IDUTILTXSERV   = $saveServiceTax ? 'S' : 'N';

                    /*
                     * Insere o registro na tabela MOVCAIXAMOB
                     */
                    $responseInsertion = $this->orderService->insertMovcaixamob($NRSEQMOVMOB, $CDFILIAL, $CDCAIXA, $CDVENDEDOR, $NRVENDAREST, $NRCOMANDA, $NRSEQVENDA, $NRMESA, $NRLUGARMESA, $CDTIPORECE, $IDTIPMOV, $VRMOV, $DSBANDEIRA, $IDTPTEF, $IDUTILTXSERV, $CDCLIENTE, $CDCONSUMIDOR);
                    if ($responseInsertion['error']) {
                        throw new \Exception($responseInsertion['message']);
                    }
                }

                /*
                 * Alterar o status da mesa para 'Conta Solicitada'
                 * $IDSTMESAAUX: 'S'
                */
                if ($position ==  'T') {
                    $IDSTMESAAUX = self::PAID;
                    $responseStatusMesa = $this->orderService->changeTableStatus($IDSTMESAAUX, $CDFILIAL, $CDLOJA, $NRMESA);
                    if ($responseStatusMesa['error']) {
                        throw new \Exception($responseStatusMesa['message']);
                    }
                } else {
                    $IDSTLUGARMESA = self::PAID;
                    $responseStatusPositions = $this->orderService->changePositionStatus($IDSTLUGARMESA, $CDFILIAL, $NRVENDAREST, $NRCOMANDA, $positionNumber);
                    if ($responseStatusPositions['error']) {
                        throw new \Exception($responseStatusPositions['message']);
                    }
                }

                $responseCpfPosition = $this->orderService->changePositionCpf($CDFILIAL, $NRVENDAREST, $NRCOMANDA, $positionNumber, $positionCPF);
                if ($responseCpfPosition['error']) {
                    throw new \Exception($responseCpfPosition['message']);
                }
                $serviceTax = 0;

                if ($saveServiceTax) {
                    $serviceTax = $this->setServiceTax($storeServiceTax, $paymentsValue);
                    /*$usesServiceTax   = $storeServiceTax['IDCOMISVENDA'] == 'S';
                    $servTaxAsProduct = $storeServiceTax['IDTRATTAXASERV'] == 'P';
                    $servTaxProduct   = $storeServiceTax['CDPRODTAXASERV'];
                    $servTaxValue     = $storeServiceTax['VRCOMISVENDA'];
                    if(($usesServiceTax) && (!empty($servTaxValue)) && (($servTaxAsProduct && !empty($servTaxProduct)) || (!$servTaxAsProduct))){
                        $serviceTax   = $paymentsValue/$servTaxValue;
                    }*/
                }

                if ($position == 'T') {
                    $this->util->updateBillServiceTax($CDFILIAL, $NRVENDAREST, $NRCOMANDA, $IDUTILTXSERV, $this->util->float($serviceTax));
                }
            }

            $this->util->commit();
        } catch (\Exception $e) {
            $this->util->rollBack();
            $message = 'Erro ao pagar posições. Erro: ' . $e->getMessage();
            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
        }
    }

    public function lockTable(Request $request, Response $response)
    {
        $inputJSON = file_get_contents('php://input');
        $this->util->saveLog($inputJSON, 'Order', 'lockTable');
        $input = json_decode($inputJSON, true);
        $this->util->beginTransaction();

        try {
            if (!isset($input)) {
                throw new \Exception('Não foram recebidos parâmetros');
            }

            $data   = array(); // dados a serem retornados

            $params    = $input;
            $checkinId = $params['checkinId'];
            $lockDate  = date('d/m/Y H:i:s');

            $this->changeTableStatus($checkinId, self::REQUEST_ACCOUNT, $lockDate);
            $this->util->commit();
        } catch (\Exception $e) {
            $this->util->rollBack();
            $message = 'Erro ao alterar situação da mesa. Erro: ' . $e->getMessage();
            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
        }
    }

    public function lockPositions(Request $request, Response $response)
    {
        $inputJSON = file_get_contents('php://input');
        $this->util->saveLog($inputJSON, 'Order', 'lockPositions');
        $input = json_decode($inputJSON, true);

        $this->util->beginTransaction();
        try {
            if (!isset($input)) {
                throw new \Exception('Não foram recebidos parâmetros');
            }

            $data   = array(); // dados a serem retornados

            $params    = $input;
            $checkinId = $params['checkinId'];
            $positions = $params['positions'];
            $status    = self::REQUEST_ACCOUNT;
            $lockDate  = date('d/m/Y H:i:s');
            $this->changePositionStatus($checkinId, $positions, $status, $lockDate);
            $this->util->commit();
        } catch (\Exception $e) {
            $this->util->rollBack();
            $message = 'Erro ao alterar situação das posicoes: ' . $e->getMessage();
            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
        }
    }

    public function unlockTable(Request $request, Response $response)
    {
        $inputJSON = file_get_contents('php://input');
        $this->util->saveLog($inputJSON, 'Order', 'unlockTable');
        $input = json_decode($inputJSON, true);
        $this->util->beginTransaction();

        try {
            if (!isset($input)) {
                throw new \Exception('Não foram recebidos parâmetros');
            }

            $data   = array(); // dados a serem retornados

            $params    = $input;
            $checkinId = $params['checkinId'];

            $this->changeTableStatus($checkinId, self::BUSY);
            $this->util->commit();
        } catch (\Exception $e) {
            $this->util->rollBack();
            $message = 'Erro ao alterar situação da mesa. Erro: ' . $e->getMessage();
            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
        }
    }

    public function unlockPositions(Request $request, Response $response)
    {
        $inputJSON = file_get_contents('php://input');
        $this->util->saveLog($inputJSON, 'Order', 'unlockPositions');
        $input = json_decode($inputJSON, true);

        $this->util->beginTransaction();

        try {
            if (!isset($input)) {
                throw new \Exception('Não foram recebidos parâmetros');
            }

            $data   = array(); // dados a serem retornados

            $params    = $input;
            $checkinId = $params['checkinId'];
            $positions = $params['positions'];
            $status    = self::BUSY;

            $this->changePositionStatus($checkinId, $positions, $status);
            $this->util->commit();
        } catch (\Exception $e) {
            $this->util->rollBack();
            $message = 'Erro ao alterar situação da mesa. Erro: ' . $e->getMessage();
            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
        }
    }

    public function changeTableStatus($checkinId, $status, $date = null)
    {
        /*
         * Verifica se o checkinId é válido
        */
        $validCheckinId = $this->orderService->isValidCheckinId($checkinId);
        if ($validCheckinId['error']) {
            throw new \Exception($validCheckinId['message']);
        }

        /*
         * Obtém CDFILIAL E NRVENDAREST partindo de um IDCHECKIN
         */
        $resultCheckinMesPos = $this->getTableInfoByCheckin($checkinId);
        /*$resultCheckinMesPos = $this->orderService->getResultCheckinMesPos($checkinId);
        if($resultCheckinMesPos['error']){
            throw new \Exception($resultCheckinMesPos['message']);
        }

        $resultCheckinMesPos = $resultCheckinMesPos['data'];*/

        $tableNumber    = $resultCheckinMesPos['NRMESA'];
        $vendarestSeq   = $resultCheckinMesPos['NRVENDAREST'];
        $comandavenSeq  = $resultCheckinMesPos['NRCOMANDA'];
        $subsidiaryCode = $resultCheckinMesPos['CDFILIAL'];
        $storeCode      = $resultCheckinMesPos['CDLOJA'];

        $tableStatus    = $this->util->getTableStatus($subsidiaryCode, $storeCode, $tableNumber);
        if (empty($tableStatus)) {
            throw new \Exception('Mesa não encontrada');
        } else if ($tableStatus !== $status) {
            $responseStatusMesa = $this->orderService->changeTableStatus($status, $subsidiaryCode, $storeCode, $tableNumber, $date);
            if ($responseStatusMesa['error']) {
                throw new \Exception($responseStatusMesa['message']);
            }
        }
    }

    public function changePositionStatus($checkinId, $positions, $status, $date = null)
    {
        /*
         * Verifica se o checkinId é válido
        */
        $validCheckinId = $this->orderService->isValidCheckinId($checkinId);
        if ($validCheckinId['error']) {
            throw new \Exception($validCheckinId['message']);
        }

        /*
         * Obtém CDFILIAL E NRVENDAREST partindo de um IDCHECKIN
        */
        $resultCheckinMesPos = $this->getTableInfoByCheckin($checkinId);
        /*$resultCheckinMesPos = $this->orderService->getResultCheckinMesPos($checkinId);
        if($resultCheckinMesPos['error']){
            throw new \Exception($resultCheckinMesPos['message']);
        }

        $resultCheckinMesPos = $resultCheckinMesPos['data'];*/

        $tableNumber    = $resultCheckinMesPos['NRMESA'];
        $vendarestSeq   = $resultCheckinMesPos['NRVENDAREST'];
        $comandavenSeq  = $resultCheckinMesPos['NRCOMANDA'];
        $subsidiaryCode = $resultCheckinMesPos['CDFILIAL'];
        $storeCode      = $resultCheckinMesPos['CDLOJA'];

        $tableStatus    = $this->util->getTableStatus($subsidiaryCode, $storeCode, $tableNumber);

        if (empty($tableStatus)) {
            throw new \Exception('Mesa não encontrada');
        } else if ($tableStatus == self::REQUEST_ACCOUNT) {
            throw new \Exception('Conta solicitada para a mesa');
        } else {
            foreach ($positions as $position) {
                $positionNumber = $position['positionNumber'];
                $positionStatus = $this->util->getPositionStatus($subsidiaryCode, $vendarestSeq, $comandavenSeq, $positionNumber);

                if (empty($positionStatus)) {
                    throw new \Exception('Posicao não encontrada');
                } else if ($positionStatus !== $status) {
                    $responseStatusMesa = $this->orderService->changePositionStatus($status, $subsidiaryCode, $vendarestSeq, $comandavenSeq, $positionNumber, $date);
                    if ($responseStatusMesa['error']) {
                        throw new \Exception($responseStatusMesa['message']);
                    }
                }
            }
        }
    }

    public function validateAddress(Request $request, Response $response)
    {
        $inputJSON = file_get_contents('php://input');
        $input = json_decode($inputJSON, true);

        try {
            if (!isset($input)) {
                throw new \Exception('Não foram recebidos parâmetros');
            }

            $data   = array(); // dados a serem retornados
            //@todo alterar getCityCode para usar o país também 
            $params        = $input;
            $country       = '0055';
            $stateCode     = $params['state'];
            $cityName      = $params['cityName'];
            $neighborhood  = $params['neighborhood'];
            $streetAddress = $params['streetAddress'];
            $number        = $params['number'];
            $locationLat   = $params['lat'];
            $locationLng   = $params['lng'];
            $responseObject = array(
                'isAttended'   => false,
                'subsidiaries' => array()
            );

            if (!empty($locationLat) && !empty($locationLng)) {
                $locationCoords = array($locationLat, $locationLng);
            } else {
                $address = $streetAddress . ', ' . $number;

                if (!empty($neighborhood)) {
                    $address   = $address . ', ' . $neighborhood;
                }

                if (!empty($cityName)) {
                    $address   = $address . ', ' . $cityName;
                }

                if (!empty($stateCode)) {
                    $address   = $address . ', ' . $stateCode;
                }

                $addressUrl = urlencode($address);
                $url        = $this->geocodeURL . $addressUrl;
                $respJSON   = file_get_contents($url);
                $resp       = json_decode($respJSON, true);

                if ($resp['status'] == 'OK') {
                    $location       = $resp['results'][0]['geometry']['location'];
                    $locationLat    = $location['lat'];
                    $locationLng    = $location['lng'];
                    $locationCoords = array($locationLat, $locationLng);
                } else {
                    throw new \Exception("Não foi possível encontrar as coordenadas do endereço", self::ERROR_MESSAGE);
                }
            }

            $deliveryAreasResp = $this->orderService->getDeliveryAreas();

            if (!$deliveryAreasResp['error']) {
                $deliveryAreas = $deliveryAreasResp['data'];

                foreach ($deliveryAreas as $area) {
                    $subsidiaryCode      = $area['CDFILIAL'];
                    $deliveryFee         = $area['VRTXENTREGA'];
                    $priority            = $area['NRORDATEND'];
                    $areaCode            = $area['CDAREA'];
                    $stringCoordinates   = $area['COORDAREA'];
                    $stringCoordinates   = str_replace("['", "", $stringCoordinates);
                    $stringCoordinates   = str_replace("']", "", $stringCoordinates);
                    $stringCoordinates   = str_replace("','", '$', $stringCoordinates);
                    $polyCoordinates     = array();
                    $explodedCoordinates = explode('$', $stringCoordinates);

                    foreach ($explodedCoordinates as $latLongStr) {
                        $latLong = explode(',', $latLongStr);
                        array_push($polyCoordinates, $latLong);
                    }

                    $inside = $this->inside($locationCoords, $polyCoordinates);

                    if ($inside) {
                        $responseObject['isAttended'] = true;
                        $subsidiaryObject = array(
                            'subsidiaryCode' => $subsidiaryCode,
                            'deliveryFee'    => $deliveryFee,
                            'priority'       => $priority,
                            'areaCode'       => $areaCode
                        );
                        array_push($responseObject['subsidiaries'], $subsidiaryObject);
                    }
                }
            } else {
                throw new \Exception("Não foi possível buscar as áreas de atendimento", self::ERROR_MESSAGE);
            }

            array_push($data, $responseObject);
            $response->addDataSet(new DataSet('', $data));
        } catch (\Exception $e) {
            $message = 'Erro ao validar endereço. Erro: ' . $e->getMessage();
            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
        }
    }

    public function inside($point, $vs)
    {
        $x = $point[0];
        $y = $point[1];

        $inside = false;
        for ($i = 0, $j = count($vs) - 1; $i < count($vs); $j = $i++) {
            $xi = (float)$vs[$i][0];
            $yi = (float)$vs[$i][1];
            $xj = (float)$vs[$j][0];
            $yj = (float)$vs[$j][1];

            $intersect = (($yi > $y) != ($yj > $y)) && ($x < ($xj - $xi) * ($y - $yi) / ($yj - $yi) + $xi);
            if ($intersect)
                $inside = !$inside;
        }
        return $inside;
    }

    public function getDeliveryDetails(Request $request, Response $response)
    {
        $inputJSON = file_get_contents('php://input');
        $input = json_decode($inputJSON, true);

        try {
            if (!isset($input)) {
                throw new \Exception('Não foram recebidos parâmetros');
            }

            $data   = array(); // dados a serem retornados

            $params         = $input;
            $areaCode       = $params['areaCode'];
            $subsidiaryCode = $params['subsidiaryCode'];
            $storeCode      = $params['storeCode'];

            $storeDeliveryTime = $this->util->getStoreDeliveryTime($storeCode, $subsidiaryCode);
            $deliveryFee       = $this->util->getDeliveryFee($areaCode, $subsidiaryCode);
            $data = array(
                array(
                    'deliveryFee'  => $this->util->removeDecimal($deliveryFee),
                    'deliveryTime' => $storeDeliveryTime
                )
            );
            $response->addDataSet(new DataSet('', $data));
        } catch (\Exception $e) {
            $message = 'Erro ao buscar detalhes de entrega da loja. Erro: ' . $e->getMessage();
            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
        }
    }

    public function recalculateItensPrice(Request $request, Response $response)
    {
        $inputJSON = file_get_contents('php://input');
        $input     = json_decode($inputJSON, true);

        try {
            if (!isset($input)) {
                throw new \Exception('Não foram recebidos parâmetros');
            }

            $params = $input;

            $subsidiaryCode = $params['subsidiaryCode'];
            $clientCode     = $params['clientCode'];
            $consumerCode   = $params['consumerCode'];
            $checkinId      = $params['checkinId'];

            $validCheckinId = $this->orderService->isValidCheckinId($checkinId);
            if ($validCheckinId['error']) {
                throw new \Exception($validCheckinId['message']);
            }
            $resultCheckinMesPos = $this->getTableInfoByCheckin($checkinId);

            $storeCode   = $resultCheckinMesPos['CDLOJA'];
            $tableNumber = $resultCheckinMesPos['NRMESA'];
            $nrvendarest = $resultCheckinMesPos['NRVENDAREST'];
            $positionNumber = $this->util->getTablePositions($subsidiaryCode, $nrvendarest);

            $data = array();

            if (!$this->util->consumerExists($clientCode, $consumerCode)) {
                throw new \Exception('Consumidor inexistente.');
            }

            foreach ($positionNumber as $position) {
                $positionData = array(
                    'subsidiaryCode' => $subsidiaryCode,
                    'clientCode'     => $clientCode,
                    'consumerCode'   => $consumerCode,
                    'positionNumber' => $position,
                    'storeCode'      => $storeCode,
                    'tableNumber'    => $tableNumber,
                    'items'          => array()
                );

                if (isset($params['items']) && !empty($params['items']))
                    $return = $this->recalculateItemsPrice($params['items'], $subsidiaryCode, $clientCode, $consumerCode, $storeCode);

                $return = $this->getProductsToRecalculate($subsidiaryCode, $tableNumber, $position, $storeCode);
                if (!$return['error']) {
                    $productsToRecalculate = $return['data'];
                    $productsData = $this->recalculateProductsPrice($productsToRecalculate, $subsidiaryCode, $tableNumber, $position, $storeCode, $clientCode, $consumerCode, true);
                    $positionData['items'] = $productsData;
                }

                $return = $this->getCombosToRecalculate($subsidiaryCode, $tableNumber, $position, $storeCode);
                if (!$return['error']) {
                    $combosToRecalulate = $return['data'];
                    $combosData = $this->recalculateCombos($combosToRecalulate, $subsidiaryCode, $storeCode, $clientCode, $consumerCode, true);
                    $positionData['items'] = array_merge($positionData['items'], $combosData);
                }

                array_push($data, $positionData);
            }

            $response->addDataSet(new DataSet('', $data));
        } catch (\Exception $e) {
            $message = 'Erro ao realizar recálculo dos itens do pedido. Erro: ' . $e->getMessage();
            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
        }
    }

    private function recalculateItemsPrice($items, $subsidiaryCode, $clientCode, $consumerCode, $storeCode)
    {
        foreach ($items as &$item) {
            $this->recalculateItemPrice($item, $subsidiaryCode, $clientCode, $consumerCode, $storeCode);
        }
    }

    private function recalculateItemPrice($item, $subsidiaryCode, $clientCode, $consumerCode, $storeCode)
    {

        if (!empty($item['childItems'])) {
            $item['unitValue'] = 0;
            $item['discount']  = 0;

            foreach ($item['childItems'] as $childItem) {
                $priceInfo = $this->recalculateItemPrice($childItem, $subsidiaryCode, $clientCode, $consumerCode, $storeCode);
                $item['unitValue'] += $priceInfo['unitValue'];
                $item['discount']  += $priceInfo['discount'];
            }
        } else {
            $priceDetails = $this->getConsumerItemPrice($subsidiaryCode, $clientCode, $item['productCode'], $storeCode, $consumerCode);
            $discount     = $this->getNewDiscount($priceDetails, $item);
            $newUnitPrice = !empty($priceDetails['newPrice']) ? $priceDetails['newPrice'] : $item['VRPRECCOMVEN'];

            $this->updateItemPrice($item['CDFILIAL'], $item['NRVENDAREST'], $item['NRCOMANDA'], $item['NRPRODCOMVEN'], $this->util->removeDecimal($newUnitPrice), $discount);

            $item['unitValue'] = $newUnitPrice;
            $item['discount']  = $discount;
            return array(
                'unitValue' => $newUnitPrice,
                'discount'  => $discount
            );
        }
    }

    private function getProductsToRecalculate($subsidiaryCode, $tableNumber, $positionNumber, $storeCode)
    {
        return $this->orderService->getProductsToRecalculate($subsidiaryCode, $tableNumber, $positionNumber, $storeCode);
    }

    private function recalculateProductsPrice($productsToRecalculate, $subsidiaryCode, $tableNumber, $positionNumber, $storeCode, $clientCode, $consumerCode, $saveRecalculatedItem = true)
    {
        $itemsData = array();

        foreach ($productsToRecalculate as $item) {

            $productCode  = $item['CDPRODUTO'];
            $priceDetails = $this->getConsumerItemPrice($subsidiaryCode, $clientCode, $productCode, $storeCode, $consumerCode);

            $discount = $this->getNewDiscount($priceDetails, $item);
            $newUnitPrice = !empty($priceDetails['newPrice']) ? $priceDetails['newPrice'] : $item['VRPRECCOMVEN'];

            $itemData = array(
                'productCode'  => $productCode,
                'productName'  => $item['NMPRODUTO'],
                'quantity'     => $this->util->float($item['QTPRODCOMVEN']),
                'unitValue'    => $this->util->removeDecimal($newUnitPrice),
                'discount'     => $this->util->removeDecimal($discount),
                'observations' => array()
            );

            if ($saveRecalculatedItem) {
                $this->updateItemPrice($item['CDFILIAL'], $item['NRVENDAREST'], $item['NRCOMANDA'], $item['NRPRODCOMVEN'], $this->util->removeDecimal($newUnitPrice), $this->util->removeDecimal($discount));
            }

            $return = $this->getObservations($item['CDFILIAL'], $item['NRVENDAREST'], $item['NRCOMANDA'], $item['NRPRODCOMVEN']);

            if (!$return['error'])
                $itemData['observations'] = $comboProductData['observations'] = $this->formatObservations($return['data']);

            array_push($itemsData, $itemData);
        }

        return $itemsData;
    }

    private function getConsumerItemPrice($subsidiaryCode, $clientCode, $productCode, $storeCode, $consumerCode)
    {

        $return = array(
            'newPrice'      => 0,
            'discountType'  => '',
            'discountValue' => 0
        );

        $priceDetails = $this->orderService->recalculateItensPrice($subsidiaryCode, $clientCode, $productCode, $storeCode, $consumerCode);
        if (!$priceDetails['error']) {
            $return['newPrice'] = ($priceDetails['PRECO'] + $priceDetails['ACRE']) - $priceDetails['DESC'];
        }

        return $return;
    }

    private function getObservations($CDFILIAL, $NRVENDAREST, $NRCOMANDA, $NRPRODCOMVEN)
    {
        return $this->orderService->getObservations($CDFILIAL, $NRVENDAREST, $NRCOMANDA, $NRPRODCOMVEN);
    }

    private function updateItemPrice($CDFILIAL, $NRVENDAREST, $NRCOMANDA, $NRPRODCOMVEN, $VRPRECCOMVEN, $VRDESCCOMVEN)
    {
        return $this->orderService->updateItemPrice($CDFILIAL, $NRVENDAREST, $NRCOMANDA, $NRPRODCOMVEN, $VRPRECCOMVEN, $VRDESCCOMVEN);
    }

    private function getCombosToRecalculate($subsidiaryCode, $tableNumber, $positionNumber, $storeCode)
    {
        return $this->orderService->getCombosToRecalculate($subsidiaryCode, $tableNumber, $positionNumber, $storeCode);
    }

    private function recalculateCombos($combosToRecalulate, $subsidiaryCode, $storeCode, $clientCode, $consumerCode, $saveRecalculatedItem = true)
    {
        $recalculatedCombos = array();

        foreach ($combosToRecalulate as $combo) {
            $comboData = array(
                'productCode'   => $combo['CDPRODPROMOCAO'],
                'productName'   => $combo['NMPRODUTO'],
                'quantity'      => 1,
                'unitValue'     => 0,
                'discount'      => 0,
                'observations'  => array(),
                'childItems'    => array()
            );

            $return = $this->getComboProductsToRecalculate($combo['CDPRODPROMOCAO'], $combo['CDFILIAL'], $combo['NRVENDAREST'], $combo['NRCOMANDA'], $combo['NRSEQPRODCOM']);
            if (!$return['error']) {
                $comboProductsToRecalculate = $return['data'];
                $comboProducts = $this->recalculateComboProducts($comboProductsToRecalculate, $subsidiaryCode, $storeCode, $clientCode, $consumerCode, $combo['NRCOMANDA'], $combo['NRVENDAREST'], $saveRecalculatedItem);
                $comboData['childItems'] = $comboProducts;
            }

            foreach ($comboData['childItems'] as $childItem) {
                $comboData['unitValue'] += $childItem['unitValue'];
                $comboData['discount']  += $childItem['discount'];
            }

            array_push($recalculatedCombos, $comboData);
        }

        return $recalculatedCombos;
    }

    private function getComboProductsToRecalculate($CDPRODPROMOCAO, $CDFILIAL, $NRVENDAREST, $NRCOMANDA, $NRSEQPRODCOM)
    {
        return $this->orderService->getComboProductsToRecalculate($CDPRODPROMOCAO, $CDFILIAL, $NRVENDAREST, $NRCOMANDA, $NRSEQPRODCOM);
    }

    private function numberFormatPrecision($number, $precision = 2, $separator = '.')
    {
        $numberParts = explode($separator, $number);
        $response    = $numberParts[0];

        if (count($numberParts) > 1) {
            $response .= $separator;
            $response .= substr($numberParts[1], 0, $precision);
        }
        return $response;
    }

    private function getNewDiscount($priceDetails, $item)
    {

        $discountType  = $priceDetails['discountType'];
        $discountValue = $priceDetails['discountValue'];

        $itemDiscount = isset($item['VRDESCCOMVEN']) ? $item['VRDESCCOMVEN'] : $item['discount'];

        if ($discountType == 'V')
            $discount = $itemDiscount + $itemDiscount - $discountValue;
        else
            $discount = $itemDiscount - ($itemDiscount * ($discountValue / 100));

        return floatval($this->numberFormatPrecision($discount));
    }

    private function recalculateComboProducts($comboProducts, $subsidiaryCode, $storeCode, $clientCode, $consumerCode, $nrcomanda, $nrvendarest, $saveRecalculatedItem)
    {
        $comboData = array();

        foreach ($comboProducts as $comboProduct) {

            $productCode     = $comboProduct['CDPRODUTO'];
            $priceDetails    = $this->getConsumerItemPrice($subsidiaryCode, $clientCode, $productCode, $storeCode, $consumerCode);
            $discount        = $this->getNewDiscount($priceDetails, $comboProduct);
            $newUnitPrice    = $priceDetails['newPrice'];
            $itcomandavenSeq = $comboProduct['NRPRODCOMVEN'];
            $newUnitPrice    = !empty($newUnitPrice) ? $newUnitPrice : $comboProduct['VRPRECCOMVEN'];

            $comboProductData = array(
                'productCode'   => $comboProduct['CDPRODUTO'],
                'productName'   => $comboProduct['NMPRODUTO'],
                'quantity'      => $this->util->float($comboProduct['QTPRODCOMVEN']),
                'unitValue'     => $this->util->removeDecimal($newUnitPrice),
                'discount'      => $this->util->removeDecimal($discount),
                'observations'  => array()
            );

            if ($saveRecalculatedItem) {
                $this->updateItemPrice($comboProduct['CDFILIAL'], $comboProduct['NRVENDAREST'], $comboProduct['NRCOMANDA'], $comboProduct['NRPRODCOMVEN'], $this->util->removeDecimal($newUnitPrice), $this->util->removeDecimal($discount));
            }

            $return = $this->getObservations($subsidiaryCode, $nrvendarest, $nrcomanda, $itcomandavenSeq);

            if (!$return['error']) {
                $comboProductData['observations'] = $this->formatObservations($return['data']);
            }

            array_push($comboData, $comboProductData);
        }

        return $comboData;
    }

    private function formatObservations($observations)
    {

        $observationsData = array();
        foreach ($observations as $observation) {

            $observationData = array(
                'obsGroupCode' => $observation['CDGRPOCOR'],
                'obsCode'      => $observation['CDOCORR'],
                'obsName'      => $observation['DSOCORR'],
                'quantity'     => 1,
                'price'        => 0
            );

            array_push($observationsData, $observationData);
        }
        return $observationsData;
    }

    public function printOrder($subsidiaryCode, $comandavenSeq, $storeCode, $items, $sellerCode, $standardTable, $pagerId, $modalidade ,$toGoString, $taaOrder)
    {
        $this->orderService->printOrder($subsidiaryCode, $storeCode, $items, $sellerCode, null, $standardTable, $pagerId, $modalidade, $toGoString, false, true, $comandavenSeq, $taaOrder);
    }

    private function formatOrderToPrint($items)
    {
        $dataToPrint = array();

        foreach ($items as $item) {
            $obsData = array();
            $itemData = $this->printItemPattern();

            if ($item['childItems']) {
                $itemData['NMPRODPROMOCAO'] = $item['productName'];
                $itemData['IDTIPOCOMPPROD'] = '3';

                foreach ($item['childItems'] as $childItem) {
                    if (!empty($childItem['observations'])) {
                        $obsData = array_merge($obsData, $this->returnTxProdComVen($childItem['observations']));
                    }
                }
            }

            if (!empty($item['observations'])) {
                $obsData = array_merge($obsData, $this->returnTxProdComVen($item['observations']));
            }

            if (!empty($obsData)) {
                $itemData['TXPRODCOMVEN'] = '-> ' . implode(PHP_EOL . '   -> ', $obsData);
            }

            $itemData['CDPRODUTO']    = $item['productCode'];
            $itemData['NMPRODUTO']    = $item['productName'];
            $itemData['QTPRODCOMVEN'] = $item['quantity'];

            array_push($dataToPrint, $itemData);
        }

        return $dataToPrint;
    }

    private function printItemPattern()
    {
        return array(
            "CDPRODUTO"      => "",
            "NMPRODUTO"      => "",
            "IDTIPOCOMPPROD" => "",
            "QTPRODCOMVEN"   => 0,
            "TXPRODCOMVEN"   => "",
            "IDIMPPROMOCAO"  => "N",
            "ORDEMIMP"       => null,
            "ATRASOPROD"     => null,
            "ORDEM"          => null,
            "ORDEMIMP"       => null,
            "NMPRODPROMOCAO" => null
        );
    }

    private function returnTxProdComVen($observations)
    {
        $obsData = array();

        foreach ($observations as $observation) {
            array_push($obsData, $observation['obsName']);
        }

        return $obsData;
    }

    private function getTableInfoByCheckin($checkinId)
    {
        $resultCheckinMesPos = $this->orderService->getResultCheckinMesPos($checkinId);

        if ($resultCheckinMesPos['error']) {
            throw new \Exception($resultCheckinMesPos['message']);
        }

        $resultCheckinMesPos = $resultCheckinMesPos['data'];

        return array(
            'CDFILIAL'      => $resultCheckinMesPos['CDFILIAL'],
            'NRVENDAREST'   => $resultCheckinMesPos['NRVENDAREST'],
            'NRPOSICAO'     => $resultCheckinMesPos['NRPOSICAO'],
            'NRCOMANDA'     => $resultCheckinMesPos['NRCOMANDA'],
            'NRMESA'        => $resultCheckinMesPos['NRMESA'],
            'CDLOJA'        => $resultCheckinMesPos['CDLOJA'],
            'CDCLIENTE'     => $resultCheckinMesPos['CDCLIENTE'],
            'CDCONSUMIDOR'  => $resultCheckinMesPos['CDCONSUMIDOR']
        );
    }

    private function verifyDailyCreditLimit($paymentValue, $clientCode, $consumerCode, $subsidiaryCode)
    {
        //$cdfamilisald = '001';
        $cdfamilisald = $this->orderService->getFamiliaSald($subsidiaryCode);

        $maxDailyBuy = $this->orderService->getConsumerMaxDaily($clientCode, $consumerCode);
        if(count($cdfamilisald) > 1){
            $balance = 0;
            foreach($cdfamilisald as $fmlsald){
                $balance = ($this->util->getConsumerBalance($clientCode, $consumerCode, $fmlsald['CDFAMILISALD'], 7, $subsidiaryCode) + $balance); 
            }

        }else{
            $balance = $this->util->getConsumerBalance($clientCode, $consumerCode, $cdfamilisald, 7, $subsidiaryCode); //calls extratocons
        }

        $todaysMovement   = $this->util->getTodaysExpenditure($clientCode, $consumerCode, 7, $subsidiaryCode); //calls extratocons

        $blockLowBalance = isset($maxDailyBuy['VRLIMCREDCONS']) ? $maxDailyBuy['VRLIMCREDCONS'] : 0;
        $blockLowBalance = $this->util->removeDecimal($blockLowBalance);

        $maxDailyBuyValue = isset($maxDailyBuy['VRMAXCREDCONS']) ? $maxDailyBuy['VRMAXCREDCONS'] : 0;
        $maxDailyBuyValue = $this->util->removeDecimal($maxDailyBuyValue); 

        if(isset($maxDailyBuy['VRLIMCREDCONS']) && ($balance - $paymentValue) < $blockLowBalance)
            return $this->returnPattern(true, 'O valor do pagamento é menor que o limite de bloqueio de lançamentos.');
        else if (!isset($maxDailyBuy['VRLIMCREDCONS']) && $paymentValue > $balance)
            return $this->returnPattern(true, 'O valor do pagamento é maior que o saldo disponível.');
        else if (($maxDailyBuyValue != 0) && ($todaysMovement + $paymentValue > $maxDailyBuyValue))
            return $this->returnPattern(true, 'O valor do pagamento é maior que o limite diário estabelecido.');
        else if (!isset($maxDailyBuy['VRLIMCREDCONS']) && ($balance < $paymentValue))
            return $this->returnPattern(true, 'O valor do pagamento é maior que o saldo atual.');
        else
            return $this->returnPattern(false, '');
    }

    private function returnPattern($error, $message)
    {
        return array(
            'error'   => $error,
            'message' => $message
        );
    }

    public function getOrderStatus(Request $request, Response $response)
    {
        $inputJSON  = file_get_contents('php://input');
        $input      = json_decode($inputJSON, true);

        try {

            if (empty($input))
                throw new \Exception('Não foram recebidos parâmetros');

            $params = $input;

            $subsidiaryCode = $params['subsidiaryCode'];
            $saleCode       = $params['saleCode'];
            $saleId         = $params['saleId'];

            $orderInfo = $this->util->getOrderStatus($subsidiaryCode, $saleCode, $saleId);

            if (empty($orderInfo))
                throw new \Exception("Pedido não encontrado.");

            $orderStatus = $this->getStatus($orderInfo);

            $data = array(
                array(
                    'orderStatus' => $orderStatus
                )
            );

            $response->addDataSet(new DataSet('', $data));
        } catch (\Exception $e) {
            $message = 'Erro ao recuperar status do pedido. Erro: ' . $e->getMessage();
            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
        }
    }

    private function getStatus($orderInfo)
    {
        $orderStatus = '2';
        if ($orderInfo['IDSTCOMANDA'] == '4')     $orderStatus = 'C';
        elseif ($orderInfo['IDSINCMSDE'] == 'N')  $orderStatus = '0';
        elseif (!empty($orderInfo['DTCHEGACMD'])) $orderStatus = '4';
        elseif (!empty($orderInfo['DTSAIDACMD'])) $orderStatus = '3';

        return $orderStatus;
    }

    public function setOrderArrival(Request $request, Response $response)
    {
        $inputJSON  = file_get_contents('php://input');
        $input      = json_decode($inputJSON, true);

        $this->util->beginTransaction();
        try {

            if (empty($input))
                throw new \Exception('Não foram recebidos parâmetros');

            $params = $input;

            $subsidiaryCode = $params['subsidiaryCode'];
            $saleCode       = $params['saleCode'];
            $saleId         = $params['saleId'];
            $arrivalTime    = !empty($params['arrivalTime']) ? $params['arrivalTime'] : date('d/m/Y H:i:s', time());;

            $retorno = $this->util->setOrderArrival($subsidiaryCode, $saleCode, $saleId, $arrivalTime);

            if (empty($retorno))
                throw new \Exception("Não foi possível encontrar o pedido.");

            $this->util->commit();

            $response->addMessage(new Message("Horário atualizado com sucesso.", self::GENERAL_MESSAGE));
        } catch (\Exception $e) {
            $this->util->rollBack();
            $message = 'Erro ao atualizar horário de chegada. Erro: ' . $e->getMessage();
            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
        }
    }

    public function getMinTime(Request $request, Response $response)
    {
        $inputJSON  = file_get_contents('php://input');
        $input      = json_decode($inputJSON, true);

        try {
            if (empty($input))
                throw new \Exception('Não foram recebidos parâmetros');

            $params = $input;

            $subsidiaryCode = $params['subsidiaryCode'];
            $storeCode      = $params['storeCode'];

            $time = $this->util->getMinTime($subsidiaryCode, $storeCode);

            $data = array(
                array(
                    'minimumTime' => $time
                )
            );

            $response->addDataSet(new DataSet('', $data));
        } catch (\Exception $e) {
            throw new \Exception("Houve um erro ao recuperar o tempo de produção desta loja.");
        }
    }

    private function validateParams($params)
    {
        $order = $params['order'];
        $data = array();

        if (!isset($order['subsidiaryCode']))
            $data['subsidiaryCode'] = true;

        if (!isset($order['items'])) {
            $data['items'] = true;
        } else {
            $items = $order['items'];

            foreach ($items as $item) {
                $this->validateItems($item);
            }
        }

        if (!empty($data))
            throw new \Exception("Alguns parâmetros obrigatórios estão vazios, verifique os dados.");
    }

    public function savePayment(Request $request, Response $response)
    {

        $inputJSON  = file_get_contents('php://input');
        $input      = json_decode($inputJSON, true);

        $saleCode    = $input['saleCode'];
        $fileName = date("Ymd-H-i-s") . "-" . $saleCode;

        $this->util->saveLogOrder($inputJSON, 'Payment/' . date("Y/m/d"), $fileName);

        $this->util->beginTransaction();

        try {
            $nrcomanda    = $input['saleCode'];
            $cdfilial     = $input['subsidiaryCode'];
            $cashBack     = $input['cashBack'];
            $payments     = $input['payments'];
            $giftCards    = isset($input['giftCards']) ? $input['giftCards'] : array();
            $orderCPF     = isset($input['orderCPF']) ? $input['orderCPF'] : null;
            $consumerName = isset($input['consumerName']) ? $input['consumerName'] : null;

            $orderCPF          = preg_replace( '/[^0-9]/is', '', $orderCPF );
            $order             = $this->util->getNrvendarest($cdfilial, $nrcomanda);
            $deliveryFee       = $this->util->removeDecimal($order['VRACRCOMANDA']);
            $nrvendarest       = $order['NRVENDAREST'];
            $orderPaymentValue = $this->util->getPaymentValue($cdfilial, $nrvendarest);

            if (!empty($orderPaymentValue))
                throw new \Exception("Esta comanda já foi paga!");

            $cdcliente    = $order['CDCLIENTE'];
            $cdconsumidor = $order['CDCONSUMIDOR'];

            if (!$nrvendarest)
                throw new \Exception("Não foi posssível encontrar o pedido.");

            if (empty($payments) && empty($giftCards) && empty($cashBack))
                throw new \Exception("Nenhuma forma de pagamento informada.");

            // Checa se o vale presente está sendo utilizado com outro tipo de recebimento com IDTIPORECE = '9'
            if (!empty($giftCards)) {
                $hasCreditPayment = false;
                foreach ($payments as $payment) {
                    if ($this->util->getPaymentId($payment['paymentCode']) == "9") $hasCreditPayment = true;
                }
                if ($hasCreditPayment == true) throw new \Exception("Não é possível utilizar junto ao vale presente outro método de pagamento com crédito pessoal.");
            }

            if(!empty($orderCPF)) {
                $this->orderService->updateOrderCpf($cdfilial, $nrcomanda, $nrvendarest, $orderCPF, $consumerName);
            }

            if (!empty($payments)) {
                $savePaymentsreturn = $this->savePayments($payments, $cdfilial, $nrvendarest, $cdcliente, $cdconsumidor, 0, true);
                if ($savePaymentsreturn['error']) {
                    $message = $savePaymentsreturn['message'];
                    throw new \Exception($message);
                }
            }

            if (!empty($giftCards)) {
                $totalValue = $this->util->removeDecimal($this->orderService->getItemsValue($cdfilial, $nrvendarest, $nrcomanda));
                $totalValue = $this->util->subtract($this->util->add($totalValue, $deliveryFee), $cashBack);
                $caixa      = $this->util->getCDCAIXA($cdfilial, $nrcomanda, $nrvendarest);
                $cdcaixa    = $caixa['CDCAIXACOLETOR'];
                $DTABERCAIX = $this->util->getPosDate($cdfilial, $cdcaixa);

                $giftCardsResponse = $this->saveGiftCards($giftCards, $cdfilial, $nrvendarest, $cdcaixa, $DTABERCAIX, $totalValue);

                if ($giftCardsResponse['error']) {
                    $message = $giftCardsResponse['message'];
                    throw new \Exception($message);
                }
            }

            $productsValue = 0;
            $items = $this->orderService->getOrderItems($nrvendarest, $nrcomanda, $cdfilial);

            foreach ($items as $item) {
                $productsValue = $this->util->add($productsValue, $item['TOTAL']);
            }

            if (!empty($cashBack)) {
                $this->validateCashBack($cdcliente, $cdconsumidor, $cashBack, $cdfilial, $items[0]["NRINSCRCONS"]);

                $realItemsQuant = count($items);
                $realItemsQuant = $this->util->div($realItemsQuant, 100);

                foreach ($items as &$item) {
                    $item['quantity']  = $item['QTPRODCOMVEN'];
                    $item['unitValue'] = $item['VRPRECCOMVEN'];
                    $item['discount']  = $item['VRDESCCOMVEN'];
                }
                unset($item);

                if ($cashBack < $productsValue) {
                    $orderDiscount = $cashBack;
                } else {
                    $cashToApply = $this->util->subtract($productsValue, $realItemsQuant);
                    $orderDiscount = $cashToApply;
                    $userCashBack = $cashBack;
                    $cashBack = $this->util->add($cashToApply, $realItemsQuant);

                    if ($userCashBack < $this->util->add($cashBack, $deliveryFee)) {
                        $cashBack = $userCashBack;
                    } else {
                        $cashBack = $this->util->add($cashBack, $deliveryFee);
                    }

                    $cdtiporece     = $this->util->getMoney();
                    $orderPayments  = $this->orderService->getOrderPayments($cdfilial, $nrvendarest);

                    if (empty($orderPayments)) {
                        $movcaixadlvSeq = str_pad(1, 10, '0', STR_PAD_LEFT);
                        $paymentValue = $this->util->add($realItemsQuant, $deliveryFee);
                        $this->insertMovCaixaDlv($cdfilial, $nrvendarest, $movcaixadlvSeq, 'E', $cdcliente, $cdconsumidor, $cdtiporece, $paymentValue, null, 1, null);
                    }
                }

                if(!empty($orderPayments)) {
                    $vrPagto = 0;

                    foreach ($orderPayments as $payment) {
                        if($payment['IDTIPOMOVIVEDLV'] === 'E') {
                            $vrPagto = $this->util->add($vrPagto, $payment['VRMOVIVEND'], 2);
                        }
                    }

                    if(((bcsub($productsValue, $orderDiscount, 2)) != $vrPagto) || $realItemsQuant < $vrPagto) {
                        throw new \Exception("Valor informado no pagamento precisa corresponder há no mínimo 0.01 por produto cobrado. Valor Esperado: {$realItemsQuant} | Valor Pago: {$vrPagto}");
                    }

                }

                $totalDiscount = $orderDiscount;
                $sumValue = 0;

                $this->shareDiscountEx($items, $orderDiscount, $productsValue);

                //foreach ($items as &$item) {
                //    $this->shareDiscount($item, $orderDiscount, $productsValue, $remaingValue, $remaingDiscount, false);
                //}

                foreach ($items as $item) {
                    $sumValue += $this->getDiscountValue($item);
                }

                $orderDiscount = $this->util->subtract($totalDiscount, $sumValue);
                unset($item);

                foreach ($items as $item) {
                    $cdproduto = $item['CDPRODUTO'];
                    $nrprodcomven = $item['NRPRODCOMVEN'];
                    $discount = $item['discount'];
                    $this->orderService->updateItemDiscount($nrcomanda, $nrvendarest, $cdfilial, $cdproduto, $nrprodcomven, $discount);
                }

                $this->orderService->saveCashBack($nrcomanda, $nrvendarest, $cdfilial, $cashBack);
            }

            //$this->validatePayments($cdfilial, $nrvendarest, $nrcomanda, $order['VRACRCOMANDA']);

            if (empty($cdconsumidor) && !empty($giftCards)) {
                $this->orderService->updateConsumerVendarest($giftCards[0]['clientCode'], $giftCards[0]['consumerCode'], $nrvendarest, $cdfilial);
            }

            $response->addMessage(new Message("Pagamento efetuado com sucesso ($cdfilial-$nrcomanda).", self::GENERAL_MESSAGE));
            $this->util->commit();
        } catch (\Exception $e) {
            $this->util->rollBack();
            $message = 'Erro ao salvar o pagamento. Erro: ' . $e->getMessage();
            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
        }
    }

    public function cancelOrder(Request $request, Response $response)
    {
        try {
            $inputJSON  = file_get_contents('php://input');
            $input      = json_decode($inputJSON, true);

            $cdfilial   = !empty($input['subsidiaryCode']) ? $input['subsidiaryCode'] : null;
            $cdcaixa    = !empty($input['posCode']) ? $input['posCode'] : null;
            $nrcomanda  = !empty($input['saleCode']) ? $input['saleCode'] : null;
            $nrseqvenda = !empty($input['saleSeq']) ? $input['saleSeq'] : null; // opcional

            if ($this->util->databaseIsOracle())
                throw new \Exception("Operação não permitida nessa versão do projeto!");

            if (empty($cdfilial))
                throw new \Exception("Numero da filial nao informada.");

            if (empty($cdcaixa))
                throw new \Exception("Numero do caixa nao informado.");

            if (empty($nrcomanda))
                throw new \Exception("Numero da comanda nao informada.");

            $controlPath = "{$this->logPath}" . DIRECTORY_SEPARATOR . ".." . DIRECTORY_SEPARATOR . "CANCEL-CTRL-FILE.dat";

            if (!file_exists($controlPath)){
                if($this->util->isRunningOnWindows()) {
                    mkdir($controlPath, 0777, true);
                } else {
                    mkdir($controlPath, 0775, true);
                }
            }

            // if (!file_exists($controlPath))
            //     throw new \Exception("Serviço de cancelamento de pedidos inativo.");

            $controlFile = json_decode(base64_decode(file_get_contents($controlPath)), true);

            if ((time() - $controlFile['expiresIn']) > self::MAX_CONTROL_FILE_UPDATE_TIME)
                throw new \Exception("Serviço de cancelamento de pedidos inativo.");

            $this->createCancelOrder($cdfilial, $cdcaixa, $nrcomanda, $nrseqvenda);

            $response->addMessage(new Message("O pedido ($cdfilial-$nrcomanda), venda (" . ((!empty($nrseqvenda)) ? $nrseqvenda : "não informada") . ") foi incluído para cancelamento.", self::GENERAL_MESSAGE));
        } catch (\Exception $e) {
            $message = 'Erro ao cancelar o pedido. Erro: ' . $e->getMessage();

            if ($e->getCode() == 1000)
                $response->addDataSet(new DataSet('', ["status" => "CANCELLED"]));

            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
        }
    }

    private function createCancelOrder($cdfilial, $cdcaixa, $nrcomanda, $nrseqvenda)
    {
        // Verifica se o pedido está pendente de fechamento 
        $pendingFile = null;
        $pendingPath = "{$this->logPath}" . DIRECTORY_SEPARATOR . "Close" . DIRECTORY_SEPARATOR . "Pending" . DIRECTORY_SEPARATOR . "";

        if (file_exists($pendingPath)) {
            chdir($pendingPath);
            $pendingFile = glob("*-" . $nrcomanda . ".json");
        }

        if (!empty($pendingFile))
            throw new \Exception("Não é possível cancelar o pedido ($cdfilial-$nrcomanda), venda (" . ((!empty($nrseqvenda)) ? $nrseqvenda : "não informada") . "), pois o pedido está em processo de fechamento.");

        //
        $receivedFile = null;
        $receivedPath = "{$this->logPath}" . DIRECTORY_SEPARATOR . "Cancel" . DIRECTORY_SEPARATOR . "Received" . DIRECTORY_SEPARATOR . date('Y') . DIRECTORY_SEPARATOR . date('m') . DIRECTORY_SEPARATOR . date('d') . DIRECTORY_SEPARATOR;

        if (file_exists($receivedPath)) {
            chdir($receivedPath);
            $receivedFile = glob("*-" . $nrcomanda . ".json");
        }

        if (!empty($receivedFile))
            throw new \Exception("Não é possível cancelar o pedido ($cdfilial-$nrcomanda), venda (" . ((!empty($nrseqvenda)) ? $nrseqvenda : "não informada") . "), pois o cancelamento já foi solicitado anteriormente.", 1000);
        //

        $order = array(
            'CDFILIAL'      => $cdfilial,
            'CDCAIXA'       => $cdcaixa,
            'NRSEQVENDA'    => $nrseqvenda,
            'CDOPERADOR'    => null,
            'CDSUPERVISOR'  => null,
            'NRCOMANDA'     => $nrcomanda
        );
        $resultReceived = $this->registerReceivedOrder('Cancel', date('dmy'), $order);

        $nrvendarest = $this->util->getNrvendarest($cdfilial, $nrcomanda);
        $nrvendarest = $nrvendarest['NRVENDAREST'];

        if (!$nrvendarest)
            throw new \Exception("Não foi posssível encontrar o pedido ($cdfilial-$nrcomanda) .", 1000);

        $saleData = $this->orderService->getComandaData($nrcomanda, $cdfilial);

        if ($saleData['IDSTCOMANDA'] == '4')
            throw new \Exception("Não é possivel realizar o cancelamento pois o pedido ($cdfilial-$nrcomanda) já foi cancelado.", 1000);

        $dthrabermesa = new \DateTime($saleData['DTHRABERMESA']);
        $diff = $dthrabermesa->diff(new \DateTime());

        $saleCancelTime = $this->maxSaleCancelTime;
        if ($diff->i >= $saleCancelTime)
            throw new \Exception("A comanda $nrcomanda não pode ser cancelada. (Execedeu os $saleCancelTime minutos)", 1000);

        // Cancelar a venda
        $saleStatus = $this->util->getSaleByOrderCode($cdfilial, $cdcaixa, $nrcomanda);
        if (!empty($saleStatus) && ($saleData['IDSTCOMANDA'] == 'X')) {

            // Não existe a venda associada a comanda
            //=================== CANCELAMENTO COMANDAVEN ===================
            $this->util->cancelComandaven($cdfilial, $nrvendarest, $nrcomanda);

            //=================== CANCELAMENTO ITCOMANDAVEN ===================
            $canceledItems = $this->util->cancelProducts($cdfilial, $nrvendarest, $nrcomanda);

            //=================== CANCELAMENTO KDS ===================
            foreach ($canceledItems as $canceledItem) {
                $kdsSeq = $this->orderService->generateKDSSeq('KDSOPERACAOTEMP');
                $dsOperacao = array(
                    'operation' => 'cancelItem',
                    'params' => array(
                        'CDFILIAL'     => $cdfilial,
                        'NRVENDAREST'  => $nrvendarest,
                        'NRCOMANDA'    => $nrcomanda,
                        'NRPRODCOMVEN' => $canceledItem['NRPRODCOMVEN']
                    )
                );
                $this->util->insertKDSTEMP($kdsSeq, $dsOperacao);
            }

            if ($saleStatus['IDSITUVENDA'] == 'C') {
                throw new \Exception("A venda $nrseqvenda (pedido $nrcomanda) já foi cancelada.", 1000);
            }


            if ($saleStatus['IDSITUVENDA'] == 'P') {
                throw new \Exception("A venda $nrseqvenda (pedido $nrcomanda) encontra-se pendente no sistema (gestão de vendas)", 1000);
            }

            if ($saleStatus['IDSITUVENDA'] == 'O') {
                $cdoperador     = $saleStatus['CDOPERADOR'];
                $idstatusnfce   = $saleStatus['IDSTATUSNFCE'];
                $cdsupervisor   = $saleStatus['CDSUPERVISOR'];
                $nrseqvenda     = $saleStatus['NRSEQVENDA'];

                $order = array(
                    'CDFILIAL'      => $cdfilial,
                    'CDCAIXA'       => $cdcaixa,
                    'NRSEQVENDA'    => $nrseqvenda,
                    'CDOPERADOR'    => $cdoperador,
                    'CDSUPERVISOR'  => $cdsupervisor,
                    'NRCOMANDA'     => $nrcomanda
                );

                $resultPending = $this->registerPendingOrder('Cancel', date('dmy'), $order);

                if ($resultPending['error']) {
                    throw new \Exception($resultPending['message']);
                }
            }
        } else if ($saleData['IDSTCOMANDA'] != 'X') {
            // Não existe a venda associada a comanda
            //=================== CANCELAMENTO COMANDAVEN ===================
            $this->util->cancelComandaven($cdfilial, $nrvendarest, $nrcomanda);

            //=================== CANCELAMENTO ITCOMANDAVEN ===================
            $canceledItems = $this->util->cancelProducts($cdfilial, $nrvendarest, $nrcomanda);

            //=================== CANCELAMENTO KDS ===================
            foreach ($canceledItems as $canceledItem) {
                $kdsSeq = $this->orderService->generateKDSSeq('KDSOPERACAOTEMP');
                $dsOperacao = array(
                    'operation' => 'cancelItem',
                    'params' => array(
                        'CDFILIAL'     => $cdfilial,
                        'NRVENDAREST'  => $nrvendarest,
                        'NRCOMANDA'    => $nrcomanda,
                        'NRPRODCOMVEN' => $canceledItem['NRPRODCOMVEN']
                    )
                );
                $this->util->insertKDSTEMP($kdsSeq, $dsOperacao);
            }

            //=================== REGISTRO DE CANCELAMENTO ===================

            //Tratar arquivo de sucesso de cancelamento

            $order = array(
                'CDFILIAL'      => $cdfilial,
                'CDCAIXA'       => $cdcaixa,
                'NRSEQVENDA'    => null,
                'CDOPERADOR'    => null,
                'CDSUPERVISOR'  => null,
                'NRCOMANDA'     => $nrcomanda
            );

            $this->finish('Cancel', date('dmy'), $receivedPath, $resultReceived['fileName'], false);
        }
    }

    private function validatePayment($totalValue, $payments, &$giftCards, $cashBack)
    {

        $money          = $this->util->getMoney();
        $moneyValue     = 0;
        $paymentValue   = 0;
        $giftCardValue  = 0;

        foreach ($payments as $payment) {
            if ($payment['paymentCode'] === $money) {
                $moneyValue = $this->util->add($payment['value'], $moneyValue);
            } else {
                $paymentValue = $this->util->add($payment['value'], $paymentValue);
            }
        }

        foreach ($giftCards as $card) {
            $giftCardValue = $this->util->add($card['value'], $giftCardValue);
        }

        if ($moneyValue >= $totalValue) {
            $giftCards = array();
        }

        if ($giftCardValue >= $totalValue && !empty($payments)) {
            throw new \Exception("Não é possível utilizar outro meio de pagamento visto que o Cartão Vale Presente é suficiente para pagar o pedido.");
        } else if ($this->util->add($moneyValue, $this->util->add($paymentValue, $this->util->add($giftCardValue, $cashBack))) < $totalValue) {
            throw new \Exception("O valor pago é menor que valor do pedido.");
        } else if (($this->util->add($paymentValue, $cashBack)) > $totalValue) {
            throw new \Exception("O valor pago é maior que o valor do pedido.");
        }
    }

    public function validateCashBack($clientCode, $consumerCode, $cashBack, $subsidiaryCode, $orderCPF = null)
    {
        $fidelityUtil = $this->util->getFidelityUtil($subsidiaryCode);
        if(!empty($fidelityUtil) && $fidelityUtil['IDUTILFIDUDS'] == "S"){
            if(empty($orderCPF))
                throw new \Exception("Não é possível utilizar o fidelidade UDS sem informar o CPF.");
        } else {
            if (empty($clientCode) || empty($consumerCode))
                throw new \Exception("Não é possível utilizar cashback sem informar o consumidor!");

            $consumerBalance          = $this->util->getConsumerFidelity($clientCode, $consumerCode);
            $orderFidelityBalance     = $this->util->getValueFidelityOrders($clientCode, $consumerCode);
            $consumerBalance['value'] = $this->util->subtract($consumerBalance['value'], $orderFidelityBalance);

            if ($consumerBalance['value'] < $cashBack)
                throw new \Exception("Não foi possível utilizar $cashBack de saldo fidelidade, consumidor possuí apenas $consumerBalance[value].");
        }
    }

    public function getTables(Request $request, Response $response)
    {
        try {
            $body = $request->getParameters();

            if (empty($body['subsidiaryCode']))
                throw new \Exception("Parâmetro subsidiayCode não encontrado.");

            if (empty($body['storeCode']))
                throw new \Exception("Parâmetro storeCode não encontrado.");
            
            if (empty($body['posCode']))
                throw new \Exception("Parâmetro posCode não encontrado.");

            $final = array('tables' => array());
            $subsidiaryCode = $body['subsidiaryCode'];
            $storeCode = $body['storeCode'];
            $posCode = $body['posCode'];

            $confTelaCaixa = $this->menuService->getConfTelaCaixaEx($subsidiaryCode, $posCode);
            
            if($confTelaCaixa['error']) throw new \Exception($confTelaCaixa['message']);

            $CDFILIALEX = $confTelaCaixa['data']['CDFILIAL'];
            $NRCONFTELA = $confTelaCaixa['data']['NRCONFTELA'];
            $DTINIVIGENCIA = $confTelaCaixa['data']['DTINIVIGENCIA'];

            $tables = $this->orderService->getTables($subsidiaryCode, $storeCode, $CDFILIALEX, $NRCONFTELA, $DTINIVIGENCIA);

            if(empty($tables)) throw new \Exception("Não foram encontradas mesas para esses parâmetros!");

            $listStatus = [
                'O' => [ 
                        "statusId"=> 1 ,
                        "status"  => "Mesa Aberta"
                        ],
                'D' => [ 
                        "statusId"=> 2 ,
                        "status"  => "Mesa Livre"
                        ],
                'S' => [ 
                        "statusId"=> 3 ,
                        "status"  => "Conta Fechada"
                        ],
                'R' => [ 
                        "statusId"=> 4 ,
                        "status"  => "Mesa em Recebimento"
                        ],
            ];

            $ambients = array();
            foreach ($tables as $table){
                if(empty($ambients[$table['CDSALA']])){
                    
                    $statusTable = isset($listStatus[$table['IDSTMESAAUX']]) ? $listStatus[$table['IDSTMESAAUX']] : null;
                    $ambient = array(
                        'roomCode' => $table['CDSALA'],
                        'roomName' => $table['NMSALA'],
                        'tables'=> [
                            $table['NRMESA'] => array(
                                'tableNumber' => $table['NRMESA'],
                                'tableName' => $table['NMMESA'],
                                'statusId' => $statusTable['statusId'],
                                'status' => $statusTable['status'],
                                'time'=> $table['TEMPOCOMSUMIR'],
                                'maxPeople' => $table['NRPESSOAMESA'],
                                'people' => $table['NRPESMESAVEN'],
                                'position' => $table['NRPOSICAOMESA']
                            )
                        ]
                    );
                    $ambients[$table['CDSALA']] = $ambient;
                } else {
                    $statusTable = isset($listStatus[$table['IDSTMESAAUX']]) ? $listStatus[$table['IDSTMESAAUX']] : null;
                    $newTable = array(
                        'tableNumber' => $table['NRMESA'],
                        'tableName' => $table['NMMESA'],
                        'statusId' => $statusTable['statusId'],
                        'status' => $statusTable['status'],
                        'time'=> $table['TEMPOCOMSUMIR'],
                        'maxPeople' => $table['NRPESSOAMESA'],
                        'people' => $table['NRPESMESAVEN'],
                        'position' => $table['NRPOSICAOMESA']
                    );
                    $ambients[$table['CDSALA']]['tables'][$table['NRMESA']] = $newTable;
                }
            }
            
            foreach($ambients as &$ambient){
                $ambient['tables'] = array_values($ambient['tables']);
            }

            $response->addDataSet(new DataSet('', array_values($ambients)));
        } catch (\Exception $e) {
            $message = 'Erro ao buscar mesas. Erro: ' . $e->getMessage();
            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
        }
    }

    public function imprimeParcial(Request $request, Response $response)
    {
        try {
            $body = $request->getParameters();

            if (empty($body['saleCode']))
                throw new \Exception("Parâmetro saleCode não encontrado.");

            if (empty($body['tableNumber']))
                throw new \Exception("Parâmetro tableNumber não encontrado.");

            $nrcomanda      = $body['saleCode'];
            $nrmesa         = $body['tableNumber'];
            $comandaInfo    = $this->util->getComandaInfo($nrcomanda, $nrmesa);
            $cdfilial       = $comandaInfo['CDFILIAL'];
            $nrvendarest    = $comandaInfo['NRVENDAREST'];
            $cdloja         = $comandaInfo['CDLOJA'];
            $dscomanda      = $comandaInfo['DSCOMANDA'];
            $nrpessoas      = $comandaInfo['NRPESMESAVEN'];
            $cdprodcouver   = $comandaInfo['CDPRODCOUVER'];
            $cdprodconsum   = $comandaInfo['CDPRODCONSUM'];
            $idcomisvenda   = $comandaInfo['IDCOMISVENDA'];
            $vrcomisvenda   = $comandaInfo['VRCOMISVENDA'];
            $idlugarmesa    = $comandaInfo['IDLUGARMESA'];
            $idstmesaaux    = $comandaInfo['IDSTMESAAUX'];
            $nmfanven       = $comandaInfo['NMFANVEN'];

            $stComandaVens   = "_" . $nrcomanda . "_";
            $stVendaRests    = "_" . $nrvendarest . "_";
            $stPosicoes      = $impPosicao = "T";
            $produtosParcial = $this->util->getProdutosParcial($cdfilial, $stComandaVens, $stPosicoes, $impPosicao);

            $totalProdutosTaxa      = 0;
            $totalProdutosTaxaDesc  = 0;
            $totalSemDesconto       = 0;
            $totalDesconto          = 0;
            $totalSubsidy           = 0;

            self::calculaValoresParcial($produtosParcial, $cdprodcouver, $cdprodconsum, $totalProdutosTaxa, $totalProdutosTaxaDesc, $totalSemDesconto, $totalDesconto, $totalSubsidy);

            $totalSemDesconto   = round($totalSemDesconto, 2);
            $totalProdutosTaxa  = round(round($totalProdutosTaxa, 2) - $totalProdutosTaxaDesc, 2);
            $total              = round($totalSemDesconto - $totalDesconto, 2);

            $fidelityValue      = 0;
            $fidelityDiscount   = 0;

            $taxaDeServico = $this->taxaDeServico($idcomisvenda, $vrcomisvenda, $cdprodcouver, $cdprodconsum, $totalProdutosTaxa, $cdfilial, $nrvendarest, $nrcomanda, $idstmesaaux, $produtosParcial);

            $total += $taxaDeServico;

            $impressoraInfo = $this->util->getDadosImpressora();

            if (empty($impressoraInfo))
                throw new \Exception("Não foi encontrada impressora parametrizada!");

            $nrseqimprloja = $impressoraInfo['NRSEQIMPRLOJA'];

            $tempoPermanencia = self::calculaTempoPermanencia($cdfilial, $nrvendarest, $nrcomanda);
            $horas = $tempoPermanencia["horas"];
            $minutos = $tempoPermanencia["minutos"];
            $segundos = $tempoPermanencia["segundos"];
            $modo = "M";
            $dadosImpressao = '';
            $totalPorPessoa = floatval($total) / floatval($nrpessoas);

            $retorno = $this->orderService->imprimeParcial(
                $produtosParcial,
                $cdfilial,
                $stVendaRests,
                $stComandaVens,
                $cdprodcouver,
                $modo,
                $nrvendarest,
                $nrcomanda,
                $cdloja,
                $nrmesa,
                $impPosicao,
                $nrpessoas,
                round($totalSemDesconto - $totalDesconto, 2),
                $taxaDeServico,
                $total,
                $totalPorPessoa,
                $horas,
                $minutos,
                $segundos,
                $dscomanda,
                $dadosImpressao,
                $fidelityValue,
                $idcomisvenda,
                $nmfanven,
                $idlugarmesa,
                $vrcomisvenda,
                $nrseqimprloja
            );

            if (isset($retorno['error']) && $retorno['error'] === false) {
                $response->addMessage(new Message("Conta impressa com sucesso.", self::GENERAL_MESSAGE));
            } else if (isset($retorno['error']) && $retorno['error'] === true) {
                throw new \Exception($retorno['message']);
            } else {
                throw new \Exception("Houve um problema ao imprimir seu pedido.");
            }
        } catch (\Exception $e) {
            $message = 'Erro ao imprimir parcial. Erro: ' . $e->getMessage();
            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
        }
    }

    public function printText(Request $request, Response $response)
    {
        try {
            $body = $request->getParameters();

            $text = $body['text'];

            $dadosImpressora = $this->util->getDadosImpressora();

            $printerParams = $this->impressaoUtil->buscaParametrosImpressora($dadosImpressora['IDMODEIMPRES']);
            $options['bold'] = true;
            $options['letterType'] = '2';
            $string = $this->impressaoUtil->centraliza($printerParams, "") . $printerParams['comandoEnter'];
            $string .= $this->impressaoUtil->centraliza($printerParams, "") . $printerParams['comandoEnter'];
            $string .= $this->impressaoUtil->centraliza($printerParams, $text) . $printerParams['comandoEnter'];
            $string .= $this->impressaoUtil->centraliza($printerParams, "") . $printerParams['comandoEnter'];
            $string .= $this->impressaoUtil->centraliza($printerParams, "") . $printerParams['comandoEnter'];

            $this->impressaoUtil->imprimeNaoFiscal($string, $dadosImpressora, false, $options);
        } catch (\Exception $e) {
            $message = 'Erro ao imprimir texto. Erro: ' . $e->getMessage();
            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
        }
    }

    private function calculaTempoPermanencia($cdfilial, $nrvendarest, $nrcomanda)
    {
        $tempoPermanencia = $this->util->getTempoPermanencia($cdfilial, $nrvendarest, $nrcomanda);

        $now = new \DateTime();
        $entrada = \DateTime::createFromFormat("Y-m-d H:i:s.u", $tempoPermanencia["DATA"]);
        $intervalo = $entrada->diff($now);
        $intervaloHoras = str_pad((int)$intervalo->format('%R%h'), 2, '0', STR_PAD_LEFT);
        $intervaloMinutos = str_pad((int)$intervalo->format('%R%i'), 2, '0', STR_PAD_LEFT);
        $intervaloSegundos = str_pad((int)$intervalo->format('%R%s'), 2, '0', STR_PAD_LEFT);

        return array(
            "horas"    => $intervaloHoras,
            "minutos"  => $intervaloMinutos,
            "segundos" => $intervaloSegundos
        );
    }

    private function calculaValoresParcial($produtosParcial, $cdprodcouver, $cdprodconsum, &$totalProdutosTaxa, &$totalProdutosTaxaDesc, &$totalSemDesconto, &$totalDesconto, &$totalSubsidy)
    {
        foreach ($produtosParcial as $item) {
            if ($item['CDPRODUTO'] != $cdprodcouver && $item['CDPRODUTO'] != $cdprodconsum) {
                $vrProd = floatval(bcdiv($item['QTPRODCOMVEN'] * $item['VRPRECCOMVEN'], 1, 2)) + $item['VRACRCOMVEN'];
                if ($item['IDCOBTXSERV'] == 'S') {
                    $totalProdutosTaxa += $vrProd;
                    $totalProdutosTaxaDesc += $item['VRDESCCOMVEN'];
                }
                $totalSemDesconto += $vrProd;
                $totalDesconto += $item['VRDESCCOMVEN'];
                $totalSubsidy += floatval(bcmul($item['QTPRODCOMVEN'], $item['VRPRECCLCOMVEN'], 2));
            }
        }
    }

    private function taxaDeServico($idcomisvenda, $vrcomisvenda, $cdprodcouver, $cdprodconsum, $totalProdutos, $cdfilial, $nrvendarest, $nrcomanda, $IDSTMESAAUX, $produtosParcial)
    {
        $taxaDeServico = 0;
        $params = array(
            'CDFILIAL' => $cdfilial,
            'NRVENDAREST' => $nrvendarest,
            'NRCOMANDA' => $nrcomanda
        );
        $dadosComissao = $this->util->getComissaoVenda($cdfilial, $nrvendarest, $nrcomanda);

        if ($idcomisvenda === "S") {
            if (is_null($dadosComissao['VRCOMISPOR'])) {
                $taxaCalculada = $this->util->truncate($totalProdutos * floatval($vrcomisvenda) / 100, 2);
            } else {
                $taxaCalculada = $this->util->truncate($totalProdutos * floatval($dadosComissao["VRCOMISPOR"]) / 100, 2);
            }
        } else if ($idcomisvenda === "G") {
            $taxaCalculada = 0;
            foreach ($produtosParcial as $item) {
                if ($item['CDPRODUTO'] != $cdprodcouver && $item['CDPRODUTO'] != $cdprodconsum) {
                    $vrProd = $item['QTPRODCOMVEN'] * ($item['VRPRECCOMVEN'] + $item['VRPRECCLCOMVEN']) - $item['VRDESCCOMVEN'];
                    if ($item['IDCOBTXSERV'] == 'S' && $item['IDVRINCIDECONTA'] == "S") {
                        $taxaCalculada += $this->util->truncate($vrProd * floatval($item['VRPERCCOMISVEND']) / 100, 2);
                    }
                }
            }
        }

        if ($idcomisvenda !== "N") {
            // caso existir, retorna o valor calculado se for diferente de zero; se não, salva e retorna
            if ($IDSTMESAAUX !== 'O') {
                if (floatval($dadosComissao['VRCOMISVENDE']) != 0 || $dadosComissao['VRCOMISPOR'] > 0) {
                    $taxaDeServico = $taxaCalculada;
                }
            } else {
                if ($taxaCalculada > 0) {
                    $params = array(
                        $taxaCalculada,
                        $cdfilial,
                        $nrvendarest,
                        $nrcomanda
                    );
                    $this->entityManager->getConnection()->executeQuery("UPDATE_COMISSAO_VENDA", $params);
                }
                $taxaDeServico = $taxaCalculada;
            }
        }

        return $taxaDeServico;
    }

    public function recalculateComboDiscount($subsidiaryCode, $items)
    {
        $recalculate = false;

        foreach ($items as &$item) {
            if (!empty($item['childItems'])) {
                $sonFullValues = array();
                $totalDiscount = 0;
                $sonsTotalPrice = 0;
                $sonsToApplyDiscount = 0;
                $quantSons = 0;
                $fatherProduct = $item['productCode'];
                $returnGroups = $this->menuService->getComboGroupsBySubsidiaryDiscount($subsidiaryCode, $fatherProduct);
                if (!$returnGroups['error']) {
                    $comboGroups = $returnGroups['data'];
                    if (empty($comboGroups)) {
                        $useSubsidiaryGroups = false;
                        $returnGroups = $this->menuService->getComboGroupsByProductDiscount($fatherProduct);
                        if (!$returnGroups['error']) {
                            $comboGroups = $returnGroups['data'];
                        }
                    }
                }
                foreach ($comboGroups as $group) {
                    if ($group['IDAPLICADESC'] === 'S')
                        $recalculate = true;
                }
                if ($recalculate === true) {
                    foreach ($item['childItems'] as &$son) {
                        $quantSons++;
                        foreach ($comboGroups as $group) {
                            $sonDiscount = $this->util->getSonDiscount($son['productCode'], $item['productCode'], $group['CDGRUPROMOC']);
                            if (!empty($sonDiscount)) {
                                if ($sonDiscount['IDPERVALORDES'] === 'P') {
                                    $percentApllied = bcsub(100, $sonDiscount['VRDESPRODPROMOC'], 2);
                                    $fullValue = bcdiv($son['unitValue'] * 100, $percentApllied, 2);
                                    $totalDiscount = bcadd($totalDiscount, bcsub($fullValue, $son['unitValue']), 2);
                                    $sonFullValues[$son['productCode']] = $fullValue;
                                } else {
                                    $fullValue = bcadd($son['unitValue'], $sonDiscount['VRDESPRODPROMOC'], 2);
                                    $totalDiscount = bcadd($totalDiscount, $sonDiscount['VRDESPRODPROMOC'], 2);
                                    $sonFullValues[$son['productCode']] = $fullValue;
                                }

                                if ($group['IDAPLICADESC'] === 'S') {
                                    $sonsToApplyDiscount++;
                                    $son['APPLYDISCOUNT'] = true;
                                    $sonsTotalPrice = bcadd($sonsTotalPrice, $sonFullValues[$son['productCode']], 2);
                                }
                            }
                        }
                    }
                    $remaingDiscount = $totalDiscount;
                    unset($son);
                    if ($quantSons > $sonsToApplyDiscount) {
                        if ($totalDiscount < $sonsTotalPrice) {
                            if ($sonsToApplyDiscount === 1) {
                                foreach ($item['childItems'] as &$son) {
                                    if (isset($sonFullValues[$son['productCode']])) {
                                        $son['unitValue'] = $sonFullValues[$son['productCode']];
                                        if (!empty($son['APPLYDISCOUNT'])) {
                                            $son['unitValue'] = bcsub($sonFullValues[$son['productCode']], $totalDiscount, 2);
                                        }
                                    }
                                }
                                unset($son);
                            } else {
                                foreach ($item['childItems'] as &$son) {
                                    if (isset($sonFullValues[$son['productCode']])) {
                                        $son['unitValue'] = $sonFullValues[$son['productCode']];
                                        if (!empty($son['APPLYDISCOUNT'])) {
                                            $productDiscount = bcdiv(
                                                (
                                                    (
                                                        ($son['unitValue']
                                                        ) * $son['quantity']
                                                    ) * $totalDiscount
                                                ) / $sonsTotalPrice,
                                                1,
                                                2
                                            );
                                            $son['unitValue'] = bcdiv(bcsub($son['unitValue'], $productDiscount, 2), 1, 2);
                                            $remaingDiscount = bcdiv(bcsub($remaingDiscount, $productDiscount, 2), 1, 2);
                                        }
                                    }
                                }
                                unset($son);
                                while ($remaingDiscount > 0) {
                                    foreach ($item['childItems'] as &$son) {
                                        if ($remaingDiscount > 0) {
                                            if ($son['unitValue'] > 0.01 && !empty($son['APPLYDISCOUNT'])) {
                                                $son['unitValue'] = bcdiv(bcsub($son['unitValue'], 0.01, 2), 1, 2);
                                                $remaingDiscount = bcdiv(bcsub($remaingDiscount, 0.01, 2), 1, 2);
                                            }
                                        }
                                    }
                                    unset($son);
                                }
                            }
                        }
                    }
                }
            }
        }
        unset($item);
        return $items;
    }

    public function getWaiters(Request $request, Response $response)
    {
        try {
            $waiters = $this->orderService->getWaiters();
            $response->addDataSet(
                new DataSet(
                    '',
                    $waiters
                )
            );
        } catch (\Exception $e) {
            $response->addMessage(new Message($e->getMessage(), self::ERROR_MESSAGE));
        }
    }

    public function getOrder(Request $request, Response $response)
    {
        try {
            if (isset($_GET['saleCode']))
                $nrcomanda = $_GET['saleCode'];
            else
                $nrcomanda = 'T';

            if (isset($_GET['tableNumber']))
                $nrmesa = $_GET['tableNumber'];
            else
                throw new \Exception('Mesa não informada.');

            $fileName = $nrcomanda;
            $this->util->saveLog($nrcomanda . '-' . $nrmesa, 'Comanda' . date("dmY"), $fileName . '-');

            $order = $this->orderService->getOrderProducts($nrcomanda, $nrmesa);
            $response->addDataSet(
                new DataSet(
                    '',
                    $order
                )
            );
        } catch (\Exception $e) {
            $message = 'Erro ao recuperar comanda. ' . $e->getMessage();
            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
        }
    }

    public function closeOrder(Request $request, Response $response)
    {
        try {
            $inputJSON = file_get_contents('php://input');
            $input = json_decode($inputJSON, true);

            if (isset($input['subsidiaryCode']))
                $cdfilial = $input['subsidiaryCode'];
            else
                throw new \Exception('Filial não informada.');

            if (isset($input['saleCode']))
                $nrcomanda = $input['saleCode'];
            else
                throw new \Exception('Comanda não informada.');

            if (isset($input['posCode']))
                $cdcaixa = $input['posCode'];
            else
                throw new \Exception('Caixa não informado.');

            $checkCloseByApi = true;
            if (isset($input['checkCloseByApi']) && $input['checkCloseByApi'] === false) $checkCloseByApi = $input['checkCloseByApi'];

            $order = $this->orderService->getOrder($cdfilial, $nrcomanda);

            if (empty($order))
                throw new \Exception("Comanda não encontrada! Verifique os dados de entrada.");

            if ($checkCloseByApi && $order['IDFECHAMENTOAPI'] == 'N')
                throw new \Exception("Esta comanda não pode ser fechada via API, o fechamento deve ser realizado atraves do PDV.");

            $nrvendarest = $order['NRVENDAREST'];
            $message = '';
            $order['CDCAIXA'] = $cdcaixa;
            $order['NRSEQMOVEXT'] = "001";

            $orderAlreadyClosed = false;
            if ($order['IDSTCOMANDA'] === '4') {
                $orderAlreadyClosed = null;
                $message = "Não é possivel realizar o fechamento pois o pedido ($cdfilial-$nrcomanda) foi cancelado.";
                $response->addMessage(new Message($message, self::ERROR_MESSAGE));
            } else if ($order['IDSTCOMANDA'] === 'X') {
                $orderAlreadyClosed = true;
                $message = "O pedido ($cdfilial-$nrcomanda) já foi fechado, a consulta encontra-se disponível no endpoint '/getSale'.";
                $response->addMessage(new Message($message, self::ERROR_MESSAGE));
            } else {
                $orderAlreadyClosed = false;
                $payments = $this->orderService->getOrderPayments($cdfilial, $nrvendarest);

                $paymentsValue = 0;
                if (!empty($payments)) {
                    foreach ($payments as $payment) {
                        if($payment['IDTIPOMOVIVEDLV'] == "E") $paymentsValue = bcadd($payment['VRMOVIVEND'], $paymentsValue, 2);
                    }
                }

                $cashBackValue = $this->orderService->getOrderCashback($cdfilial, $nrvendarest, $nrcomanda);
                $items = $this->orderService->getOrderItems($nrvendarest, $nrcomanda, $cdfilial);

                // $productsValue = 0;
                // foreach ($items as $item) {
                //     $productsValue = $this->util->add($productsValue, $item['TOTAL']);
                // }
                $productsValue = $this->orderService->calculaValorDaVenda($items)['total'];

                $totalPayment = $this->util->float($this->util->add($cashBackValue['VRDESCFID'], $paymentsValue));
                $totalOrderValue = $this->util->float($this->util->add($productsValue, $order['VRACRCOMANDA']));

                if ($totalPayment < $totalOrderValue) {
                    $remaingValue = $this->util->subtract($totalOrderValue, $totalPayment);
                    throw new \Exception("Ainda devem ser pago(s) R$ $remaingValue para que a comanda seja fechada.");
                }

                $receivedFile = null;
                $receivedPath = "{$this->logPath}" . DIRECTORY_SEPARATOR . "Close" . DIRECTORY_SEPARATOR . "Received" . DIRECTORY_SEPARATOR . date('Y') . DIRECTORY_SEPARATOR . date('m') . DIRECTORY_SEPARATOR . date('d') . DIRECTORY_SEPARATOR . "";
                if (file_exists($receivedPath)) {
                    chdir($receivedPath);
                    $receivedFile = glob("*-" . $nrcomanda . ".json");
                }

                if (!empty($receivedFile))
                    throw new \Exception("Não é possível solicitar o fechamento do pedido ($cdfilial-$nrcomanda), pois o fechamento já foi solicitado anteriormente. Favor direcionar-se ao caixa.");

                $this->registerReceivedOrder('Close', date('dmy'), $order);
                $this->registerPendingOrder('Close', date('dmy'), $order);


                $response->addMessage(new Message("O pedido ($cdfilial-$nrcomanda) foi incluído para fechamento.", self::GENERAL_MESSAGE));
            }

            $response->addDataSet(
                new DataSet(
                    '',
                    array('orderAlreadyClosed' => $orderAlreadyClosed)
                )
            );

        } catch (\Exception $e) {
            $message = 'Erro ao solicitar o fechamento da comanda. ' . $e->getMessage();
            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
        }
    }

    public function registerReceivedOrder($function, $today, $order) {
        $receivedPath = "{$this->logPath}/$function/Received/".date('Y')."/".date('m')."/".date('d')."/";
        
        if(!is_dir($receivedPath)) {
            if($this->util->isRunningOnWindows()) {
                mkdir($receivedPath, 0777, true);
            } else {
                mkdir($receivedPath, 0775, true);
            }
        }

        $fileName = date('Ymd-H-i-s') . "-" . $order['NRCOMANDA'] . ".json";
        $fileData = array(
            "PATH" => $receivedPath,
            "FILENAME" => $fileName,
            "order" => $order
        );

        file_put_contents($receivedPath . $fileName, json_encode($fileData));
        return array('error' => false, 'fileName' => $fileName, 'message' => "");
    }

    public function registerPendingOrder($function, $today, $order){
        $pendingPath = "{$this->logPath}/$function/Pending/";

        if(!is_dir($pendingPath)) {
            if($this->util->isRunningOnWindows()) {
                mkdir($pendingPath, 0777, true);
            } else {
                mkdir($pendingPath, 0775, true);
            }
        }

        $fileName = date('Ymd-H-i-s') . "-" . $order['NRCOMANDA'] . ".json";
        $fileData = array(
            "PATH" => $pendingPath,
            "FILENAME" => $fileName,
            "order" => $order
        );

        file_put_contents($pendingPath . $fileName, json_encode($fileData));
        return array('error' => false, 'message' => "");
    }

    private function validatePayments($cdfilial, $nrvendarest, $nrcomanda, $deliveryFee)
    {
        $productsValue = $this->orderService->getItemsValue($cdfilial, $nrvendarest, $nrcomanda);
        $orderPayments = $this->orderService->getOrderPayments($cdfilial, $nrvendarest);

        $paymentsValue = 0;
        $money = $this->util->getMoney();
        $hasMoney = false;

        foreach ($orderPayments as $orderPayment) {
            if($orderPayment['IDTIPOMOVIVEDLV'] == "E"){
                $hasMoney = $orderPayment['CDTIPORECE'] === $money;
                $paymentsValue = $this->util->add($paymentsValue, $orderPayment['VRMOVIVEND']);
            }
        }

        $totalOrderValue = $this->util->add($productsValue, $deliveryFee);
        if ($paymentsValue != $totalOrderValue) {
            if ($paymentsValue > $totalOrderValue) {
                if ($hasMoney === false) {
                    throw new \Exception("Valor do pagamento inválido. Valor total da compra: R$$totalOrderValue.");
                }
            } else if ($paymentsValue < $totalOrderValue) {
                throw new \Exception("Valor do pagamento inválido. Valor total da compra: R$$totalOrderValue.");
            }
        }
    }

    public function fechaPedidosParalelo()
    {
        $controlPath = "{$this->logPath}" . DIRECTORY_SEPARATOR . ".." . DIRECTORY_SEPARATOR . "CLOSE-CTRL-FILE.dat";
        /*
        if (!file_exists($controlPath)){
            if(!$this->util->isRunningOnWindows()) {
                mkdir($controlPath, 0775, true);
            } 
        }
        */

        if (!file_exists($controlPath))
            $this->updateControlFile(false);


        $controlFile = json_decode(base64_decode(file_get_contents($controlPath)), true);

        $controlDiffTime = (time() - filemtime($controlPath));
        if ($controlFile['lock'] && ($controlDiffTime <= 55))
            die;

        if ($this->executionDebug)
            $this->util->saveLog('Início serviço fechamento de pedidos', 'Execution/' . date("Y/m"), 'CLOSE-');

        $this->updateControlFile(true);
        $processingErrors = array();

        $running = null;
        $handles = [];
        $c = 0;

        $numberOfPendingOrders = 0;

        try {
            $pedidos = $this->recuperaPedidos('Close');

            $numberOfPendingOrders = count($pedidos);

            echo "Numero total de pedidos pendentes de fechamento: $numberOfPendingOrders\n";
            echo "Numero de pedidos a serem fechados no lote: " . (($numberOfPendingOrders <= $this->maxParallelClose) ? $numberOfPendingOrders : $this->maxParallelClose) . "\n";

            if ($numberOfPendingOrders > 0) {

                $this->util->saveLog("Numero total de pedidos pendentes de fechamento: $numberOfPendingOrders", 'Execution/' . date("Y/m"), 'CLOSE-');
                $this->util->saveLog("Numero de pedidos a serem fechados no lote: " . (($numberOfPendingOrders <= $this->maxParallelClose) ? $numberOfPendingOrders : $this->maxParallelClose), 'Execution/' . date("Y/m"), 'CLOSE-');

                foreach ($pedidos as $pedido) {
                    $handles[$c] = curl_init();

                    curl_setopt_array($handles[$c], array(
                        CURLOPT_URL             => $this->projectUrl . 'internal/closeOrderById',
                        CURLOPT_RETURNTRANSFER  => true,
                        CURLOPT_TIMEOUT         => 50,
                        CONNECTION_TIMEOUT     => 5,
                        CURLOPT_SSL_VERIFYPEER     => false,
                        CURLOPT_HTTP_VERSION    => CURL_HTTP_VERSION_1_1,
                        CURLOPT_CUSTOMREQUEST   => 'POST',
                        CURLOPT_POSTFIELDS => json_encode($pedido),
                        CURLOPT_HTTPHEADER => array(
                            'Content-Type: application/json'
                        )
                    ));

                    $c++;

                    if ($c >= $this->maxParallelClose) {
                        break;
                    }
                }

                $mh = curl_multi_init();
                for ($i = 0; $i < $c; $i++) {
                    curl_multi_add_handle($mh, $handles[$i]);
                }

                do {
                    curl_multi_exec($mh, $running);
                } while ($running > 0);

                for ($i = 0; $i < $c; $i++) {
                    curl_multi_remove_handle($mh, $handles[$i]);
                }

                curl_multi_close($mh);
                echo "\n\n";
            }
        } catch (\Exception $e) {
            $this->util->saveLog($e->getMessage(), 'Error', 'CLOSE-');
            $this->util->saveLog("[ERRO] " . $e->getMessage(), 'Execution/' . date("Y/m"), 'CLOSE-');
        }

        if ($this->executionDebug)
            $this->util->saveLog('Fim serviço fechamento de pedidos.', 'Execution/' . date("Y/m"), 'CLOSE-');

        $this->updateControlFile(false);
    }

    public function closeOrderById(Request $request, Response $response)
    {
        set_time_limit(50);

        $hasError = false;

        try {
            $inputJSON = file_get_contents('php://input');
            $pedido = json_decode($inputJSON, true);

            if (empty($pedido["FILENAME"]) || !isset($pedido["FILENAME"])) throw new \Exception("Nome do arquivo não informado.");
            if (empty($pedido["PATH"]) || !isset($pedido["PATH"])) throw new \Exception("Caminho do arquivo não informado.");
            if (empty($pedido["order"]) || !isset($pedido["order"])) throw new \Exception("Informações do pedido não enviado.");
            $initTime = microtime(1);
            $fechamento = $this->orderService->closeOrder($pedido['order'], $pedido['order']['CDFILIAL'], $pedido['order']['NRCOMANDA'], $pedido['order']['NRSEQMOVEXT']);
            $endTime = microtime(1);
            $execTime = round(($endTime - $initTime) * 1000);

            if (!empty($fechamento['mensagemNfce']))
                $fechamento['message'] .= $fechamento['mensagemNfce'];

            if ($fechamento['error'] || (!empty($fechamento['IDSTATUSNFCE']) && $fechamento['IDSTATUSNFCE'] == null)) {

                $message = $pedido['order']['DSCOMANDA'] . ' - ' . $pedido['order']['NRCOMANDA'] . ' - ' . $fechamento['message'];
                $this->util->saveLog($message, 'Error', 'CLOSE-');

                $hasError = true;
                $this->error('Close', date('dmy'), $pedido['PATH'], $pedido['FILENAME'], $fechamento['message']);

                if ((!empty($fechamento['IDSTATUSNFCE'])) || (strpos($fechamento['message'], 'SAT') !== false)) {

                    if (empty($fechamento['NRSEQVENDA']))
                        $fechamento['NRSEQVENDA'] = "";

                    if(!(strpos($pedido['order']['IDORGCMDVENDA'], "CMD_") !== false)) $this->createCancelOrder($pedido['order']['CDFILIAL'], $pedido['order']['CDCAIXA'], $pedido['order']['NRCOMANDA'], $fechamento['NRSEQVENDA']);
                }

                $this->util->beginTransaction();
                try {
                    $this->util->insertLogOperFos($pedido['order']['CDFILIAL'], $pedido['order']['CDCAIXA'], 'FEC_PED', "Erro Fechamento " . $pedido['order']['CDFILIAL'] . " - " . $pedido['order']['NRCOMANDA'] . " - " . $fechamento['message']);
                    $this->util->commit();
                } catch (\Exception $e) {
                    $this->util->rollBack();
                }
            } else {
                $this->finish('Close', date('dmy'), $pedido['PATH'], $pedido['FILENAME'], true, $execTime, $fechamento['tempoTransmissao']);
                // if(strpos($pedido['order']['DSCOMANDA'], "DLV_") === false) $this->deleteOrder($pedido['order']['NRVENDAREST'], $pedido['order']['CDFILIAL'], $pedido['order']['NRCOMANDA']);
            }

            $response->addDataSet(new DataSet('', array('error' => false, 'pedido' => $pedido, 'fechamento' => $fechamento)));
        } catch (\Exception $e) {
            $message = 'Erro closeOrderById: ' . $e->getMessage();

            if (!$hasError)
                $this->error('Close', date('dmy'), $pedido['PATH'], $pedido['FILENAME'], $message);

            $this->util->saveLog($e->getMessage(), 'Error', 'CLOSE-');
            $response->addMessage(new Message("Erro no fechamento do pedido: " . $e->getMessage(), self::ERROR_MESSAGE));
        }
    }

    public function cancelaPedidos()
    {
        $controlPath = "{$this->logPath}" . DIRECTORY_SEPARATOR . ".." . DIRECTORY_SEPARATOR . "CANCEL-CTRL-FILE.dat";

        if (!file_exists($controlPath))
            $this->updateCancelControlFile(false);

        $controlFile = json_decode(base64_decode(file_get_contents($controlPath)), true);

        $controlDiffTime = (time() - filemtime($controlPath));
        if ($controlFile['lock'] && ($controlDiffTime <= 60))
            die;

        $executionErrors = array();

        if ($this->executionDebug)
            $this->util->saveLog('Início serviço cancelamento de pedidos', 'Execution/' . date("Y/m"), 'CANCEL-');

        $this->updateCancelControlFile(true);
        $processingErrors = array();

        try {
            $pedidos = $this->recuperaPedidos('Cancel');

            foreach ($pedidos as $pedido) {

                try {
                    $cancelamento = $this->orderService->cancelOrder($pedido['order']['CDFILIAL'], $pedido['order']['CDCAIXA'], $pedido['order']['NRSEQVENDA'], $pedido['order']['CDOPERADOR'], $pedido['order']['CDSUPERVISOR']);
                    if (!empty($cancelamento['mensagemNfce']))
                        $cancelamento['message'] .= $cancelamento['mensagemNfce'];

                    if ($cancelamento['error']) {

                        $message = $pedido['order']['NRCOMANDA'] . ' - ' . $cancelamento['message'];
                        $this->util->saveLog($message, 'Error', 'CANCEL-');
                        $this->util->beginTransaction();

                        try {
                            $this->util->insertLogOperFos($pedido['order']['CDFILIAL'], $pedido['order']['CDCAIXA'], 'CAN_PED', "Erro Cancelamento " . $pedido['order']['CDFILIAL'] . " - " . $pedido['order']['NRCOMANDA'] . " - " . $pedido['order']['NRSEQVENDA'] . ": " . $cancelamento['message']);
                            $this->util->commit();
                        } catch (\Exception $e) {
                            $this->util->rollBack();
                        }

                        $this->error('Cancel', date('dmy'), $pedido['PATH'], $pedido['FILENAME'], $cancelamento['message']);
                        $this->util->saveLog("[ERRO] FILIAL: {$pedido['order']['CDFILIAL']} - NRCOMANDA {$pedido['order']['NRCOMANDA']} - NRSEQVENDA: {$pedido['order']['NRSEQVENDA']} ERROR_MSG: {$cancelamento['message']}.", 'Execution/' . date("Y/m"), 'CANCEL-');
                    } else {

                        // Cancelar comanda caso esteja fechada
                        $saleData = $this->orderService->getComandaData($pedido['order']['NRCOMANDA'], $pedido['order']['CDFILIAL']);
                        if ((!empty($saleData)) && ($saleData['IDSTCOMANDA'] == 'X')) {

                            //=================== CANCELAMENTO DE ITENS + KDS
                            // $canceledItems = $this->util->cancelProducts($pedido['order']['CDFILIAL'], $saleData['NRVENDAREST'], $saleData['NRCOMANDA']);
                            // foreach ($canceledItems as $canceledItem) {
                            //     $kdsSeq = $this->orderService->generateKDSSeq('KDSOPERACAOTEMP');
                            //     $dsOperacao = array(
                            //         'operation' => 'cancelItem',
                            //         'params' => array(
                            //             'CDFILIAL'     => $pedido['order']['CDFILIAL'],
                            //             'NRVENDAREST'  => $saleData['NRVENDAREST'],
                            //             'NRCOMANDA'    => $saleData['NRCOMANDA'],
                            //             'NRPRODCOMVEN' => $canceledItem['NRPRODCOMVEN']
                            //         )
                            //     );
                            //     $this->util->insertKDSTEMP($kdsSeq, $dsOperacao);
                            // }
                            // //=================== CANCELAMENTO DE ITENS + KDS

                            // $this->util->cancelComandaven($pedido['order']['CDFILIAL'], $saleData['NRVENDAREST'], $saleData['NRCOMANDA']);
                        }

                        $this->finish('Cancel', date('dmy'), $pedido['PATH'], $pedido['FILENAME']);
                        $this->util->saveLog("FILIAL: {$pedido['order']['CDFILIAL']} - NRCOMANDA {$pedido['order']['NRCOMANDA']} - NRSEQVENDA: {$pedido['order']['NRSEQVENDA']} CANCELADO.", 'Execution/' . date("Y/m"), 'CANCEL-');
                    }
                } catch (\Exception $e) {
                    array_push($processingErrors, $pedido['order']['NRCOMANDA'] . " - " . $e->getMessage());
                    $this->error('Cancel', date('dmy'), $pedido['PATH'], $pedido['FILENAME'], $e->getMessage());
                }
            }

            if (!empty($processingErrors)) {
                $message = "";

                foreach ($processingErrors as $error)
                    $message .= $error;

                throw new \Exception($message);
            }
        } catch (\Exception $e) {
            $this->util->saveLog($e->getMessage(), 'Error', 'CANCEL-');
            $this->util->saveLog('[ERRO] ' . $e->getMessage(), 'Execution/' . date("Y/m"), 'CANCEL-');
        }

        if ($this->executionDebug)
            $this->util->saveLog('Fim serviço cancelamento de pedidos.', 'Execution/' . date("Y/m"), 'CANCEL-');

        $this->updateCancelControlFile(false);
    }

    public function getSale(Request $request, Response $response)
    {
        ini_set('memory_limit', '512M');

        try {
            $inputJSON = file_get_contents('php://input');
            $input = json_decode($inputJSON, true);

            if (isset($input['subsidiaryCode']))
                $cdfilial = $input['subsidiaryCode'];
            else
                throw new \Exception('Filial não informada.');

            if (isset($input['saleCode']))
                $nrcomanda = $input['saleCode'];
            else
                throw new \Exception('Comanda não informada.');

            if (isset($input['posCode'])) $cdcaixa = $input['posCode'];
            else throw new \Exception('Caixa não informado.');

            $controlPath = "{$this->logPath}" . DIRECTORY_SEPARATOR . ".." . DIRECTORY_SEPARATOR . "CLOSE-CTRL-FILE.dat";

            if (!file_exists($controlPath))
                throw new \Exception("Serviço de fechamento de pedidos inativo.");

            $controlFile = json_decode(base64_decode(file_get_contents($controlPath)), true);

            if ((time() - $controlFile['expiresIn']) > self::MAX_CONTROL_FILE_UPDATE_TIME)
                throw new \Exception("Serviço de fechamento de pedidos inativo.");

            $order = $this->orderService->getOrder($cdfilial, $nrcomanda);

            $nrvendarest = $order['NRVENDAREST'];
            $errorFile = null;
            $errorPath = "{$this->logPath}" . DIRECTORY_SEPARATOR . "Close" . DIRECTORY_SEPARATOR . "Error" . DIRECTORY_SEPARATOR . date('Y') . DIRECTORY_SEPARATOR . date('m') . DIRECTORY_SEPARATOR . date('d') . DIRECTORY_SEPARATOR . "";

            if (file_exists($errorPath)) {
                chdir($errorPath);
                $errorFile = glob("*-" . $nrcomanda . ".json");
            }

            if ($order['IDSTCOMANDA'] === 'X') {
                $cdcaixa = $order['CDCAIXA'];
                $saleDetails     = $this->util->getSaleDetails($cdfilial, $nrcomanda, $nrvendarest, $this->logPath, $cdcaixa);
                $nrseqvenda      = $saleDetails['NRSEQVENDA'];
                $cdcaixa         = $saleDetails['CDCAIXA'];
                $status          = $saleDetails['IDSTATUSNFCE'];
                $dtentrvenda     = $saleDetails['DTENTRVENDA'];
                $qrCode          = $saleDetails['DSQRCODENFCE'];
                $nfceAccessCode  = $saleDetails['NRACESSONFCE'];
                $nfceCode        = $saleDetails['NRNOTAFISCALCE'];
                $xml             = $saleDetails['DSARQXMLNFCE'];
                $xml             = str_replace('"', "'", $xml);
                $message = "";
                
                $generateQrCodeCat = $this->orderService->generateQrCodeCat($cdfilial, $cdcaixa, $nrseqvenda, $dtentrvenda);
                if ($generateQrCodeCat['error']) {
                    $this->util->saveLog("Erro ao gerar QrCode de liberação da catraca. {$generateQrCodeCat['message']}", 'Error', 'CLOSE-');
                }

                $generateXmlCat = $this->orderService->generateXmlCatraca($cdfilial, $order['DSCOMANDA'], $order['NRCOMANDA'], 'L');
                if ($generateXmlCat['error']) {
                    $this->util->saveLog("Erro ao gerar XML de liberação da catraca. {$generateXmlCat['message']}", 'Error', 'CLOSE-');
                }

                if (!empty($errorFile)) {
                    $file = json_decode(file_get_contents($errorPath . $errorFile[0]), true);
                    $message = $file['error'];
                }
            } else if (!empty($errorFile)) {
                $file = json_decode(file_get_contents($errorPath . $errorFile[0]), true);
                $validate = $this->orderService->validateOrder($order);
                $message = $validate['error'] ? $validate['message'] : $file['error'];
                $response->addMessage(new Message("Erro no fechamento do pedido: " . $message, self::ERROR_MESSAGE));

            } else {
                $response->addMessage(new Message("O pedido ($cdfilial-$nrcomanda) está em processo de fechamento, tente novamente mais tarde.", self::GENERAL_MESSAGE));
            }

            $storeName = $this->util->getStoreName($cdfilial, $cdcaixa, $order['CDLOJA']);
            
            $fidelityPoints = null;
            $personalCredit = null;
            
            // if (!empty($order['CDCONSUMIDOR']) && $storeName['IDEXTCONSONLINE'] == "S") {
            if (!empty($order['CDCONSUMIDOR'])) {
                $personalCredit = $this->util->getConsumerBalance($order['CDCLIENTE'], $order['CDCONSUMIDOR'], '001', 10, $cdfilial);
                if($storeName['IDEXTCONSONLINE'] == "S"){
                    $fidelityPoints = $this->util->getConsumerFidelity($order['CDCLIENTE'], $order['CDCONSUMIDOR']);
                    if (!empty($fidelityPoints)) $fidelityPoints = $fidelityPoints['value'];
                }
            }
            
            $response->addDataSet(
                new DataSet(
                    '',
                    array(
                        'message'        => $message,
                        'consumerName'   => $order['NMCONSUMIDOR'],
                        'saleDate'       => $dtentrvenda,
                        'fidelityPoints' => $fidelityPoints,
                        'personalCredit' => $personalCredit,
                        'saleSeq'        => $nrseqvenda,
                        'posCode'        => $cdcaixa,
                        'posName'        => $storeName['NMCAIXA'],
                        'subsidiaryCode' => $cdfilial,
                        'subsidiaryName' => $storeName['NMFILIAL'],
                        'storeCode'      => $order['CDLOJA'],
                        'storeName'      => $storeName['NMLOJA'],
                        'couponFooter'   => $storeName['DSOBSCUPFIS'],
                        'status'         => $status,
                        'pagerId'        => $order['CDSENHAPED'],
                        'nfceCode'       => $nfceCode,
                        'nfceAccessCode' => $nfceAccessCode,
                        'qrCodeCat'      => $generateQrCodeCat['qrCodeCat'],
                        'qrCode'         => $qrCode,
                        //'xml'          => (($status == "R") || ($status == "C")) ? "" : $xml
                        'xml'            => $xml
                    )
                )
            );
        } catch (\Exception $e) {
            $message = 'Erro ao buscar o fechamento da comanda. ' . $e->getMessage();
            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
        }
    }

    public function recuperaPedidos($function)
    {
        $notas = array();
        $directory = "{$this->logPath}/$function/Pending/";

        if (is_dir($directory)) {
            $scanned_directory = array_diff(scandir($directory), array('..', '.'));

            foreach ($scanned_directory as $file) {
                $fileData = json_decode(file_get_contents("$directory/$file"), true);
                array_push($notas, $fileData);
            }
        }

        return $notas;
    }

    public function finish($function, $dataProcessamento, $path, $fileName, $removeFile = true, $execTime = null, $transmitTime = null)
    {
        $processedPath = "{$this->logPath}/$function/Processed/" . date("Y/m/d") . "/";

        if (!is_dir($processedPath)) {
            try {
                if($this->util->isRunningOnWindows()) {
                    mkdir($processedPath, 0777, true);
                } else {
                    mkdir($processedPath, 0775, true);
                } 
            } catch (\Exception $e) {
                $this->util->saveLog($e->getMessage(), 'Execution/' . date("Y/m"), 'FOLDERS-');
            }
        }

        $file = file_get_contents($path . $fileName);
        $file = json_decode($file, true);
        $file['execTime'] = $execTime;
        $file['transmitTime'] = $transmitTime;

        file_put_contents($processedPath . $fileName, json_encode($file));

        if ($removeFile)
            unlink(realpath($path . $fileName));
    }

    public function error($function, $dataProcessamento, $path, $fileName, $message)
    {
        $errorPath = "{$this->logPath}/$function/Error/" . date('Y') . "/" . date('m') . "/" . date('d') . "/";

        if(!is_dir($errorPath)) {
            if($this->util->isRunningOnWindows()) {
                mkdir($errorPath, 0777, true);
            } else {
                mkdir($errorPath, 0775, true);
            }
        }

        $file = file_get_contents($path . $fileName);
        $file = json_decode($file, true);
        $file['error'] = $message;

        file_put_contents($errorPath . $fileName, json_encode($file));

        unlink(realpath($path . $fileName));
    }

    public function updateControlFile($alreadyRunning)
    {
        $value = array(
            'expiresIn' => time() + 60,
            'lock'      => $alreadyRunning
        );
        
        if(!file_exists($this->logPath)) {
            if($this->util->isRunningOnWindows()) {
                mkdir($this->logPath, 0777, true);
            } else {
                mkdir($this->logPath, 0775, true);
            }
        }

        file_put_contents("{$this->logPath}" . DIRECTORY_SEPARATOR . ".." . DIRECTORY_SEPARATOR . "CLOSE-CTRL-FILE.dat", base64_encode(json_encode($value)));
        //file_put_contents("{$this->logPath}" . DIRECTORY_SEPARATOR  . "CLOSE-CTRL-FILE.dat", base64_encode(json_encode($value)));

    }

    public function updateCancelControlFile($alreadyRunning)
    {
        $value = array(
            'expiresIn' => time() + 60,
            'lock'      => $alreadyRunning
        );

        file_put_contents("{$this->logPath}" . DIRECTORY_SEPARATOR . ".." . DIRECTORY_SEPARATOR . "CANCEL-CTRL-FILE.dat", base64_encode(json_encode($value)));
    }

    public function V3_getStorePrinters(Request $request, Response $response){
        try{
            $inputJSON = file_get_contents('php://input');
            $input = json_decode($inputJSON, true);
            if (isset($input['subsidiaryCode'])) $subsidiaryCode = $input['subsidiaryCode'];
            else throw new \Exception('Filial não informada.');

            if (isset($input['storeCode'])) $store = $input['storeCode'];
            else throw new \Exception('Loja não informadas.');
            
            $returnStorePrinters = $this->orderService->getStorePrinters($subsidiaryCode, $store);

            $printers = array();
            foreach($returnStorePrinters as $storePrinter){
                $printer = array(
                    'subsidiaryCode' => $storePrinter["CDFILIAL"],
                    'storeCode' => $storePrinter["CDLOJA"],
                    'seqStorePrinter' => $storePrinter["NRSEQIMPRLOJA"],
                    'printerCode' => $storePrinter["CDIMPRESSORA"],
                    'name' => $storePrinter["NMIMPRLOJA"],
                    'port' => $storePrinter["CDPORTAIMPR"],
                    'ipPrinter' => $storePrinter["DSIPIMPR"],
                    'ipPonte' => $storePrinter["DSIPPONTE"]
                );
                $printers[] = $printer;
            }

            $response->addDataSet(new DataSet('', array('printers' => $printers)));            
        } catch (\Exception $e) {
            $this->util->saveLog($e->getMessage(),'Error','STORE-PRINTERS-');
            $message = 'Erro ao recuperar as impressoras de uma loja. ' . $e->getMessage();
            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
        }
    }

    private function correctQuantityChildItems(&$childItems){
        
        for($i = 0; $i < count($childItems); $i++){
            for($j = 0; $j < count($childItems); $j++){
                if($childItems[$i]['productCode'] == $childItems[$j]['productCode'] && $i != $j){
                    if($childItems[$i]['subProductsGroupPrint'] == "S" && $childItems[$j]['subProductsGroupPrint'] == "S"){
                        $childItems[$i]['insertItComVen']  = $childItems[$j]['insertItComVen'] = true;
                        $childItems[$i]['insertItComEst'] = true;
                        $childItems[$j]['insertItComEst'] = false;
                        if(!empty($childItems[$i]['quantityItComandaEst'])){
                            $childItems[$i]['quantityItComandaEst'] += $childItems[$j]['quantity'];   
                        } else {
                            $childItems[$i]['quantityItComandaEst'] = $childItems[$i]['quantity'] + $childItems[$j]['quantity'];
                        }
                    } else if($childItems[$i]['subProductsGroupPrint'] == "S" && $childItems[$j]['subProductsGroupPrint'] == "N"){
                        $childItems[$i]['insertItComVen'] = true; 
                        $childItems[$i]['insertItComEst'] = true;
                        $childItems[$j]['insertItComVen'] = false;
                        $childItems[$j]['insertItComEst'] = false;
                        if(!empty($childItems[$i]['quantityItComandaEst'])){
                            $childItems[$i]['quantityItComandaEst'] += $childItems[$j]['quantity'];   
                        } else {
                            $childItems[$i]['quantityItComandaEst'] = $childItems[$i]['quantity'] + $childItems[$j]['quantity'];
                        }
                    } else if($childItems[$i]['subProductsGroupPrint'] == "N" && $childItems[$j]['subProductsGroupPrint'] == "N"){
                        $childItems[$i]['insertItComVen'] = false; 
                        $childItems[$i]['insertItComEst'] = true;
                        $childItems[$j]['insertItComVen'] = false;
                        $childItems[$j]['insertItComEst'] = false;
                        if(!empty($childItems[$i]['quantityItComandaEst'])){
                            $childItems[$i]['quantityItComandaEst'] += $childItems[$j]['quantity'];   
                        } else {
                            $childItems[$i]['quantityItComandaEst'] = $childItems[$i]['quantity'] + $childItems[$j]['quantity'];
                        }
                    }
                }
            }
        }
    }

    public function closePos(Request $request, Response $response) {
		ini_set('memory_limit', '512M');
		set_time_limit(100);
        try {
            $inputJSON = file_get_contents('php://input');
            $input = json_decode($inputJSON, true);

            if (isset($input['subsidiaryCode'])) 
              $cdfilial = $input['subsidiaryCode'];
            else 
              throw new \Exception('Filial não informada.');

            if (isset($input['posCode'])) 
              $cdcaixa = $input['posCode'];
            else 
              throw new \Exception('Caixa não informado.');


            $usesOpeningLock = 'N';
            if (isset($input['usesOpeningLock']) && $input['usesOpeningLock'] == 'S') 
              $usesOpeningLock = 'S';
            $retornoFechamento = $this->orderService->closePos($cdfilial, $cdcaixa,  $usesOpeningLock);

            $response->addDataSet(
                new DataSet(
                    '',
                    array(
                        'closing'       => $retornoFechamento
                    )
                )
            );
            
        } catch (\Exception $e) {
            $message = 'Erro ao realizar o fechamento do caixa. ' . $e->getMessage();
            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
        }
    }

    public function openPos(Request $request, Response $response) {
		ini_set('memory_limit', '512M');
		
        try {
            $inputJSON = file_get_contents('php://input');
            $input = json_decode($inputJSON, true);

            if (isset($input['subsidiaryCode'])) 
                $cdfilial = $input['subsidiaryCode'];
            else 
                throw new \Exception('Filial não informada.');

            if (isset($input['posCode'])) $cdcaixa = $input['posCode'];
            else throw new \Exception('Caixa não informado.');
            
            $retornoAbertura = $this->orderService->openPos($cdfilial, $cdcaixa);

            $response->addDataSet(
                new DataSet(
                    '',
                    array(
                        'opening'       => $retornoAbertura
                    )
                )
            );
            
        } catch (\Exception $e) {
            $message = 'Erro ao realizar a abertura do caixa. ' . $e->getMessage();
            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
        }
    }

    public function getPosStatus(Request $request, Response $response) {
		ini_set('memory_limit', '512M');
		
        try {
            $inputJSON = file_get_contents('php://input');
            $input = json_decode($inputJSON, true);

            if (isset($input['subsidiaryCode'])) 
                $cdfilial = $input['subsidiaryCode'];
            else 
                throw new \Exception('Filial não informada.');

            if (isset($input['posCode'])) $cdcaixa = $input['posCode'];
            else throw new \Exception('Caixa não informado.');
            
            $retornoAbertura = $this->orderService->getPosStatus($cdfilial, $cdcaixa);

            $response->addDataSet(
                new DataSet(
                    '',
                    array(
                        'status'       => $retornoAbertura
                    )
                )
            );
            
        } catch (\Exception $e) {
            $message = 'Erro ao retornar o status do caixa. ' . $e->getMessage();
            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
        }
    }

    public function cancelSale(Request $request, Response $response){
        try{
            $inputJSON = file_get_contents('php://input');
            $input = json_decode($inputJSON, true);
            if (isset($input['subsidiaryCode'])) $subsidiaryCode = $input['subsidiaryCode'];
            else throw new \Exception('Filial não informada.');

            if (isset($input['posCode'])) $posCode = $input['posCode'];
            else throw new \Exception('Caixa não informado.');

            if (isset($input['nfceCode'])) $nfceCode = $input['nfceCode'];
            else throw new \Exception('Número da Nota Fiscal não informado.');

            if (isset($input['saleSeq'])) $saleSeq = $input['saleSeq'];
            else $saleSeq = null;

            $caixaInfo = $this->orderService->getCaixaInfo($subsidiaryCode, $posCode);

            if(empty($caixaInfo)) throw new \Exception("Caixa não encontrado.");

            if($caixaInfo['IDTPEMISSAOFOS'] == "FNC") $nfceCode = str_pad($input['nfceCode'], 9, '0', STR_PAD_LEFT);

            $returnSaleRouter = $this->orderService->getSaleRouter($subsidiaryCode, $posCode, $saleSeq, $nfceCode);

            if(empty($returnSaleRouter)) throw new \Exception("Venda não encontrada. Verifique os dados de entrada.");

            if($returnSaleRouter['IDSITUVENDA'] == "C") throw new \Exception("Não é possivel realizar o cancelamento da venda pois a mesma já foi cancelada.");

            if ($returnSaleRouter['IDSITUVENDA'] == "P") throw new \Exception("A venda encontra-se pendente no sistema (gestão de vendas)", 1000);

            $dtvenda = new \DateTime($returnSaleRouter['DTVENDA']);
            $diff = $dtvenda->diff(new \DateTime());
            $saleCancelTime = $this->maxSaleCancelTime;
            if ($diff->i >= $saleCancelTime && $returnSaleRouter['IDSTATUSNFCE'] == "A") throw new \Exception("A venda não pode ser cancelada. (Execedeu os $saleCancelTime minutos)", 1000);
            
            if($caixaInfo["IDHABCAIXAVENDA"] != "TAA" && $caixaInfo["IDHABCAIXAVENDA"] != "APC" && $returnSaleRouter['IDTPEMISVEND'] == "S" && (empty($returnSaleRouter['ENDVENDAROUTER']) || !isset($returnSaleRouter['ENDVENDAROUTER']))){
                throw new \Exception("Cancelamento de venda não permitido. Caixa não vinculado a router.");
            }
            
            $nrseqvenda = $returnSaleRouter['NRSEQVENDA'];
            $cdoperador = $returnSaleRouter['CDOPERADOR'];
            $cdsupervisor = $returnSaleRouter['CDSUPERVISOR'];
            $cancelamento = $this->orderService->cancelOrder($subsidiaryCode, $posCode, $nrseqvenda, $cdoperador, $cdsupervisor);

            if (!empty($cancelamento['mensagemNfce']))
                        $cancelamento['message'] .= $cancelamento['mensagemNfce'];

            if ($cancelamento['error']){
                $return = array(
                        'mensagemNfce' => $cancelamento['mensagemNfce'],
                        'dadosImpressao' => $cancelamento['dadosImpressao']
                );
            }
            //throw new \Exception($cancelamento['message']);

            $response->addDataSet(new DataSet('', array('cancel' => isset($return) ? $return : $cancelamento)));            
        } catch (\Exception $e) {
            $this->util->saveLog($e->getMessage(),'Error','CANCEL-SALE-');
            $message = 'Erro ao cancelar a venda. ' . $e->getMessage();
            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
        }
    }

    public function getOrderData(Request $request, Response $response) {
		ini_set('memory_limit', '512M');
		
        try {
            $inputJSON = file_get_contents('php://input');
            $input = json_decode($inputJSON, true);

            if (isset($input['subsidiaryCode'])) 
                $cdfilial = $input['subsidiaryCode'];
            else 
                throw new \Exception('Filial não informada.');

            if (isset($input['saleDescription'])) 
                $saleDescription = $input['saleDescription'];
            else 
                $dscomanda = null;
            
            if (isset($input['saleCode'])) 
                $nrcomanda = $input['saleCode'];
            else 
                $nrcomanda = null;
            
            $cdMifare = $this->orderService->getOrderDscomandaFos($saleDescription, $cdfilial);
            if(!empty($cdMifare)){
                $dscomanda = str_pad($cdMifare['DSCOMANDAFOS'], 10, '0', STR_PAD_LEFT);
            }else{
                $dscomanda = str_pad($saleDescription, 10, '0', STR_PAD_LEFT);
            }
            if(empty($dscomanda) && empty($nrcomanda)) throw new \Exception("Comanda não informada.");
            
            if ($dscomanda != "0000000000" && !empty($dscomanda)){
                $orderByDscomanda = $this->orderService->getOrderDscomanda($dscomanda, $cdfilial);
                if(empty($orderByDscomanda)) throw new \Exception("Comanda não encontrada.");
                $nrcomanda = $orderByDscomanda['NRCOMANDA'];
            }

            $order = $this->orderService->getOrder($cdfilial, $nrcomanda);

            if(empty($order)) throw new \Exception("Comanda não encontrada!");

            $clientCode = $order['CDCLIENTE'];
            $consumerCode = $order['CDCONSUMIDOR'];
            $storeCode = $order['CDLOJA'];
            $tableNumber = $order['NRMESA'];
            
            $return = $this->orderService->getProductsToRecalculateEx($cdfilial, $nrcomanda, $storeCode);
            if (!$return['error']) {
                $productsToRecalculate = $return['data'];
                $this->recalculateProductsPrice($productsToRecalculate, $cdfilial, $tableNumber, '01', $storeCode, $clientCode, $consumerCode, true);
                // $positionData['items'] = $productsData;
            }
            
            $return = $this->orderService->getCombosToRecalculateEx($cdfilial, $nrcomanda, $storeCode);
            if (!$return['error']) {
                $combosToRecalulate = $return['data'];
                $this->recalculateCombos($combosToRecalulate, $cdfilial, $storeCode, $clientCode, $consumerCode, true);
                // $positionData['items'] = array_merge($positionData['items'], $combosData);
            }

            $items = $this->orderService->getOrderItems($order['NRVENDAREST'], $nrcomanda, $cdfilial);
            $itemsValue = 0;
            foreach($items as $item){
                $itemData = array(
                    'number' => $item['NRPRODCOMVEN'],
                    'productName'       => $item['NMPRODUTO'],
                    'unit'              => $item['SGUNIDADE'],
                    'productCode'       => $item['CDPRODUTO'],
                    'familyCode'        => $item['CDFAMILISALD'],
                    'quantity'          => $this->util->float($item['QTPRODCOMVEN']),
                    'unitValue'         => $this->util->removeDecimal($item['VRPRECCOMVEN']),
                    'discount'          => $this->util->removeDecimal($item['VRDESCCOMVEN']),
                    'addition'          => $this->util->removeDecimal($item['VRACRCOMVEN']),
                    'itemValue'         => $this->orderService->calculaTotalItem($item),
                );
                $itemsValue += $itemData['itemValue'];
                $order['items'][] = $itemData;
            }
            $itemsValue = $this->orderService->calculaValorDaVenda($items)['total'];
            
            $payments = $this->orderService->getOrderPayments($cdfilial, $order['NRVENDAREST']);
            $order['payments'] = [];
            $totalPaid = 0;
            foreach($payments as $payment){
                if($payment['IDTIPOMOVIVEDLV'] == "E"){
                    $paymentData = array(
                        'paymentCode'       => $payment['CDTIPORECE'],
                        'paymentName'       => $payment['NMTIPORECE'],
                        'value'             => $this->util->float($payment['VRMOVIVEND']),
                        'nsu'               => $payment['CDNSUHOSTTEF']
                    );
                    $totalPaid += $paymentData['value'];
                    $order['payments'][] = $paymentData;
                }
            }

            $serviceTax = $order['VRCOMISVENDE'];
            if($serviceTax == ".000" || empty($serviceTax)){
                $storeServiceTax = $this->util->getStoreServiceTax($cdfilial, $order['CDLOJA']);

                if(!empty($storeServiceTax) && $storeServiceTax['IDCOMISVENDA'] == "S"){
                    $perc = !empty($storeServiceTax['VRCOMISVENDA']) ? $storeServiceTax['VRCOMISVENDA'] : 0;
                    $serviceTax = ($itemsValue * $perc) / 100;
                } else if(!empty($storeServiceTax) && $storeServiceTax['IDCOMISVENDA'] == "N"){
                    $serviceTax = null;
                }
            }

            $consumers = array(array(
                'clientCode' => $order['CDCLIENTE'],
                'consumerCode' => $order['CDCONSUMIDOR'],
                'consumerId' => $order['CDIDCONSUMID'],
                'consumerName' => $order['NMCONSUMIDOR']
            ));
            $orderConsumers = $this->orderService->getJoinedOrdersConsumers($cdfilial, $order['DSCOMANDA']);
            if(!empty($orderConsumers)) {
                foreach($orderConsumers as $consumer){
                    $consumerData = array(
                        'clientCode' => $consumer['CDCLIENTE'],
                        'consumerCode' => $consumer['CDCONSUMIDOR'],
                        'consumerId' => $consumer['CDIDCONSUMID'],
                        'consumerName' => $consumer['NMCONSUMIDOR']
                    );
                    if(!in_array($consumerData, $consumers, true)) array_push($consumers, $consumerData);
                }
            }

            $response->addDataSet(
                new DataSet(
                    '',
                    array(
                        'order' => array(
                            'subsidiaryCode' => $order['CDFILIAL'],
                            'storeCode'      => $order['CDLOJA'],
                            'posCode'        => $order['CDCAIXA'],
                            'pagerId'        => $order['CDSENHAPED'],
                            'saleDescription'=> $order['DSCOMANDA'],
                            'saleCode'       => $order['NRCOMANDA'],
                            'tableNumber'    => $order['NRMESA'],
                            'people'         => $order['NRPESMESAVEN'],
                            'sellerCode'     => $order['CDVENDEDOR'],
                            'operatorCode'   => $order['CDOPERADOR'],
                            'origin'         => $order['IDORGCMDVENDA'],
                            'status'         => $order['IDSTCOMANDA'],
                            'openingDate'    => $order['DTHRABERMESA'],
                            // 'clientCode'     => $order['CDCLIENTE'],
                            // 'consumerCode'   => $order['CDCONSUMIDOR'],
                            'consumers'      => $consumers,
                            'items'          => $order['items'],
                            'productsValue'  => $itemsValue,
                            'additionialValue' => $this->util->removeDecimal($order['VRACRCOMANDA']),
                            'serviceTax'     => ($serviceTax == null) ? null : $this->util->float($serviceTax),
                            'totalValue'     => $this->util->removeDecimal($itemsValue + $order['VRACRCOMANDA'] + $serviceTax),
                            'orderCPF'       => $order['NRINSCRCONS'],
                            'payments'       => $order['payments'],
                            'totalPaid'     => $totalPaid
                        ), 
                    )
                )
            );
            
        } catch (\Exception $e) {
            $message = 'Erro ao buscar os dados da comanda. ' . $e->getMessage();
            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
        }
    }

    public function editServiceTax(Request $request, Response $response) {
		ini_set('memory_limit', '512M');
		$this->util->beginTransaction();
        try {
            $inputJSON = file_get_contents('php://input');
            $input = json_decode($inputJSON, true);

            if (isset($input['subsidiaryCode'])) 
                $cdfilial = $input['subsidiaryCode'];
            else 
                throw new \Exception('Filial não informada.');
            
            if (isset($input['saleCode'])) 
                $nrcomanda = $input['saleCode'];
            else 
                throw new \Exception('Comanda não informada.');

            if (isset($input['serviceTax'])) 
                $vrcomisvende = $input['serviceTax'];
            else 
                throw new \Exception('Taxa de serviço não informada.');

            $order = $this->orderService->getOrder($cdfilial, $nrcomanda);

            if(empty($order)) throw new \Exception("Comanda não encontrada!");

            if($order['IDSTCOMANDA'] == "X") throw new \Exception("Essa comanda foi fechada e não pode ter o valor da taxa de serviço alterada.");
            else if($order['IDSTCOMANDA'] == "4") throw new \Exception("Essa comanda foi cancelada e não pode ter o valor da taxa de serviço alterada.");
            
            $this->orderService->editServiceTax($cdfilial, $nrcomanda, $vrcomisvende);
        
            $response->addMessage(new Message("O pedido ($cdfilial-$nrcomanda) teve sua taxa de serviço alterada para $vrcomisvende.", self::GENERAL_MESSAGE));
            $this->util->commit();
        } catch (\Exception $e) {
            $this->util->rollBack();
            $this->util->saveLog($e->getMessage(),'Error','EDIT-SERVICE-TAX-');
            $message = 'Erro ao editar o valor da taxa de serviço. ' . $e->getMessage();
            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
        }
    }

    private function deleteOrder($nrvendarest, $cdfilial, $nrcomanda) {
        $this->util->deleteItcomandaest($nrvendarest, $cdfilial, $nrcomanda);
        $this->util->deleteItcomandaven($nrvendarest, $cdfilial, $nrcomanda);
        $this->util->deleteComandaven($nrvendarest, $cdfilial, $nrcomanda);
        $this->util->deleteMovCaixaDlv($nrvendarest, $cdfilial);
        $this->util->deleteVendarest($nrvendarest, $cdfilial);
    }

    public function joinOrders(Request $request, Response $response) {
		ini_set('memory_limit', '512M');

		$this->util->beginTransaction();
        try {
            $inputJSON = file_get_contents('php://input');
            $input = json_decode($inputJSON, true);

            if (isset($input['subsidiaryCode'])) 
                $cdfilial = $input['subsidiaryCode'];
            else 
                throw new \Exception('Filial não informada.');
            
            if (isset($input['saleCode'])) 
                $nrcomandaPai = $input['saleCode'];
            else 
                throw new \Exception('Comanda não informada.');

            if (isset($input['childSaleCode'])) 
                $nrcomandaFilha = $input['childSaleCode'];
            else 
                throw new \Exception('Comanda a ser unida não informada.');

            if($nrcomandaPai == $nrcomandaFilha) throw new \Exception("Comandas a serem unidas são a mesma comanda.");

            $order = $this->orderService->getOrder($cdfilial, $nrcomandaPai);
            $childOrder = $this->orderService->getOrder($cdfilial, $nrcomandaFilha);

            if(empty($order)) throw new \Exception("Comanda não encontrada!");
            if(empty($childOrder)) throw new \Exception("Comanda a ser unida não encontrada!");

            // valida se a junção de comandas pode ser feita
            if($order['IDSTCOMANDA'] == "X" || $childOrder['IDSTCOMANDA'] == "X") throw new \Exception("A comanda ".($childOrder['IDSTCOMANDA'] == "X" ? $nrcomandaFilha : $nrcomandaPai)." foi fechada e não pode ser unida à outra comanda.");
            else if($order['IDSTCOMANDA'] == "4" || $childOrder['IDSTCOMANDA'] == "4") throw new \Exception("A comanda ".($childOrder['IDSTCOMANDA'] == "4" ? $nrcomandaFilha : $nrcomandaPai)." foi cancelada e não pode ser unida à outra comanda.");
            
            // valida se a comanda a ser juntada não está vinculada a outra comanda
            if($childOrder['DSCOMANDAPRI'] != null || $childOrder['DSCOMANDAPRI'] != "") throw new \Exception("A comanda ".$nrcomandaFilha." já está unida a outra comanda ".$childOrder['DSCOMANDAPRI']." .");

            // atualiza a comanda filha com idstcomanda cancelado e dscomandapri = dscomanda da comanda pai
            $this->orderService->updateChildOrder($cdfilial, $nrcomandaFilha, $order['DSCOMANDA'], '4');
        
            // select nos itens da comanda filha
            $childOrderItems = $this->orderService->getOrderItems($childOrder['NRVENDAREST'], $nrcomandaFilha, $cdfilial);

            // select nos itens da comanda pai
            $orderItems = $this->orderService->getOrderItems($order['NRVENDAREST'], $nrcomandaPai, $cdfilial);
            
            $itemsSeq = intval( end($orderItems)['NRPRODCOMVEN'] ) + 1;

            foreach($childOrderItems as $item){
                $item['CDFILIAL']     = $cdfilial;
                $item['DSCOMANDAORI'] = $childOrder['DSCOMANDA'];
                $item['NRCOMANDAORI'] = $nrcomandaFilha;
                $item['NRCOMANDA']    = $nrcomandaPai;
                $item['NRVENDAREST']  = $order['NRVENDAREST'];
                $item['NRPRODCOMORI'] = $item['NRPRODCOMVEN'];
                $item['NRPRODCOMVEN']  = str_pad($itemsSeq, 6, '0', STR_PAD_LEFT);
                $this->orderService->addItComandaven($item);
                $itemsSeq++;
            }
            $this->util->commit();
            $response->addMessage(new Message("A comanda ($cdfilial-$nrcomandaFilha) foi unida à comanda ($cdfilial-$nrcomandaPai).", self::GENERAL_MESSAGE));
        } catch (\Exception $e) {
            $this->util->rollBack();
            $this->util->saveLog($e->getMessage(),'Error','JOIN-ORDERS-');
            $message = 'Erro ao realizar a junção de comandas. ' . $e->getMessage();
            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
        }
    }

    public function splitOrders(Request $request, Response $response) {
		ini_set('memory_limit', '512M');

		$this->util->beginTransaction();
        try {
            $inputJSON = file_get_contents('php://input');
            $input = json_decode($inputJSON, true);

            if (isset($input['subsidiaryCode'])) 
                $cdfilial = $input['subsidiaryCode'];
            else 
                throw new \Exception('Filial não informada.');
            
            if (isset($input['saleCode'])) 
                $nrcomandaPai = $input['saleCode'];
            else 
                throw new \Exception('Comanda não informada.');

            if (isset($input['childSaleCode'])) 
                $nrcomandaFilha = $input['childSaleCode'];
            else 
                throw new \Exception('Comanda a ser desvinculada não informada.');

            if($nrcomandaPai == $nrcomandaFilha) throw new \Exception("Comandas a serem unidas são a mesma comanda.");

            $order = $this->orderService->getOrder($cdfilial, $nrcomandaPai);
            $childOrder = $this->orderService->getOrder($cdfilial, $nrcomandaFilha);

            if(empty($order)) throw new \Exception("Comanda não encontrada!");
            if(empty($childOrder)) throw new \Exception("Comanda a ser desvinculada não encontrada!");

            // valida se o desvinculo de comandas pode ser feito
            if($order['IDSTCOMANDA'] == "X" || $childOrder['IDSTCOMANDA'] == "X") throw new \Exception("A comanda ".($childOrder['IDSTCOMANDA'] == "X" ? $nrcomandaFilha : $nrcomandaPai)." foi fechada e não pode ser desvinculada à outra comanda.");
            else if($order['IDSTCOMANDA'] == "4") throw new \Exception("A comanda $nrcomandaPai) foi cancelada e não pode ser desvinculada à outra comanda.");
            
            // valida se a comanda a ser desvinculada está vinculada a outra comanda
            if(!isset($childOrder['DSCOMANDAPRI'])) throw new \Exception("A comanda ".$nrcomandaFilha." não está unida a outra comanda.");

            $idcontroprod = $this->getIdcontroprod($childOrder['CDLOJA'], $cdfilial);
            $idstcomanda  = '3';
            if ($idcontroprod === 'S') $idstcomanda = '1';
            // atualiza a comanda filha com idstcomanda 1 e dscomandapri = vazio ( quebra vinculo com comanda pai )
            $this->orderService->updateChildOrder($cdfilial, $nrcomandaFilha, '', $idstcomanda);

            // deleta os itens da comanda pai que tem referência a comanda filha ( tira os itens da comanda pai que são da comanda filha )
            $this->orderService->deleteJoinOrderItems($cdfilial, $order['NRVENDAREST'], $nrcomandaPai, $childOrder['DSCOMANDA'], $nrcomandaFilha);

            $this->util->commit();
            $response->addMessage(new Message("A comanda ($cdfilial-$nrcomandaFilha) foi desvinculada à comanda ($cdfilial-$nrcomandaPai).", self::GENERAL_MESSAGE));
        } catch (\Exception $e) {
            $this->util->rollBack();
            $this->util->saveLog($e->getMessage(),'Error','UNJOIN-ORDERS-');
            $message = 'Erro ao realizar o desvinculo de comandas. ' . $e->getMessage();
            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
        }
    }

    public function cleanOrder(Request $request, Response $response) {
		ini_set('memory_limit', '512M');

		$this->util->beginTransaction();
        try {
            $inputJSON = file_get_contents('php://input');
            $input = json_decode($inputJSON, true);

            if (isset($input['subsidiaryCode'])) 
                $cdfilial = $input['subsidiaryCode'];
            else 
                throw new \Exception('Filial não informada.');
            
            if (isset($input['saleCode'])) 
                $nrcomanda = $input['saleCode'];
            else 
                throw new \Exception('Comanda não informada.');

            $order = $this->orderService->getOrder($cdfilial, $nrcomanda);

            if(empty($order)) throw new \Exception("Comanda não encontrada!");

			$childOrders = $this->orderService->getChildOrders($cdfilial, $order['DSCOMANDA']);

			foreach($childOrders as $child){
				$this->deleteOrder($child['NRVENDAREST'], $cdfilial, $child['NRCOMANDA']);
			}

            $this->deleteOrder($order['NRVENDAREST'], $cdfilial, $nrcomanda);

            $this->util->commit();
            $response->addMessage(new Message("A comanda ($cdfilial-$nrcomanda) e as comandas vinculadas a mesma foram deletadas do sistema.", self::GENERAL_MESSAGE));
        } catch (\Exception $e) {
            $this->util->rollBack();
            $this->util->saveLog($e->getMessage(),'Error','CLEAN-ORDER-');
            $message = 'Erro ao limpar os dados da comanda. ' . $e->getMessage();
            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
        }
    }

    public function cancelPayments(Request $request, Response $response) {
		ini_set('memory_limit', '512M');

		$this->util->beginTransaction();
        try {
            $inputJSON = file_get_contents('php://input');
            $input = json_decode($inputJSON, true);

            if (isset($input['subsidiaryCode'])) 
                $cdfilial = $input['subsidiaryCode'];
            else 
                throw new \Exception('Filial não informada.');
            
            if (isset($input['saleCode'])) 
                $nrcomanda = $input['saleCode'];
            else 
                throw new \Exception('Comanda não informada.');

            // if (isset($input['nsu'])) 
            //     $nsu = $input['nsu'];
            // else 
            //     throw new \Exception('NSU do pagamento não informado.');

            $order = $this->orderService->getOrder($cdfilial, $nrcomanda);

            if(empty($order)) throw new \Exception("Comanda não encontrada!");

            if($order['IDSTCOMANDA'] == "X") throw new \Exception("Não é possível cancelar os pagamentos do pedido, pois o mesmo já foi fechado.");

			$payments = $this->orderService->getOrderPayments($cdfilial, $order['NRVENDAREST']);
            
            if(empty($payments)) throw new \Exception("Não existem pagamentos para esse pedido.");

            // $foundNsu = false;
			// foreach($payments as $payment){
			// 	if($payment['CDNSUHOSTTEF'] == $nsu){
            //         $foundNsu = true;
            //         if($payment['IDTIPOMOVIVEDLV'] == "C") throw new \Exception("Pagamento informado já foi cancelado.");
            //     }
			// }
            // if($foundNsu == false) throw new \Exception("Pagamento não encontrado.");

            $this->orderService->cancelPayments($cdfilial, $order['NRVENDAREST']);

            $this->util->commit();
            $response->addMessage(new Message("Os pagamentos da comanda ($cdfilial-$nrcomanda) foram cancelados.", self::GENERAL_MESSAGE));
        } catch (\Exception $e) {
            $this->util->rollBack();
            $this->util->saveLog($e->getMessage(),'Error','CANCEL-PAYMENTS-');
            $message = 'Erro ao cancelar os pagamentos. ' . $e->getMessage();
            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
        }
    }

    public function editConsumer(Request $request, Response $response) {
		ini_set('memory_limit', '512M');
		$this->util->beginTransaction();
        try {
            $inputJSON = file_get_contents('php://input');
            $input = json_decode($inputJSON, true);

            if (isset($input['subsidiaryCode'])) 
                $cdfilial = $input['subsidiaryCode'];
            else 
                throw new \Exception('Filial não informada.');
            
            if (isset($input['saleCode'])) 
                $nrcomanda = $input['saleCode'];
            else 
                throw new \Exception('Comanda não informada.');

            if (isset($input['clientCode'])) 
                $cdcliente = $input['clientCode'];
            else 
                throw new \Exception('Código do cliente não informado.');

            if (isset($input['consumerCode'])) 
                $cdconsumidor = $input['consumerCode'];
            else 
                throw new \Exception('Código do consumidor não informado.');

            $order = $this->orderService->getOrder($cdfilial, $nrcomanda);

            if(empty($order)) throw new \Exception("Comanda não encontrada!");

            if($order['IDSTCOMANDA'] == "X") throw new \Exception("Essa comanda foi fechada e não pode ter o consumidor alterado.");
            else if($order['IDSTCOMANDA'] == "4") throw new \Exception("Essa comanda foi cancelada e não pode ter o consumidor alterado.");
            
            $this->orderService->editConsumer($cdfilial, $order['NRVENDAREST'], $cdcliente, $cdconsumidor);
        
            $response->addMessage(new Message("O pedido ($cdfilial-$nrcomanda) teve seu consumidor alterado para ($cdcliente-$cdconsumidor).", self::GENERAL_MESSAGE));
            $this->util->commit();
        } catch (\Exception $e) {
            $this->util->rollBack();
            $this->util->saveLog($e->getMessage(),'Error','EDIT-CONSUMER-');
            $message = 'Erro ao editar o consumidor do pedido. ' . $e->getMessage();
            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
        }
    }

    public function printCancel(Request $request, Response $response) {
		ini_set('memory_limit', '512M');
        try {
            $inputJSON = file_get_contents('php://input');
            $input = json_decode($inputJSON, true);

            if (isset($input['subsidiaryCode'])) 
                $cdfilial = $input['subsidiaryCode'];
            else 
                throw new \Exception('Filial não informada.');
            
            if (isset($input['posCode'])) 
                $cdcaixa = $input['posCode'];
            else 
                throw new \Exception('Caixa não informado.');

            if (isset($input['saleCode'])) 
                $nrcomanda = $input['saleCode'];
            else 
                throw new \Exception('Comanda não informada.');

            $order = $this->orderService->getOrder($cdfilial, $nrcomanda);

            if(empty($order)) throw new \Exception("Comanda não encontrada!");

            if($order['IDSTCOMANDA'] != "4") throw new \Exception("Essa comanda não está cancelada e não pode ter a impressão do cancelamento realizada.");
            
            $items = $this->orderService->getItemsPrintCancel($cdfilial, $nrcomanda, $order['NRVENDAREST']);

            $pos = $this->orderService->getPosPrintCancel($cdfilial, $cdcaixa);

            if(!empty($pos)) $pos = ['IDIMPPEDPRCIGE' => $pos[0]['IDIMPPEDPRCIGE']];
            else $pos = ['IDIMPPEDPRCIGE' => "N"];
            
            $modo = "";
            if(!(strpos($order['IDORGCMDVENDA'], "CMD_") !== false)){
                $modo = 'C';
            } 

            $print = $this->orderService->printCancel($order, $items, $modo, $order['DSCOMANDA'], $pos);

            $text = !empty($print[0]['text']) ? $print[0]['text'] : "";

            $response->addDataSet(
                new DataSet(
                    '',
                    array('text' => $text)));

            if($print['error']) throw new \Exception($print['message']);        
            
            $response->addMessage(new Message("O pedido ($cdfilial-$nrcomanda) teve seu cancelamento impresso.", self::GENERAL_MESSAGE));
        } catch (\Exception $e) {
            $this->util->saveLog($e->getMessage(),'Error','PRINT-CANCEL-');
            $message = 'Erro ao imprimir o cancelamento do pedido. ' . $e->getMessage();
            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
        }
    }
    
    public function getStatusTable(Request $request, Response $response) {
      ini_set('memory_limit', '512M');
      
          try {
              $inputJSON = file_get_contents('php://input');
              $input = json_decode($inputJSON, true);
  
              if(isset($input['subsidiaryCode']))
                  $cdFilial = $input['subsidiaryCode'];
              else
                  throw new \Exception('Código da filial não informado.'); 
  
              if(isset($input['storeCode']))
                  $cdLoja = $input['storeCode'];
              else
                  throw new \Exception('Código da loja não informado.');
              
              if(isset($input['orderType']))
                  $orderType = $input['orderType'];
              else
                  throw new \Exception('Identificador de modalidade não informado.');  
  
              if(isset($input['tableId'])){
                  $nrMesa = str_pad($input['tableId'], 4, 0, STR_PAD_LEFT);
              }else{
                  if($orderType === 'MES')
                      throw new \Exception('Mesa não informada.'); 
              }
  
              if(isset($input['nrComanda'])){
                  $nrComanda = $input['nrComanda'];
                  $nrComanda = str_pad($nrComanda, 10, 0, STR_PAD_LEFT);
              }else{
                  if($orderType === 'CMD')
                      throw new \Exception('Comanda não informada.'); 
              }
  
              if($orderType === 'MES'){          
                  $statusTable = $this->orderService->getStatusTable($cdFilial, $cdLoja, $nrMesa);
                  if($statusTable['IDSTMESAAUX'] != 'D'){
                    $data = $this->orderService->getOrderByNrMesa($cdFilial, $nrMesa);
                  }
  
                  $listStatus = [
                      'O' => [ 
                              "statusId"    => 1 ,
                              "status"      => "Mesa Aberta",
                              "saleCode"    => isset($data['NRCOMANDA']) ? $data['NRCOMANDA'] : null,
                              "saleDescription" => isset($data['NRCOMANDA']) ? $data['DSCOMANDA'] : null,
                              "personalQTD" => isset($data['NRPESMESAVEN']) ? $data['NRPESMESAVEN'] : null

                              ],
                      'D' => [ 
                              "statusId"    => 2 ,
                              "status"      => "Mesa Livre",
                              "saleCode"    => isset($data['NRCOMANDA']) ? $data['NRCOMANDA'] : null,
                              "saleDescription" => isset($data['NRCOMANDA']) ? $data['DSCOMANDA'] : null,
                              "personalQTD" => isset($data['NRPESMESAVEN']) ? $data['NRPESMESAVEN'] : null
                              ],
                      'S' => [ 
                              "statusId"    => 3 ,
                              "status"      => "Conta Fechada",
                              "saleCode"    => isset($data['NRCOMANDA']) ? $data['NRCOMANDA'] : null,
                              "saleDescription" => isset($data['NRCOMANDA']) ? $data['DSCOMANDA'] : null,
                              "personalQTD" => isset($data['NRPESMESAVEN']) ? $data['NRPESMESAVEN'] : null
                              ],
                      'R' => [ 
                              "statusId"    => 4 ,
                              "status"      => "Mesa em Recebimento",
                              "saleCode"    => isset($data['NRCOMANDA']) ? $data['NRCOMANDA'] : null,
                              "saleDescription" => isset($data['NRCOMANDA']) ? $data['DSCOMANDA'] : null,
                              "personalQTD" => isset($data['NRPESMESAVEN']) ? $data['NRPESMESAVEN'] : null
                              ],
                  ];
  
                  $Status = isset($listStatus[$statusTable['IDSTMESAAUX']]) ? $listStatus[$statusTable['IDSTMESAAUX']] : null;
              }
              else{
                  $data = $this->orderService->getStatusCmd($cdFilial, $cdLoja, $nrComanda);
                  $statusCMD = empty($data) ? 2 : 1; 
                
                  $listStatus = [
                      '1' => [ 
                              "statusId"    => 1 ,
                              "status"      => "Comanda Aberta",
                              "saleCode"    => isset($data['NRCOMANDA']) ? $data['NRCOMANDA'] : null ,
                              "saleDescription" => isset($data['NRCOMANDA']) ? $data['DSCOMANDA'] : null,
                              "personalQTD" => '1'
                              ],
                      '2' => [ 
                              "statusId"    => 2 ,
                              "status"      => "Comanda Disponível",
                              "saleCode"    => null,
                              "personalQTD" => '1'
                              ]
                  ];
                  
                  $Status = isset($listStatus[$statusCMD]) ? $listStatus[$statusCMD] : null;
                  
              }
  
              $response->addDataSet(
                     new DataSet(
                          '',
                          array(
                          'status' => $Status,
                          )
                      )
              );
  
      } catch (\Exception $e) {
              $message = 'Erro ao buscar status da mesa/comanda. ' . $e->getMessage();
              $response->addMessage(new Message($message, self::ERROR_MESSAGE));
          }
      }
  
    public function openTable(Request $request, Response $response) {
      ini_set('memory_limit', '512M');
      $this->util->beginTransaction();
  
          try {
              $inputJSON = file_get_contents('php://input');
              $input = json_decode($inputJSON, true);
  
              if(isset($input['subsidiaryCode']))
                  $cdFilial = $input['subsidiaryCode'];
              else
                  throw new \Exception('Código da filial não informado.'); 
  
              if(isset($input['storeCode']))
                  $cdLoja = $input['storeCode'];
              else
                  throw new \Exception('Código da loja não informado.'); 
              
              if(isset($input['personalQTD']))
                  $personalQtd = $input['personalQTD'];
              else
                  throw new \Exception('Mesa não informada.');
              
              if(isset($input['clientCode']) && !empty($input['clientCode']))
                  $cdCliente = $input['clientCode'];
              else
                  $cdCliente = $this->util->getDefaultClient($cdFilial);
              
              if(isset($input['orderType']))
                  $orderType = $input['orderType'];
              else
                  throw new \Exception('Mesa não informada.');
              
              if(isset($input['tableId'])){
                  $nrMesa = $input['tableId'];
              }
              else{
                  if($orderType ==='MES')
                      throw new \Exception('Mesa não informada.');
                  
                  $nrMesa = null;
              }
              if(isset($input['nrComanda'])){
                $nrComanda = $input['nrComanda'];
              }else{
                if($orderType ==='CMD')
                    throw new \Exception('Comanda não informada.');

                $nrComanda = null;
              }

              $check = $orderType == 'MES' ? 'CMD' : 'MES';
  
              $this->orderService->checkTable($cdFilial, $cdLoja, $nrMesa, $check);
              $statusTable = $this->orderService->getStatusTable($cdFilial, $cdLoja, $nrMesa);
              if($statusTable['IDSTMESAAUX'] != 'D')
                  throw new \Exception('Mesa indisponível.');

              $statusCmd = $this->orderService->getOrderDscomanda($nrComanda, $cdFilial);
              if(!empty($statusCmd))
                  throw new \Exception('Comanda indisponível.');
              
              $vendarestSeq  = $this->orderService->generateVendarestSeq($cdFilial);
              $comandavenSeq = $this->orderService->generateComandavenSeq($cdFilial);
              $orderType = $this->validateOrderType($orderType, false, false);
              $dscomanda  = isset($input['nrComanda']) ? $input['nrComanda'] : null;
              $toGo = 'N';
              if (empty($dscomanda)) {
                  if ($orderType == "MES_MOB" || $orderType == "CMD_MOB") {
                      $dscomanda = "PKR_$comandavenSeq";
                  } else {
                      $dscomanda = "DLV_$comandavenSeq";
                  }
              } else {
                  $dscomanda = str_pad($dscomanda, 10, 0, STR_PAD_LEFT);
              }
  
              $ordeData = $this->orderService->openTable($cdFilial, $cdLoja, $nrMesa, $personalQtd, $cdCliente, $vendarestSeq, $comandavenSeq, $orderType, $dscomanda, $toGo);
              $this->util->commit();
              
              if ($orderType === 'CMD') {
                  $generateXmlCat = $this->orderService->generateXmlCatraca($cdFilial, $dscomanda, $comandavenSeq, 'C');
                  if ($generateXmlCat['error']) {
                      $this->util->saveLog("Erro ao gerar XML da catraca. {$generateXmlCat['message']}", 'Error', 'OPEN_CMD-');
                  }
              }

              $response->addDataSet(
                new DataSet(
                    '',
                    array(
                      $ordeData
                    )
                  )
              );
              
              $response->addMessage(new Message("Mesa/Comanda Aberta com sucesso.", self::GENERAL_MESSAGE));
      } catch (\Exception $e) {
              $this->util->rollBack();
              $message = 'Erro ao abrir mesa/comanda. ' . $e->getMessage();
              $response->addMessage(new Message($message, self::ERROR_MESSAGE));
          }
    }

    public function getSalesByPeriod(Request $request, Response $response) {
      ini_set('memory_limit', '512M');
      
          try {
  
              $today = date("d/m/Y H:i:s");
              $timeLimit = date("d/m/Y H:i:s",strtotime('-30 minutes'));
              $orders = $this->orderService->getLatestSales($today, $timeLimit);
  
              $response->addDataSet(
                  new DataSet(
                      '',
                      array($orders)
                  )
              );
              
          } catch (\Exception $e) {
              $message = 'Erro ao buscar vendas. ' . $e->getMessage();
              $response->addMessage(new Message($message, self::ERROR_MESSAGE));
          }
    }

    public function sellerLogin(Request $request, Response $response) {
        ini_set('memory_limit', '512M');
        $inputJSON = file_get_contents('php://input');
        $input = json_decode($inputJSON, true);

        try {

            if(isset($input['Password']) && !empty($input['Password']))
                $passwd = $input['Password'];
            else
                throw new \Exception('Password inválido ou não informado.'); 
        
            if(isset($input['cashierCode']) && !empty($input['cashierCode'])){
                $user = $input['cashierCode'];
                $user = str_pad($user, 12, 0, STR_PAD_LEFT);
                $userDetails = $this->orderService->getCashierByCode($user);
            }elseif(isset($input['sellerCode']) && !empty($input['sellerCode'])){
                $user = $input['sellerCode'];
                $user = str_pad($user, 4, 0, STR_PAD_LEFT);
                $userDetails = $this->orderService->getCashierBySellerCode($user);
            }elseif(isset($input['email']) && !empty($input['email'])){
                $user = $input['email'];
                $userDetails = $this->orderService->getCashierByEmail($user);
            }else
                throw new \Exception('Usuário inválido ou não informado.'); 
            
            if(!$userDetails)
                throw new \Exception('Nenhum operador encontrado com os dados fornecidos.'); 
            
            $passMD5 = $userDetails['CDSENHAOPERMD5'];
            $passOperWeb = $userDetails['CDSENHAOPERWEB'];
            $passcriptoPlus = $userDetails['CDSENHOPER'];

            if($passMD5 !== md5($passwd)){
                if($passOperWeb !== md5($passwd)){
                    if($passcriptoPlus !== $this->orderService->encryptDelphi($passwd))
                        throw new \Exception('Senha inválida');
                }
            }
            
            $return = array(
                'cashierCode'=> $userDetails['CDOPERADOR'],
                'sellerCode' => $userDetails['CDVENDEDOR']            
            ); 
            
            $response->addDataSet(
                new DataSet(
                    '',
                    array($return)
                )
            );
            
        } catch (\Exception $e) {
            $message = 'Erro no login de operador - vendedor. ' . $e->getMessage();
            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
        }
        
      }
    public function sendingDanfeByEmail(Request $request, Response $response){
        ini_set('memory_limit', '512M');
        $this->util->beginTransaction();
        
      try {
        $inputJSON = file_get_contents('php://input');
        $input = json_decode($inputJSON, true);

        if(isset($input['subsidiaryCode']))
            $subsidiaryCode = $input['subsidiaryCode'];
        else
            throw new \Exception('Código da filial não informado.'); 
        
        if(isset($input['posCode']))
            $posCode = $input['posCode'];
        else
            throw new \Exception('Código do caixa não informado.');
        
        if(isset($input['saleSeq']))
            $saleSeq = $input['saleSeq'];
        else
            throw new \Exception('Sequencial da Venda não informado.');
        
        if(isset($input['nfceCode']))
            $nfceCode = $input['nfceCode'];
        else
            throw new \Exception('Cupom não informado não informado.');
        
        if(isset($input['email']))
            $email = $input['email'];
        else
            throw new \Exception('Email não informado não informado.');

        $checkSaleByEmail = $this->orderService->checkSaleByEmail($subsidiaryCode, $posCode, $saleSeq, $nfceCode, $email);
        if(isset($checkSaleByEmail['error']) && $checkSaleByEmail['error'])
            throw new \Exception($checkSaleByEmail['message']);
        $updateSale = $this->orderService->updateSaleEmail($subsidiaryCode, $posCode, $saleSeq, $nfceCode, $email);
        $saleWasExported = $this->orderService->saleWasExported($subsidiaryCode, $posCode, $saleSeq, $nfceCode);

        if(isset($saleWasExported['error']) && $saleWasExported['error'])
            throw new \Exception('Erro. ');

        if($saleWasExported['data']['IDIMPVENDA'] == 'S')
            throw new \Exception('Falha na alteração de status da venda. ');

        $this->util->commit();
        $response->addMessage(new Message("Cupom inserido na fila para envio por email.", self::GENERAL_MESSAGE));
      } catch (\Exception $e) {
            $this->util->rollBack();
            $message = 'Erro ao inserir cupom para envio por email. ' . $e->getMessage();
            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
        }
    }

    public function saveOrder(Request $request, Response $response)
    {
        $inputJSON = file_get_contents('php://input');
        $input     = json_decode($inputJSON, true);

        try {
            if (!isset($input))
                throw new \Exception('Não foram recebidos parâmetros');
        } catch (\Exception $e) {
            $message = 'Erro ao inserir comanda delivery. ' . $e->getMessage();
            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
            return;
        }

        

        $data   = array();
        $params = $input;

        $order    = $params['order'];
        $fileName = date("Ymd-H-i-s") . "-";
        $pagerId  = (isset($order['pagerId']) && !empty($order['pagerId'])) ? $order['pagerId'] : (string)$this->geraSenhaProducao($order['storeCode'], $order['subsidiaryCode'], $order['posCode']);
        $fileName .= empty($pagerId) ? $order['subsidiaryCode'] : $pagerId;

        $this->util->saveLogOrder($inputJSON, 'Order/' . date("Y/m/d"), $fileName);

        $subsidiaryCode = $order['subsidiaryCode'];
        $comandavenSeq  = isset($order['saleCode']) ? $order['saleCode'] : null;
        if(empty($comandavenSeq)){
            throw new \Exception("SaleCode não informdo!");
        }

        $dscomanda = isset($order['orderId']) ? $order['orderId'] : null;
        $nrmesa    = isset($order['tableNumber']) ? $order['tableNumber'] : null;
        
        if((empty($dscomanda)) || (empty($nrmesa))){
            $orderDetails = $this->orderService->getComandaData($comandavenSeq, $subsidiaryCode);
            $nrmesa       = $orderDetails['NRMESA'];
            $dscomanda    = $orderDetails['DSCOMANDA'];
            $vendarestSeq = $orderDetails['NRVENDAREST'];
        }

        $now = strtotime("now");

        $this->util->beginTransaction();

        try {
            $consumerCode     = !empty($order['consumerCode']) ? $order['consumerCode'] : null;
            $clientCode       = $order['clientCode'];
            $billParams       = $order['orderDescription'];
            $orderType        = $billParams['orderType'];
            $orderDiscount    = isset($billParams['orderDiscount']) ? $billParams['orderDiscount'] : 0;
            $holdTime         = isset($billParams['holdTime']) ? $billParams['holdTime'] : 0;
            $productsValue    = $billParams['productsValue'];
            $remaingValue     = $billParams['productsValue'];
            $toTravel         = !empty($billParams['toTravel']);
            $driveThru        = !empty($billParams['driveThru']);
            $orderDescription = isset($billParams['description']) ? $billParams['description'] : null;
            $sellerCode       = isset($order['sellerCode']) ? $order['sellerCode'] : null;
            $closeByApi       = isset($order['closeByApi']) ? $order['closeByApi'] : false;
            $items            = $order['items'];

            $closeByApi = !empty($closeByApi) ? 'S' : 'N';

            $standardTable = null;
            if (!empty($nrmesa)) {
                $this->orderService->validateTable($nrmesa);
            }

            // --- Consumer description
            $consumerDescription = isset($order['consumerId']) ? $order['consumerId'] : null;

            // verifica a existência da filial
            if (!$this->util->subsidiaryExists($subsidiaryCode)) {
                throw new \Exception('Filial inexistente.');
            }

            if (!empty($clientCode) && !empty($consumerCode)) {
                if (!$this->util->consumerExists($clientCode, $consumerCode)) {
                    throw new \Exception('Consumidor inexistente.');
                }
            } else {
                $defaultClient = $this->util->getDefaultClient($subsidiaryCode);

                if (!empty($defaultClient))
                    $clientCode = $defaultClient;
                else
                    throw new \Exception('Cliente padrão inexistente.');
            }

            $orderType     = $this->validateOrderType($orderType, $toTravel, $driveThru);
            $storeCode     = null;
            $sellerPos     = null;
            $idposobsped   = null;
            $idlugarmesa   = null;
            $modalidade    = substr($orderType,0,1);

            if (isset($order['storeCode'])) {
                $storeCode   = $order['storeCode'];
                $storeReturn = $this->util->getStoreDetails($subsidiaryCode, $storeCode);

                if (!empty($storeReturn)) {
                    if (empty($standardTable)) $standardTable = $storeReturn['NRMESA'];
                    $idposobsped = $storeReturn['IDPOSOBSPED'];
                    $idlugarmesa = $storeReturn['IDLUGARMESA'];
                }
            }

            if (isset($order['posCode'])) {
                $posCode   = $order['posCode'];
                $sellerPos = $posCode;
            }

            if (empty($sellerCode))
                $sellerReturn = $this->orderService->getStandardSeller($subsidiaryCode);
            else
                $sellerReturn = $this->orderService->getVendorInfo($sellerCode);

            if (!$sellerReturn['error']) {

                $standardSeller = $sellerReturn['data'];
                $sellerCode     = $standardSeller['CDVENDEDOR'];
                $operatorCode   = $standardSeller['CDOPERADOR'];

                if ($storeCode == null)
                    $storeCode = $standardSeller['CDLOJA'];

                if ($sellerPos == null)
                    $sellerPos = $standardSeller['CDCAIXA'];

                if ($standardTable == null)
                    $standardTable = $standardSeller['NRMESA'];

                if ($idposobsped == null)
                    $idposobsped = $standardSeller['IDPOSOBSPED'];

                if ($idlugarmesa == null)
                    $idlugarmesa = $standardSeller['IDLUGARMESA'];

                $table = !empty($nrmesa) ? $nrmesa : $standardTable;

                if (!empty($dscomanda))
                    $oderByDscomanda = $this->orderService->getOrderDscomanda($dscomanda, $subsidiaryCode);

                if ($this->isToGo($orderType)) {
                    $toGo = 'S';
                    $subsidiaryAddress = $this->util->getSubsidiaryAddress($subsidiaryCode, self::PRINCIPAL_ADDRESS);

                    if (!empty($subsidiaryAddress)) {
                        $cityCode      = $subsidiaryAddress['CDMUNICIPIO'];
                        $state         = $subsidiaryAddress['SGESTADO'];
                        $country       = $subsidiaryAddress['CDPAIS'];
                    } else {
                        $cityCode      = null;
                        $state         = null;
                        $country       = null;
                    }

                    $zipCode           = null;
                    $streetAddress     = null;
                    $neighborhood      = null;
                    $number            = null;
                    $reference         = null;
                    $phone             = null;
                    $cellphone         = null;
                    $addressLine2      = null;
                    $fullStreetAddress = null;
                    $deliveryFee       = 0;
                } else {
                    $toGo           = 'N';
                    $addressDetails = $billParams['deliveryAddress'];
                    $state          = $addressDetails['state'];
                    $cityCodeExt    = $addressDetails['city'];
                    $cityCode       = $this->util->getCityCode($cityCodeExt, $state);
                    $zipCode        = $addressDetails['zipCode'];
                    $streetAddress  = $addressDetails['streetAddress'];
                    $neighborhood   = $addressDetails['neighborhood'];
                    $number         = $addressDetails['number'];
                    $reference      = $addressDetails['reference'];
                    $phone          = $addressDetails['phone'];
                    $cellphone      = $addressDetails['cellphone'];
                    $addressLine2   = $addressDetails['addressLine2'];
                    $deliveryFee    = (float)$billParams['deliveryFee'];
                    $fullStreetAddress = $streetAddress . ', ' . $number;
                    $country        = '0055';
                    $phoneToUpdate  = !empty($phone) ? $phone : $cellphone;
                    $this->util->updateConsumerPhone($clientCode, $consumerCode, $phoneToUpdate);
                }

                $idcontroprod = $this->getIdcontroprod($storeCode, $subsidiaryCode);
                $IDSTCOMANDA  = '3';

                if ($idcontroprod === 'S')
                    $IDSTCOMANDA = '1';

                $orderToken = sha1($subsidiaryCode . $vendarestSeq . $comandavenSeq . $clientCode . $consumerCode);

                $scheduledDate = null;
                $isScheduled   = $billParams['isScheduled'];
                $scheduledDate = $isScheduled ? $billParams['scheduledDate'] : $scheduledDate;
                $IDSTCOMANDA   = $isScheduled ? self::SCHEDULED_ORDER : $IDSTCOMANDA;
                $syncScheduled = $isScheduled ? self::PENDING_SCHED_ORDER : self::NOT_PENDING_SCHED_ORDER;
                $kdsChangeStatusOnExp = '0';

                // ******Validar Usabilidade modalidades Mesa/Comanda******
                /*
                $taaOrder = strpos($orderType, 'TAA') !== false;

                if ($this->printMode == 'EXPEDICAO') {

                    if ($toGo == 'S' || $taaOrder)
                        $kdsChangeStatusOnExp = '1';
                } else if ($this->printMode == 'INICIO') {
                    if ($taaOrder)
                        $kdsChangeStatusOnExp = '2';
                    else if ($toGo == 'S')
                        $kdsChangeStatusOnExp = '1';
                }*/

                // ---> new saveItems <---
                $itemsArray = $this->orderService->formatItems($items, $orderType, $subsidiaryCode, $vendarestSeq, $comandavenSeq, $storeCode, $idcontroprod, $sellerPos, $sellerCode, null, null, $pagerId, $holdTime);
                if($orderDiscount > 0){
                    $this->orderService->shareDiscount($orderDiscount, $itemsArray['ITCOMANDAVEN']);
                }
                $this->orderService->saveOrderItems($itemsArray);

                $this->util->commit();

                if (!empty($consumerDescription) && !empty($pagerId))
                    $pagerId .= " - $consumerDescription";
                
                $this->util->beginTransaction();
                $printItems = $this->formatItemsToPrint($items);
                
                $toGoString = ($toGo == 'S') ? '' : 'PEDIDO PARA ENTREGA DELIVERY';

                
                $this->printOrder($subsidiaryCode, $comandavenSeq, $storeCode, $printItems, $sellerCode, $table, $pagerId, $modalidade ,$toGoString, null);
                $this->util->commit();
                
                $response->addDataSet(
                    new DataSet(
                        '',
                        array(
                            array(
                                'saleId' => $orderToken,
                                'saleCode' => $comandavenSeq,
                                'pagerId' => $pagerId
                            )
                        )
                    )
                );
            }
        } catch (\Exception $e) {
            $this->util->rollBack();

            $message = 'Erro ao inserir comanda delivery. ' . $e->getMessage();
            $this->util->saveLogOrder($message, 'Error/' . date("Y/m/d"), $fileName);

            $response->addMessage(new Message($message, self::ERROR_MESSAGE));
        }
    }

    public function geraSenhaProducao($cdloja, $cdfilial, $cdcaixa) {
		
		$IDSENHACUP = $this->orderService->getIdSenhaCup($cdloja, $cdfilial, $cdcaixa);
        $maxSeqDia = $this->orderService->getNrMaxSenhaPed($cdloja, $cdfilial, $cdcaixa);
		$IDSENHACUP = $IDSENHACUP['IDSENHACUP'];

		if ($IDSENHACUP === 'S') {
			$today = date("d/m/Y");
            
			return $this->orderService->geraSeqCdsenhaped($cdfilial, $today, $maxSeqDia);

		} else if ($IDSENHACUP === 'A' || $IDSENHACUP === 'L') {

            $CDSENHAPED = null;
			$getSenhaPed = $this->orderService->getCdSenhaPed($cdloja, $cdfilial);
            if (isset($getSenhaPed['CDSENHAPED'])) {
			    $CDSENHAPED = array_map("intval", array_column($getSenhaPed, 'CDSENHAPED'));
            }

			$max = $IDSENHACUP === 'A' ? 99999 : 999;
            $nrMinSenhaAl = $this->orderService->getNrMinSenhaAl($cdloja, $cdfilial, $cdcaixa);
            if ($nrMinSenhaAl['NRMINSENHAAL'] != null) {
              return $this->geraSenhaPedido($max, $CDSENHAPED, $nrMinSenhaAl['NRMINSENHAAL']);
            }

			$nrPedido = $this->geraSenhaPedido($max, $CDSENHAPED);

			return $nrPedido;
		}
	}

	private function geraSenhaPedido($max, $CDSENHAPED, $min=1) {
        $min = $max > $min ? $min : 1;
		$randomNrpedido = mt_rand($min, $max);

		if (empty($CDSENHAPED)) 
            return $randomNrpedido;

		if (!in_array($randomNrpedido, $CDSENHAPED)) {
			return $randomNrpedido;

		} else {
			$nrPedido = $this->geraSenhaPedido($max, $CDSENHAPED);
		}

		return $nrPedido;
	}
}
